/*
 * ActionsGUIInterface.h
 * John Davin
 *
 * Sends commands to the Actions display GUI through the message queue
 * This is not part of the actionsgui program, it's basically a short 
 * function library for the player and the logplayer to use to send
 * commands to the GUI. player and logplayer need to compile ActionsGUIInterface.C 
 * on their own. 
 */

#ifndef ACTIONSGUIINTERFACE_H
#define ACTIONSGUIINTERFACE_H


/* If you don't want to use the GUI, and don't want to waste time sending commands 
 * to the queue, comment out the following line. 
 * CoachablePlayer will still call the functions, but the function bodies will be empty 
 */
#define USE_GUI


/* The following is a hack to get g++ to allow us to redefine struct msgbuf (in ActionsGUIInterface.C) 
 * over the definition in sys/msg.h. 
 *
 */
#ifdef __USE_GNU
#undef __USE_GNU
#include <sys/msg.h>
#define __USE_GNU
#else
#include <sys/msg.h>
#endif

#include <iostream>
#include <fstream>
using namespace std;

#define NUM_PLAYERS 11

#define GUI_MSGSIZE 128 // includes fgcolor, bgcolor, msg, playernum

#define GUI_MAX_CYCLES 10000 // since 0 is a valid cycle #, GUI_MAX_CYCLES-1 is the max cycle #

#define GUI_COLOR_LEN 15

#define GUI_ACTIONSRC_LEN 5

#define GUI_MAX_LOG_LINE (GUI_MSGSIZE+4)
// the +4 is for cyclenum and spaces

#define LOG_FILENAME_LEN 30

#define LOG_DIR "gui_logs/"
#define LOG_FILENAME "actionslog.log" // playernum will be prepended. 



class ActionsGUIInterface {

 public:
  ActionsGUIInterface(bool writelogs = false, bool usegui = false);
  ~ActionsGUIInterface();

  /* send( ):
   * params: 
   * cyclenum - the cycle number of this message
   * actionsource - a character indicating the action source (eg, C for Coach, R for rule file)
   * msg  - the text you want to send (eg, "intercept")
   * fgcolor - the text color, 
   * bgcolor - the background color. 
   * 
   * Note: Color names must be in your system's .../X11/rgb.txt file.
   * NULL specifies to use the default color. 
   * Current implementation is limited to one word colors though. */
  /* also, cyclenum is only used by the players to specify the current cyclenum used
   * for logging. pnum is the playernum and is only specified by logplayer (because players
   * set their playernum with setPlayernum). */
  void send(int cyclenum, char *actionsource, char *msg, 
	    char *fgcolor=NULL, char *bgcolor=NULL, int pnum=-1);

  /* 
   * This is only used by player and is mainly for speed/convenience. 
   * We could specify playernum every call to send(), but it's better to just 
   * set it this way upon player initialization. 
   * This also opens the logfile if the player is using one. 
   */
  void setPlayernum(int num);



 private: 
  int playernum;
  int msgid; // id of message queue
  int errorcnt; // just used to stop printing errors when we go over some #

  bool WriteLogs; // set whether we log messages to actions log files. 
  ofstream logfile; // the logfile we'll write to

  bool useGUI; // sets whether ActionsGUIInterface will actually send commands

};



#endif
