/*
  * Setup the message queue in this thread and run a loop to continually check the queue
  * Handle the messages received from the queue. 
  *
  * Author: John Davin
  */
#include <sys/types.h>
#include <sys/ipc.h>
#include <stdio.h>
#include <iostream>
using namespace std;

#include <qtable.h>
#include "CheckQueue.h"
#include "ColorTableItem.h"
#include "ActionsGUIInterface.h"


#define QUEUE_KEY 1234

/* 
 * Initialize message queue
 *
 */
CheckQueue::CheckQueue(ActionsDisplay *ptr, QTimer *timer)
  : actionsdisp(ptr), Timer(timer)
{
  idleInterval = 0;
  for(int p = 0; p < NUM_PLAYERS; p++)
    lastreceived[p] = 0;
   
  if ((MsqID = msgget((key_t)QUEUE_KEY, 0666 | IPC_CREAT)) < 0) {
    perror("msgget");
    cout << strerror(errno) <<endl;
    exit(1);
  }
  cout << "opened queue with id "<<MsqID <<endl;
  
}


CheckQueue::~CheckQueue(){ }


/* 
 * Init queue: If queue has entries, flush the queue by deleting it and then recreating it
 */
void CheckQueue::init(){
  struct msqid_ds buf;
  message_buf  rbuf;
    
  // first check if the queue already has stuff  
  msgctl(MsqID, IPC_STAT, &buf);
  cout << "Queue has "<<buf.msg_qnum<<" entries. last one sent at: "<<buf.msg_stime<<endl;
  if(buf.msg_qnum > 1){
    cout << "flushing queue"<<endl;

    for(unsigned int i=0; i<buf.msg_qnum; i++)
      msgrcv(MsqID, &rbuf, MSGSIZE, 1, IPC_NOWAIT);

    cout << "Done flushing queue."<<endl;

    /* Don't delete queue b/c a player may already have a msgid for it 
    msgctl(MsqID, IPC_RMID, &buf);
    if ((MsqID = msgget(QUEUE_KEY, 0666 | IPC_CREAT)) < 0) {
      perror("msgget");
      cout << strerror(errno) <<endl;
      exit(1);
    }
    cout << "reopened queue with id "<<MsqID <<endl;
    */
  }
}

/*
 * get one message from the queue
 *
 */
void CheckQueue::run(){
  struct msqid_ds buf;
  message_buf  rbuf;	

  /* first check how many entries are in queue */ 
  msgctl(MsqID, IPC_STAT, &buf);
  if(buf.msg_qnum >= NUM_PLAYERS){
    cout << "[CheckQueue]: We're behind on reading queue! Entries in queue: "<<buf.msg_qnum<<endl;
  }

  if(time(NULL) - buf.msg_stime > 3){ // haven't received anything in 3 seconds
    // so probably players are not executing
    // slow down our idle checking to not consume processor. 
    idleInterval = 3000; //milliseconds
    Timer->changeInterval(idleInterval);

  }
  else{
    if(idleInterval != 0){
      idleInterval = 0;
      Timer->changeInterval(idleInterval);
    }
  }

  /* now check if any of our player cells have not been updated in a while
   * (indicating that we are not receiving messages from the player and should
   * gray out the cell to warn the user) */
  for(int p = 0; p < NUM_PLAYERS; p++){
    if( lastreceived[p] - time(NULL) >= 1){
      cout << "[CheckQueue]: Player "<<p+1<<" has not updated in at least 1 second."<<endl;
      applyStyle(p-1, actionsdisp->table->text(p-1,1), 
		 actionsdisp->table->text(p-1,0), GRAYED_OUT, "-");
    }
  }
  

  // query msgrcv for a message of type 1 
  // this call is non-blocking (IPC_NOWAIT) so that 
  // we won't halt redrawing of the GUI. 
  if (msgrcv(MsqID, &rbuf, MSGSIZE, 1, IPC_NOWAIT) < 0) {
    if(errno != ENOMSG) // ignore error from IPC_NOWAIT. 
      perror("msgrcv error");
    return;
  }

  handleMsg(rbuf.mtext);
  //printf("%s\n", rbuf.mtext);
  //printf(".");fflush(stdout);

}


/*
  * msg should be of format: %d %s %s %s message
  * where %d is the player number, the 1st %s is the action source 
  * (C for coach, D for default, R for rule file), 2nd %s is the foreground color, 
  *  and the 3rd %s is background color.
  * Ex. "02 blue white shoot"  would mean show "shoot" in player 2 cell 
  *  with blue text and white background.
  * "-" indicates no color (default). 
  */
void CheckQueue::handleMsg(char *msg){
    int playernum;
    char actionsrc[GUI_ACTIONSRC_LEN];
    char fgcolor[16];
    char bgcolor[16];
    
    //printf("got %s\n", msg);
    sscanf(msg, "%d", &playernum);
    msg += 2; // skip past playernum

    sscanf(msg, "%5s", actionsrc);
    msg += strlen(actionsrc) + 1; // +1 is to skip over the space

    sscanf(msg, "%15s", fgcolor);
    msg += strlen(fgcolor) + 1; // +1 is to skip over the space
    
    //printf("fgcolor: %s: %d\n", fgcolor, strlen(fgcolor));
    sscanf(msg, "%15s", bgcolor);
    msg += strlen(bgcolor) +1;
    //printf("bgcolor: %s: %d\n", bgcolor, strlen(bgcolor));
    
    if(playernum < 1 || playernum > 11){
	cerr << "[handleMsg]: invalid playernum "<<playernum <<endl;
	return;
    }	
    
    applyStyle(playernum-1, actionsrc, msg, fgcolor, bgcolor);
    lastreceived[playernum-1] = time(NULL); // update last received time.
    
}

/* 
  * Apply the style specified by 'code' to the cell at row 'row'
  */
void CheckQueue::applyStyle(int row, const char *actionsrc, const char *msg, char *fgcolor, char *bgcolor){
    QColor fg;
    QColor bg;
    QColor blank = QColor();
    
    if(!strcmp(fgcolor, "-"))
	fg = QColor(); // invalid color
    else
	fg = QColor(fgcolor);
    if(!strcmp(bgcolor, "-"))
	bg = QColor(); // invalid color
    else
	bg = QColor(bgcolor);
 
    ColorTableItem *it = new ColorTableItem(actionsdisp->table,QTableItem::Never, msg, fg, bg);
    actionsdisp->table->setItem(row,0, it);

    ColorTableItem *actionSrcItem = new ColorTableItem(actionsdisp->table,
						       QTableItem::Never,actionsrc, blank,blank);
    actionsdisp->table->setItem(row,1, actionSrcItem);

    
}



