/* Dongryeol Lee
 *
 * ActionPicker.C
 *
 * This file implements the ActionPicker Class defined in the 
 * ActionPicker.h file.
 */

#include <iostream>
#include <string>
#include <stdlib.h>
#include <stdio.h>
#include "clangaction.h"
#include "ActionPicker.h"
#include "config.h"
#include "DirVisitor.h"
#include "msgrulestore.h"
#include "LevelStorage.h"
#include "Filter.h"
#include "ActionVisitor.h"
#include "ReductionActionVisitor.h"
#include "ScoringActionVisitor.h"
#include "WorldModelAccessor.h"
#include "Logger.h"
#include "BasicPlayer.h"

using rcss::clang;


extern Filter filter;
extern LevelStorage levels;
extern WorldModelAccessor MyWorldModel;



/* Default constructor for ActionPicker */
ActionPicker::ActionPicker()
  : time_last_home_set()
{
  /* Intialize the previously executed home advice action ID */
  prevHomeID = 0;
  

  /* Initialize the list of default actions */
  

  /* Default Action 1: Dribble toward the opponent's goal */
  defaultActions.push_back(ActionBinding
			   (new ActDribble
			    (std::auto_ptr <Region>	
			     (new RegRec
			      (std::auto_ptr<Point>
			       (new PointSimple(40.0, -20.0)), 
			       std::auto_ptr<Point>
			       (new PointSimple(52.0, 20.0))))),
			    VarBindings(), ActionBinding::DEFAULT));


  /* Default Action 2: Intercept ball */
  defaultActions.push_back(ActionBinding(new ActIntercept(), VarBindings(),
					 ActionBinding::DEFAULT));


  /* Default Action 3: Shoot */
  defaultActions.push_back(ActionBinding(new ActShoot(), VarBindings(),
					 ActionBinding::DEFAULT));


}

/*
 * This is called each evaluation (in pickActions()) to add 
 * actions based on communicated information (eg, WantPass messages). 
 */
void ActionPicker::addCommunicationActions(){

  // RECENT_TIME is the number of cycles that we will consider a world model
  // update to be current/relevant. 
#define RECENT_TIME 5

  /////// if a teammate wants a pass, add an action to pass to that player ///////

  // only use if a player requested a pass within last 5 (RECENT_TIME) cycles
  Time lasttime = trilearn_wm->getWantPassTime();
  if(lasttime != -1 && lasttime >= MyWorldModel.getCurrentTime() - RECENT_TIME){

    ObjectT passToObj = trilearn_wm->getWantPassPlayer();
    assert(passToObj != OBJECT_PLAYER_UNKNOWN);
    UNum::unum_t passTo = MyWorldModel.getUNumFromObjectT(passToObj);

    LOGOUT(50) << passTo << " wants a pass, so adding a passto action"<<endl;
    addAction(new ActPassUNum(UNumSet(passTo)), 
	      VarBindings(), ActionBinding::COMMUNICATION );
  }


  /////// if a teammate is passing to us, add an action to intercept ball /////

  // only use if a player said he's passing to us within last 5 cycles
  lasttime = trilearn_wm->getPassToPlayerTime();
  if(lasttime != -1 && lasttime >= MyWorldModel.getCurrentTime() - RECENT_TIME){

    ObjectT passer = trilearn_wm->getPassToPlayerSender();
    assert(passer != OBJECT_PLAYER_UNKNOWN);

    char passername[10];
    SoccerTypes::getObjectStr(passername, passer, "");
    LOGOUT(50) << passername << " is passing to us, so adding an intercept action"<<endl;
    // getPassToPlayerSender() is currently not used. just printing it for 
    // informational/debugging purposes


    addAction(new ActIntercept(), 
	      VarBindings(), ActionBinding::COMMUNICATION );
  }
  //LOGOUT(50) << "[ActionPicker]: "<<trilearn_wm->getCurrentTime()
  //    <<": PassToPlayerTime="<<trilearn_wm->getPassToPlayerTime()<<endl;


  ////// or if a teammate is passing to a point close to us, intercept ball /////

  // only use if a player said he's passing to coord within last 5 cycles
  lasttime = trilearn_wm->getPassToCoordTime();
  if(lasttime != -1 && lasttime >= MyWorldModel.getCurrentTime() - RECENT_TIME){

    ObjectT passer = trilearn_wm->getPassToCoordSender();
    assert(passer != OBJECT_PLAYER_UNKNOWN);

    VecPosition coord = trilearn_wm->getPassToCoord();

    // getPassToPlayerSender() is currently not used. just printing it for 
    // informational/debugging purposes

    // if we're within 10 units of the coordinate, assume the pass 
    // was intended for us and try to intercept it. 
    if( coord.getDistanceTo(trilearn_wm->getAgentGlobalPosition()) < 10.){
      char passername[10];
      SoccerTypes::getObjectStr(passername, passer, "");
      LOGOUT(50) << passername << " is passing to coord " <<coord
	   <<", so adding an intercept action"<<endl;

      addAction(new ActIntercept(), 
		VarBindings(), ActionBinding::COMMUNICATION );
    }
  }
  //  LOGOUT(50) << "[ActionPicker]: "<<trilearn_wm->getCurrentTime()
  //    <<": PassToCoordTime="<<trilearn_wm->getPassToCoordTime()<<endl;

}


/* Push the action into the list of storage */
void ActionPicker::addAction(Action *addThisAction, 
			     const VarBindings &bindings, 
			     ActionBinding::SourceType source)
{
  listOfActions.push_back(ActionBinding(addThisAction, bindings, source));
}


const std::vector <ActionBinding> ActionPicker::reduceAndFilter()
{
  /* New advice set after filtering */
  vector <ActionBinding> newAdviceSet;


  /* First, go through each advice in the advice set and apply reductions */

  /* Retrieve the number of advice */
  int numAdvice = listOfActions.size();

  
  //actionlog(0) << spades::ende << "Now picking action..." << spades::ende;
  actionlog(10) << "Before filtering: " << numAdvice << " available" <<spades::ende;
  

  
  //actionlog(10) << "Now reducing and filtering actions..." << spades::ende;
  

  /* Go through each advice */
  for(int i = 0; i < numAdvice; i++) {


    /* Declare ReductionActionVisitor object */
    ReductionActionVisitor r(listOfActions[i]);
    
    
    /* Call accept on the visitor to apply reduction */
    (listOfActions[i].theaction)->accept(r);
    
    /* Apply reduction to the current advice */
    vector <ActionBinding> atomizedList = r.getAtomizedActions();
    
    
    /* Number of atomized actions */
    int numAtomized = atomizedList.size();
    
    
    
    /*
    actionlog(20) << *(listOfActions[i].theaction) << " in source type: " <<
      listOfActions[i].sourceType <<
      " generated " << numAtomized << " actions" << spades::ende;
    */
    
    
    /* Now go through our atomized list, */
    for(int j = 0; j < numAtomized; j++) {

      
      /*
      actionlog(30) << "Considering the atomized " << 
      *(atomizedList[j].theaction);
      */


      /* Add to our new advice set if passes the filtering */
      if( !filter.shouldBeFiltered(atomizedList[j]) ) {
	
	//actionlog(40) << "Passed!!!" << spades::ende;


	newAdviceSet.push_back(atomizedList[j]);
      }
      else{
	actionlog(30) << "Action filtered out: "<<*atomizedList[j].theaction<<spades::ende;
      }

    }
      
  }

  /* First clear our old Advice list */
  clear();

  actionlog(10) << "After filtering: " << newAdviceSet.size() << " available" <<spades::ende;

  /* Return our atomized/filtered action advice list */
  return newAdviceSet;
}

/* Consider the atomized list and try to match with the given level:
   Returns true if matching was successful */
bool ActionPicker::considerAdviceSet(const int &level)
{
  /* The best advice encountered so far */
  ActionBinding bestAbinding;
  
  /* The best score encountered so far */
  double bestScore = -1;
  
  /* Retrieve the size of the advice set */
  int numNewAdvice = listOfActions.size();

  
  /* Consider each advice in the current advice set */
  for(int adv = 0; adv < numNewAdvice; adv++) {
    
    /* Retrieve the current advice in consideration */
    ActionBinding currAbinding = listOfActions[adv];


    actionlog(30) << level<< ": Now considering " << *currAbinding.theaction <<
      spades::ende;
    
    
    /* If the current advice is relevant to the current level, then */
    if( levels.matchLevel(currAbinding, level ) ) {
      
      //actionlog(30) << "Matched!!!!!!!!!!" << spades::ende;
            
      /* Declare ScoringActionVisitor */
      ScoringActionVisitor s(currAbinding.sourceType);
      
      /* Score the advice */
      (currAbinding.theaction)->accept(s);
      
      /* Retrieve the score */
      double currScore = s.getScore();
      
      //actionlog(40) << "Score of: " << currScore << spades::ende;
      //if(currScore == -1.)
      //	actionlog(30) <<"Score of -1 so rejecting action: "
      //	      <<*currAbinding.theaction<<spades::ende;


      /* Replace the best advice/score set if better */
      if( currScore > bestScore ) {
	
	bestScore = currScore;
	bestAbinding = currAbinding;
      }
    }
  }
  
  /* If a best action is found, then immediately execute it */
  if( bestScore >= 0 )   {

    /* If this is a home advice action, ... */
    if(levels.getActionType(level) == ActionVisitor::HOME) {

      /* Get the ID of the home advice action */
      Action::ActionID currID = (bestAbinding.theaction)->getID();

      if (time_last_home_set == trilearn_wm->getCurrentTime())
        {
          // we already set the home region this cycle, so let's act as if this
          // doesn't match
          return false;
        }
      else
        {
          time_last_home_set = trilearn_wm->getCurrentTime();
          if (currID != prevHomeID)
            {
              /* This is the only action the agent should execute */
              MY_PLAYER->executeCommand(bestAbinding);
              //actionlog(30) << "Executing: "<<*bestAbinding.theaction << spades::ende;
	
              /* Set the previous Home ID */
              prevHomeID = currID;

              return true;
            }
          else
            {
              /* Set home region */
              SoccerCommand soc = 
                MY_PLAYER->setHomeRegion( (Region *) 
                                          ((ActHome *) (bestAbinding.theaction))
                                          ->getRegion());
              // If home didn't do anything, then let's let other levels have their say
              if (soc.commandType == CMD_ILLEGAL)
                return false;
              MY_PLAYER->ACT->putCommandInQueue(soc);
              return true;
            }
       }
    }
      
    /* If it is not a home advice level, just execute in normal way */
    else {
      actionlog(30) << "Executing: "<<*bestAbinding.theaction << spades::ende;

      MY_PLAYER->executeCommand(bestAbinding);
      
      return true;
    }
  }

  /* Otherwise, return false */
  return false;
}

/* Consider the list of actions in our default action list. */
bool ActionPicker::considerDefaultAdvice(const int& level) const
{
  /* Get the number of default advice */
  int numDefaultAdvice = defaultActions.size();
 

  /* If you are still here, try to match the default actions */
  for(int adv = 0; adv < numDefaultAdvice; adv++) {
    
    /* Retrieve the current advice in consideration */
    ActionBinding currAbinding = defaultActions[adv];

    
    actionlog(30) << level<< ": Now considering default " << *currAbinding.theaction <<
      spades::ende;

    
    /* If the current advice is relevant to the current level, then */
    if( !filter.shouldBeFiltered(currAbinding) &&
	levels.matchLevel(currAbinding, level ) ) {

      actionlog(30) << "Match! Executing: " << *currAbinding.theaction<< spades::ende;

      /* Immediately execute it */
      MY_PLAYER->executeCommand(currAbinding);
      
      /* And return */
      return true;
	  
    }
  }

  /* This shouldn't happen since all the default actions should be in the
     storage, but this is for sanity check */
  return false;
}

  
/* Pick action from the list of actions: give preference to the previously
   picked action. If the ID of the previously selected exists within the action
   list, execute that action. Otherwise, randomize the choice */
void ActionPicker::pickAction()
{
  //pfr: big feakin hack
  Rectangle forbid_area(VecPosition(-42, 0), VecPosition(42, 11.3));
  if (MY_PLAYER->PS->getForbidMoveArea())
    {
      if (forbid_area.isInside(MY_PLAYER->WM->getAgentGlobalPosition()))
        {
          VecPosition new_pos = forbid_area.expand(1.0).getClosestEdgePoint(MY_PLAYER->WM->getAgentGlobalPosition());
          //std::cerr << "SMURF: " << MY_PLAYER->WM->getAgentGlobalPosition() << " -> " << new_pos << std::endl;
          MY_PLAYER->ACT->putCommandInQueue( MY_PLAYER->moveToPos(new_pos, MY_PLAYER->PS->getPlayerWhenToTurnAngle()) );
          cleanup();
          return;
        }
    }

  // add in a few actions that are based on communicated information
  addCommunicationActions();
 

  /* First, atomize and filter our list of advice in the ActionPicker */
  listOfActions = reduceAndFilter();


  /* Retrieve the number of levels defined */
  int numLevels = levels.getNumLevels();
  
  
  //actionlog(10) << "After: " << listOfActions.size() <<" advice is valid" << 
  //spades::ende << spades::ende;
  
  //actionlog(10) << "Now doing level matching..." << spades::ende;
  
  /* Now go through each advice and try to match a level, starting from 0 and
     onward */
  for(int l = 0; l < numLevels; l++)    {

    //actionlog(20) << "Now on Level " << l << " ";
    //actionlog(20) << levels.getLevel(l) << spades::ende;

    
    /* First determine if the current level is a default action level or not,
       if not, match with the advice in the atomzied advice set */    
    if( levels.getSourceType(l) != ActionBinding::DEFAULT)  {
      
      /* Consider our atomized/filtered advice set, and if a best action
	 has been executed, consider no more */
      if( considerAdviceSet(l) ){
	cleanup();
	return;
      }

    }
    
    /* If the level is a default action level, then go through our default
       action list */
    else {
      
      /* If a default action has been found, then break out */
      if( considerDefaultAdvice(l) ){
	cleanup();
	return;
      }

    }

  }

  actionlog(30) << "Executing default (NULL) action"<<spades::ende;
  /* Execute default action */
  MY_PLAYER->executeCommand(NULL);
  cleanup();

}

/* clear the action list vector */
void ActionPicker::clear()
{
  listOfActions.clear();
}

/* 
 * Helper function to pickAction()
 * frees memory in the listOfActions 
 */
void ActionPicker::cleanup(){
  // we need to delete dynamically allocated memory for the Communications actions
  for(unsigned int i=0; i<listOfActions.size(); i++){

    if(listOfActions[i].sourceType == ActionBinding::COMMUNICATION)
      delete listOfActions[i].theaction;

  }
  clear();
}
