/* Dongryeol Lee
 *
 * ApplicabilityActionVisitor.C
 *
 * ApplicabilityActionVisitor class is implemented here.
 *
 */


#include "ApplicabilityActionVisitor.h"
#include "WorldModelAccessor.h"
#include "clangutil.h"
#include <iostream>


#include "Logger.h"



/* Access to our player's world model */
extern WorldModelAccessor MyWorldModel;



/* Applicability function for ActPos actions */
void ApplicabilityActionVisitor::visit(rcss::clang::ActPos *)
{
  /* I think ActPos should have no restriction, but need to make sure */
  IsApplicable = true;
}

/* Applicability function for ActHome actions */
void ApplicabilityActionVisitor::visit(rcss::clang::ActHome *)
{
  /* No restriction in setting home region */
  IsApplicable = true;
}

/* This action is obsolete, so ignore */
void ApplicabilityActionVisitor::visit(rcss::clang::ActBallToReg *)
{
  IsApplicable = false;
}

/* Same with this action */
void ApplicabilityActionVisitor::visit(rcss::clang::ActBallToPlayer *)
{
  IsApplicable = false;
}

/* Applicability function for ActMark actions */
void ApplicabilityActionVisitor::visit(rcss::clang::ActMark *)
{
  IsApplicable = true; 
}

/* Applicability function for ActMarkLinePlayer actions */
void ApplicabilityActionVisitor::visit(rcss::clang::ActMarkLinePlayer *)
{
  IsApplicable = true;
}

/* This action will be reduced to ActMarkLinePlayer, so ignore */
void ApplicabilityActionVisitor::visit(rcss::clang::ActMarkLineReg *)
{
  IsApplicable = false;
}

/* Applicability function for ActOffsidesLine actions */
void ApplicabilityActionVisitor::visit(rcss::clang::ActOffsidesLine *)
{
  /* Maybe this is applicable anytime */
  IsApplicable = true;
}

/* This is an obsolete action, so ignore */
void ApplicabilityActionVisitor::visit(rcss::clang::ActHetType *)
{
  IsApplicable = false;
}

/* Ignore named action */
void ApplicabilityActionVisitor::visit(rcss::clang::ActNamed *)
{
  IsApplicable = false;
}

/* Same thing with ActMarkLineReg actions, this will be reduced to
   its UNUMSet equivalent actions, so ignore */
void ApplicabilityActionVisitor::visit(rcss::clang::ActPassReg *)
{
  IsApplicable = MyWorldModel.ballConfidenceGood() &&
    MyWorldModel.isBallKickable();
}

/* Applicability function for ActPassUNum actions */
void ApplicabilityActionVisitor::visit(rcss::clang::ActPassUNum *)
{

    if( MyWorldModel.ballConfidenceGood () )
      actionlog(30) << "Ball confidence good\n";
    else
      actionlog(30) << "Ball confidence bad\n";

    if( MyWorldModel.isBallKickable() )
      actionlog(30) << "Ball kickable\n";
    else
      actionlog(30) << "Ball not kickable\n";

    
    actionlog(30) << spades::ende;



  /* The ball has to be kickable at this moment */
  IsApplicable = MyWorldModel.ballConfidenceGood() &&
    MyWorldModel.isBallKickable();
}

/* Applicability function for ActDribble actions */
void ApplicabilityActionVisitor::visit(rcss::clang::ActDribble *)
{
  /* The agent should have an above average stamina */
  
  
  /* Tiredness of the agent */
  TiredNessT t = MyWorldModel.getAgentTiredNess();


  /* Check if all of the prerequisites are fulfilled */
  IsApplicable = MyWorldModel.ballConfidenceGood() &&
    (t == TIREDNESS_GOOD || t == TIREDNESS_AVERAGE);
  

  /* Also, define the threshold distance as a macro */
#ifndef NEAR_BALL
#define NEAR_BALL   5.0
#endif

  
  /* The agent position */
  VecPosition agentPos = MyWorldModel.getAgentGlobalPosition();
  
  /* Retrieve the current ball position */
  VecPosition ballPos = MyWorldModel.getBallPos();
  
  /* Calculate the distance from the agent to the bal */
  double dist = agentPos.getDistanceTo(ballPos);
  
  /* Calculate applicability */
  IsApplicable = IsApplicable && ( dist < NEAR_BALL );

}

/* Applicability function for ActClear actions */
void ApplicabilityActionVisitor::visit(rcss::clang::ActClear *)
{
  /* Applicable anytime */
  IsApplicable = MyWorldModel.ballConfidenceGood() &&
    MyWorldModel.isBallKickable();
}

/* Applicability function for ActShoot actions */
void ApplicabilityActionVisitor::visit(rcss::clang::ActShoot *)
{
  /* The agent has to be near the opponent goal */
  
  /* Also, define the threshold distance as a macro */
#ifndef NEAR_OPPGOAL
#define NEAR_OPPGOAL   25.0
#endif
  
  /* Retrieve the opponent goal position */
  VecPosition oppgoalPos = MyWorldModel.getPosOpponentGoal();
  
  /* Retrieve the agent position */
  VecPosition agentPos = MyWorldModel.getAgentGlobalPosition();
  
  
  /* Calculate the distance from the agent to the opponent goal
     position */
  double dist = agentPos.getDistanceTo(oppgoalPos);
  
  
  /* Check if the distance is within the threshold */
  IsApplicable = MyWorldModel.ballConfidenceGood() &&
    (dist < NEAR_OPPGOAL) && MyWorldModel.isBallKickable();
}

/* Applicability function for ActHold actions */
void ApplicabilityActionVisitor::visit(rcss::clang::ActHold *)
{
  IsApplicable = MyWorldModel.ballConfidenceGood() && 
    MyWorldModel.isBallKickable();
}

/* Applicability function for ActIntercept actions */
void ApplicabilityActionVisitor::visit(rcss::clang::ActIntercept *)
{
  IsApplicable = true;
}

/* Applicability function for ActTackle actions
   Important assumption: The UNumSet in ActTackle contains only one UNum
 */
void ApplicabilityActionVisitor::visit(rcss::clang::ActTackle *)
{
  /* Applicable anytime */
  IsApplicable = true;
}
