/* Dongryeol Lee
 *
 * Source file: CoachablePlayer.C
 * Description: This file is an implementation of CoachablePlayer
 *              defined in CoachablePlayer.h
 */


#include "CoachablePlayer.h"
#include "RegionVisitor.h"
#include "WorldModelAccessor.h"
#include "Spar.h"
#include "SoccerTypes.h"
#include "config.h"
#include "Logger.h"
#include "LevelStorage.h"


using namespace rcss::clang;

typedef std::vector< UNum > container;
typedef container::const_iterator const_iterator;


/* Our reference to the world model */
extern WorldModelAccessor MyWorldModel;
// pfr: 2/17/2004: part of a hack below
extern LevelStorage levels;



typedef list <double>::iterator angle_iterator;

#ifndef PI
#define PI 3.1415926
#endif

/*
 * Simple helper function which sends message to gui. 
 * Saves me from having to type MyWorldModel.getCurrentTime for each guiSend
 */
void CoachablePlayer::guiSend(char *msg, char *fgcolor, char *bgcolor){
  char actionSrcStr[5];
  if(actionsource == ActionBinding::COACH)
    snprintf(actionSrcStr, 5, "C");
  else if(actionsource == ActionBinding::RULEFILE)
    snprintf(actionSrcStr, 5, "R");
  else if(actionsource == ActionBinding::DEFAULT)
    snprintf(actionSrcStr, 5, "D");
  else if(actionsource == ActionBinding::ANY)
    snprintf(actionSrcStr, 5, "A");
  else if(actionsource == ActionBinding::COMMUNICATION)
    snprintf(actionSrcStr, 5, "S"); // S for Say
  else{
    cerr << "[CoachablePlayer::guiSend]: unknown action sourcetype"<<endl;
    return;
  }

  gui->send(MyWorldModel.getCurrentTime(), actionSrcStr, msg, fgcolor, bgcolor);
}


/*
 * This is an initialization function which is called from the Player 
 * class constructor
 * bool useGUI determines whether we send commands to the actions GUI
 */
void CoachablePlayer::InitCoachablePlayer(bool useGUI){ 
  
  // give the player a default home region:
  homeRegion = new RegRec(std::auto_ptr<Point>(new PointSimple(-40., -20.)), 
			  std::auto_ptr<Point>(new PointSimple(-5., 20.)));
  OffsideTrap = 0;
  LastTime = -1;
  lastActionType = ActionVisitor::TACKLE;
  
  // true passed to ActionsGUIInterface specifies to log commands
  gui = new ActionsGUIInterface(true, useGUI);

  levels.init();
}


/* This method tells whether the agent is in its home region */
bool CoachablePlayer::AmIHome()
{
  return homeRegion->isPtIn
    (MyWorldModel.getAgentGlobalPosition(),
     &MyWorldModel, myBindings);
}


/* This skill enables the agent to scan the field by rotating its body by
   its current view angle; same method as searchball, but will turn in the
   same direction */
SoccerCommand CoachablePlayer::scanField()
{
  guiSend("scanField", "yellow4");

  // calculate maximum turn angle
  AngDeg ang = WM->getAgentViewAngleValue( WM->getAgentViewAngle() );
  ang        = WM->getAngleForTurn( ang - ang/10.0, WM->getAgentSpeed() );

  if(BasicPlayer::isTimeToTurn()){
    return SoccerCommand( CMD_TURN, ang );
  }
  else{
    return SoccerCommand( CMD_TURN, 1 );
  }

}


/* This skill enables the agent to search for the player's location if it 
   cannot be found */
SoccerCommand CoachablePlayer::searchForPlayer(const UNum::unum_t& player, 
					       bool ourSide)
{
  static int timeLastSearch;
  int  iSign       = 1;

  char str[20];
  snprintf(str, 20, "searchForPlayer(%d)", player);
  guiSend(str, "yellow");

  // calculate maximum turn angle
  AngDeg ang = WM->getAgentViewAngleValue( MyWorldModel.getAgentViewAngle() );
  ang        = MyWorldModel.getAngleForTurn( ang - ang/10.0, 
					     MyWorldModel.getAgentSpeed() );
  
  
  // calculate whether expected player position is to the left or right of body
  // do this by making angle in directon body and back of body and check
  // whether estimated ball prediction lies between these two angles
  // only change sign when not searched in previous cycles, otherwise could
  // turn left, right, left, right and never find the player
  AngDeg angBody = MyWorldModel.getAgentGlobalBodyAngle();
  AngDeg angMax  = VecPosition::normalizeAngle( angBody + 180 );
  AngDeg angBall = (MyWorldModel.getGlobalPosition(player, ourSide) -
		    MyWorldModel.getAgentGlobalPosition()).getDirection();
  

  if( MyWorldModel.getCurrentTime() - timeLastSearch  > 3 )
    iSign = isAngInInterval( angBall, angBody, angMax ) ? 1: -1;
  timeLastSearch = MyWorldModel.getCurrentTime();


  if(BasicPlayer::isTimeToTurn()){
    return SoccerCommand( CMD_TURN, iSign * ang );
  }
  else{
    return SoccerCommand( CMD_TURN, 1 );
  }


}


/* Precondition of the function: Assumes that the UNumSet does not
   contain the invalid value, uMax in it */
UNum::unum_t CoachablePlayer::selectRandomUNum(const UNumSet &uset)
{
  UNum playerUNum;   /* UNum of the player selected */
  UNum::unum_t retUNum;   /* UNum value of the player */

  /* If the UNum set contains at least one player, */
  if(uset.size() > 0) {

    /* Now, randomly select player. Repeat selection if the selection
       is not valid or it is a wild card selection */
    /* generate random number between 1 and 11. */
    int r = (int) (rand() % uset.size());
    
    /* Retrieve the UNum of the player */
    retUNum = uset.getIndexAt(r).getUNum();
    
    /* If the player selected is all of them, arbitrarily select the first
       player */
    if(retUNum == UNum::uAll)
      return UNum::u1;
    else
      return retUNum;
  }
  
  /* If it doesn't, arbitrarily return the first player's UNum */
  else 
    return UNum::u1;
}

/* Decide which pass mode should be used */
SoccerCommand CoachablePlayer::selectPassMode(const UNum::unum_t& playerUNum)
{
  /* To-Do: Implementing the decision process for deciding pass modes

     1) Direct pass: normal and fast mode
     2) Leading pass
     3) Thorough pass
  */
  
  /* Retrieve the location of the passee */
  VecPosition pos = MyWorldModel.getGlobalPosition(playerUNum, true);
  
  /* Return pass mode */
  
  /* We will only do direct pass for now, but for deciding speed,... */

  /* Calculate the distance to the passee */
  double dist = VecPosition(MyWorldModel.getBallPos()).getDistanceTo( pos );

  // set the pass to player and its coord in the world model so that CoachableAgentComm
  // can communicate this information if needed.
  WM->setPassingToPlayer(MyWorldModel.getObjectTFromUNum(playerUNum, true), pos);
  
  /* If the distance is less than 10.0 m, we probably want normal pass */
  if( dist < 10.0 )
    return directPass(pos, PASS_NORMAL);
  else
    return directPass(pos, PASS_FAST);
}


/* This method implements passing to a chosen player given the 
   UNumSet. */
SoccerCommand CoachablePlayer::passToUNumSet(const UNumSet &uset)
{
  /* Important Note: The Action Picker will pass two kinds of pass actions
     
     1) Pass action to a teammate whose location is known with high confidence
     2) Pass action to a teammate whose location is known with low confidence
     
     The basic strategy is, to let the Coachable agent pass to the teammate
     if it knows where the teammate is. Otherwise, search for the player.

     Important Preconditions: 
     1) We always get a UNumSet uset containing only one UNum, because all 
     UNumSet actions are atomized
     2) Once this method, passToUNumSet is called, it is guaranteed that
     the agent will attempt to either search for the passee/pass the ball
  */
  

  /* First, check if the ball is kickable, if not, intercept */
  if(!MyWorldModel.isBallKickable())
    return interceptBall();

  //cout << WM->getCurrentTime() << " " << WM->getPlayerNumber() << ": passing to " << uset << endl;

  /* Retrieve the UNum of the passee */
  UNum::unum_t passeeUNum = uset.getIndexAt(0).getUNum();

  /* First, check the location of the passee and see if it's known with
     high confidence. If not, search for that player */
  if( !MyWorldModel.playerConfidenceGood(passeeUNum, true )) {

    ACT->putCommandInQueue( searchForPlayer(passeeUNum, true) );
    return alignNeckWithBody( ); // search for it

  }

  /* If it is in high confidence, go ahead with passing */
  else {
    char str[15];
    snprintf(str, 15, "pass %d", passeeUNum);
    guiSend(str, "black", "SkyBlue");
    return selectPassMode(passeeUNum);
  }
}


/* Pass to players within the region
   This module is obsolete due to reduction of pass reg -> pass UNum
*/
SoccerCommand CoachablePlayer::passToRegion(Region *)
{
  return SoccerCommand(CMD_ILLEGAL);
}


/*
 * Return a number indicating the danger (in terms of proximity to opponents)
 * of the line going from src to dest. 
 * Opponents closer to the src are more dangerous than ones closer to dest. 
 * Return value is higher for more dangerous lines and lower for safer lines. 
 */
double CoachablePlayer::evaluateLineDanger(VecPosition src, VecPosition dest){
  double danger = 0.0;
  float l1, l2;
  
  // construct a line between the src and dest
  Line srcLine = Line::makeLineFromTwoPoints(src, dest);


  for(ObjectT opp = OBJECT_OPPONENT_1; opp <= OBJECT_OPPONENT_11; opp = (ObjectT)(opp+1)) {
    // For all opponents whose locations are known....
    if( !WM->isConfidenceGood(opp)) {
      continue;
    }

    VecPosition oppPos = WM->getGlobalPosition(opp);

    VecPosition intersectOnLine = srcLine.getPointOnLineClosestTo( oppPos );

    // special case: if intersection is (0,0) that means there is no intersection, so 
    // lines must be parallel - ie, oppPos is on line srcLine. 
    if (intersectOnLine.getX() == 0. && intersectOnLine.getY() == 0.){
      intersectOnLine = oppPos;
    }


    // get the length of each segment of the src-dest line that intersectOnLine bisects
    l1 = (intersectOnLine - src).getMagnitude();
    l2 = (dest - intersectOnLine).getMagnitude();

#define CLOSENESS_FACTOR 5.0
    // include a marginal distance past the dest
    l2 += CLOSENESS_FACTOR;


    // if l1 is negative, then the intersect is before the src, which means 
    // the opponent is behind us and we can ignore him because he doesn't effect 
    // danger of our path. 
    if(l1 <= 0.0){
      PR(88, "[evaluateLineDanger]: skipping opp %d because he's behind us", opp);
      continue;
    }


    // if intersectOnLine is more than CLOSENESS_FACTOR past dest, it will be negative
    // and we consider this opponent out of range (doesn't contribute to danger)
    //if( l2 <= 0.0 ){
    //  PR(88, "[evaluateLineDanger]: skipping opp %d because he's past dest", opp);
    //  continue;
    //}

    // also need the distance from the opponent to the src-dest line
    float y = fabs( oppPos.getDistanceTo(intersectOnLine) );
    
    float dist = l1 + y;    //(l1 / (l1+l2))*y;
    // the factor before y gets greater as the intersection point gets further from src
    // this causes y to contribute to increasing the distance more, which in turn means
    // we have a lower danger; this is appropriate because the y distance is less significant
    // the further we are from the src and therefore should cause lower increases in danger
    assert(dist != 0.0);


    // if the distance is greater than the length of src-dest line + CLOSENESS_FACTOR,
    // we consider the opponent out of range (doesn't contribute to danger)
    if( dist > l1+l2 ){
      PR(88, "[evaluateLineDanger]: skipping opp %d because he's %d away (range=%d)", 
	 opp, dist, l1+l2);
      continue;
    }

    
    danger += 1./dist; //(inverted because danger increases as dist decreases)

    
  }


  //  cout <<WM->getCurrentTime()<<" dest "<<dest<<" has danger of "<<danger<<endl;
  return danger;
}


/* This method commands the player to dribble the ball to a given region
   This implements the "Smart Dribbling" method */
SoccerCommand CoachablePlayer::dribbleToRegion(Region *reg)
{
  /* First determine if the agent has the ball or not, else order him
     to intercept first */
  if( !MyWorldModel.isBallKickable() ) 
    return interceptBall();

  /* if the ball is moving fast, hold it first instead of dribbling */
  if (MyWorldModel.getBallSpeed() > .5){
    return hold();
  }


  VecPosition ballPos = MyWorldModel.getBallPos();
  VecPosition agentpos  = MyWorldModel.getAgentGlobalPosition();

  /*
    Use RegionIterator to get points X distance apart in the region 
    and then evaluate each point as a dribble to target. 
    Pick one that has best chance of reaching. (evaluate worse for lines with players
    closer to us than players farther away)
  */

  VecPosition bestPt(0, 0);
  double bestPtEval = 99999; // (we're trying to find a min)

  //cout << "Region is: "<<*reg<<", using 4.0 spacing"<<endl;
  RegionIterator *ri = reg->newIterator(4.0, &MyWorldModel, myBindings);

  /* Iterate all the points in the region, and get the best one */
  while( ri->valid() ) {
    
    VecPosition currPt = ri->getPt(&MyWorldModel, myBindings);
    double temp = evaluateLineDanger(ballPos, currPt);
    if(temp < bestPtEval) {
      bestPt = currPt;
      bestPtEval = temp;
    }
    // if eval is equal, pick the one that is closer
    else if (temp == bestPtEval){ 
      if( (currPt-ballPos).getMagnitude() < (bestPt-ballPos).getMagnitude()){
	bestPt = currPt;
	bestPtEval = temp;
      }
    }

    /* Go to the next point */
    ri->moveToNextPt();
  }

  delete ri;

  // if we're already within 5 distance of our ideal point, just hold ball
  if(bestPt.getDistanceTo(agentpos) <= 5.0){

    // also require that we actually be in the region.
    if( reg->isPtIn(agentpos, &MyWorldModel, myBindings) ){
      guiSend("dribble:hold", "purple");
      return hold();
    }

    // *Or*, if the region is small (especially RegPt's) 
    // we don't require the agent to be inside the region because it would be too hard
    // to position exactly inside a Pt or small rectangle
    RegionVisitor rvisitor;
    reg->accept(rvisitor);
    if( rvisitor.getRegionArea() <= 4.0 ){
      guiSend("dribble:hold", "purple");
      return hold();
    }
  }

  //cout << WM->getCurrentTime()<<" dribbling to point "<<bestPt<<endl;
  guiSend("dribble", "purple");

  // now we should dribble to our bestPt
  return dribble( (bestPt - ballPos).getDirection(), DRIBBLE_FAST);


#if 0
  /* If no opponent is found (region is clear to dribble to), */
  if( oppCount == 0 )

    /* Dribble fast toward the opponent goal in this cycle! */
    return dribble((MyWorldModel.getPosOpponentGoal() - ballPos).
		   getDirection(), DRIBBLE_FAST);
  
  /* Otherwise, if the opponent count < 4, we should probably
     dribble fast */
  else if( oppCount < 4)

    return dribble(bestangle, DRIBBLE_FAST );

  /* If the opponent count is < 7, we should probably dribble slow */
  else if( oppCount < 7 )
    
    return dribble(bestangle, DRIBBLE_SLOW );

  /* Otherwise, we dribble with the ball */
  else
    return dribble(bestangle, DRIBBLE_WITHBALL);
#endif

}

/* This method commands the player to shoot the ball at the opponent's
   goal */
SoccerCommand CoachablePlayer::shoot()
{
  /* First determine if the agent has the ball or not, else order him
     to intercept first */
  if(!MyWorldModel.isBallKickable()) 
    return interceptBall();
  
  VecPosition *finalShootLoc = NULL;
  VecPosition ballLoc = WM->getBallPos();
  
  
  /* Here does not consider self distance to goal. So no matter how 
     far away from the opp goal, just select a route to shoot at */
     
  unsigned int numShootRoute = 64;
  /* discritize the goal width to generate num_shoot_route + 1 shoot point */
  float quantizeStep = SS->getGoalWidth() / numShootRoute;
  
  float probMax = -1.0;
  
  /* check each shooting line */
  for(int i = 0; i <= numShootRoute; i ++) {
  	
    /* shoot point */
    VecPosition shootLoc = VecPosition(53, -SS->getGoalWidth()/2 + quantizeStep * i);
    /* shooting line angle */
    double shootAngle = (shootLoc - ballLoc).getDirection();
    
    /* from ball to the shoot point */
    VecPosition ball2goal = shootLoc - ballLoc;
    /* from ball to the left goalpost */
    VecPosition ball2l = VecPosition(53, -SS->getGoalWidth()/2) - ballLoc;
    /* from ball to the right goalpost */
    VecPosition ball2r = VecPosition(53, SS->getGoalWidth()/2) - ballLoc;
    
    /* Shooting evaluation is based on the following paper:
     * Jelle R. Kok, Remco de Boer, Nikos Vlassis, and F.C.A. Groen. 
     * Towards an optimal scoring policy for simulated soccer agents. 
     * In G. Kaminka, P.U. Lima, and R. Rojas, editors, RoboCup 2002: 
     * Robot Soccer World Cup VI, pages 296-303, Fukuoka, Japan, 2002. 
     *
     * The idea is to decompose the scoring policy into two independent subproblems:
     * - Determining the probability that the ball will enter the goal 
     *   when shot to a specific point in the goal from a given position.
     * - Determining the probability of passing the goalkeeper in a given situation.
     */
     
    /* here we only implement their basic idea:
     * i.e. select the one with the maximum product of the above two probabilities
     * . 
     * TODO: implement the complete evaluation method based on 
     * ThuAeolus' goingthrough-speed-calculation
     */

    
    /* angle between shoot line and the left shoot limit */
    double angleL = fabs(VecPosition::normalizeAngle(ball2l.getDirection()
                         - ball2goal.getDirection()));
                         
    /* angle between shoot line and the right shoot limit */                         
    double angleR = fabs(VecPosition::normalizeAngle(ball2r.getDirection()
                         - ball2goal.getDirection()));
         
    /* closest distance from left/right goalpost to the shooting line */
    float distL = ball2l.getMagnitude() * sin(angleL * PI / 180);
    float distR = ball2r.getMagnitude() * sin(angleR * PI / 180);      
       
    /* excerpt from uva's paper, a standard deviation of the y-coordinate 
       when the ball enters the goal. */         
    float sigmaL = -1.88 * log(1 - ball2l.getMagnitude() / 45);
    float sigmaR = -1.88 * log(1 - ball2r.getMagnitude() / 45);
    
    /* gaussian density function */
    float gL = 0.5 * erf(distL / (sigmaL * sqrt(2.0)));
    float gR = 0.5 * erf(distR / (sigmaR * sqrt(2.0)));
    
    /* the probability that the ball will enter the goal 
     * when shot to a specific point in the goal from a given position. 
     */
    float probSuccGoal = gL + gR;
   
    /* obtain goalie's position */
    /* TODO: when the confidence of goalie's position is low, what to do */
    ObjectT     objGoalie = WM->getOppGoalieType();
    VecPosition posGoalie = WM->getGlobalPosition( objGoalie );
    
    /* the distance between the ball and the goalkeeper */
    float d = (posGoalie - ballLoc).getMagnitude();
    /* the absolute angle between the goalkeeper and the shooting point */
    float a = fabs(VecPosition::normalizeAngle(shootAngle - (posGoalie - ballLoc).getDirection()));
    
    /* resulting discriminant function */
    float u = 0.043 * a + 0.09 * d - 2.1323; 
    
    /* the probability of passing the goalkeeper in a given situation */
    float probPassGoalie = 1 / (1 + exp(-9.5 * u));
    
    /* the assumption of independence gives the total probability of 
     * scoring in a given situation as the product of these two values */
    float prob = probPassGoalie * probSuccGoal;
    
    if(prob > probMax) {
      finalShootLoc = &shootLoc;
      probMax = prob;
    }     
  }
  
  if(finalShootLoc != NULL /* && probMax > 0.8 */){
    guiSend("shoot", "black", "red");
    return kickTo(*finalShootLoc, SS->getBallSpeedMax());
  }
  else
    return hold();  
}

/* This method commands the player to intercept the ball and if the ball is kickable, to tackle it */
SoccerCommand CoachablePlayer::intercept_tackle()
{ 
  VecPosition player_2_ball = WM->getBallPos() - WM->getAgentGlobalPosition();
  AngDeg angBallRelToPlayer = player_2_ball.getDirection();


  double tackle_dist;

  if(angBallRelToPlayer > 90 || angBallRelToPlayer < -90)
    tackle_dist = .5;
  else
    tackle_dist = 2.5;

  double fail_prob = pow((player_2_ball.getX() / tackle_dist), 6) + pow((abs(player_2_ball.getY()) / 1.25), 6);

  if(fail_prob < .5){
    guiSend("tackle", "white", "black");
    return tackleBall(100);
  }
  else
    return interceptBall();

}

/* This method commands the player to hold the ball in its place */
SoccerCommand CoachablePlayer::hold()
{ 
  if(!MyWorldModel.isBallKickable())
    return interceptBall();

  guiSend("hold", "black", "yellow3");

  if(MyWorldModel.getBallSpeed() < 0.05)
    return kickBallCloseToBody(30.0);
  else
    return freezeBall();
}

/* This method commands the player to intercept the ball */
SoccerCommand CoachablePlayer::interceptBall()
{
  /* Cheap hack: first determine whether the agent knows where the ball
     is. If not, search for the ball first */
  if(!MyWorldModel.ballConfidenceGood()) {
    //PR(8, "[interceptBall]: don't know where the ball is. searching\n");
    //cout <<"["<<MyWorldModel.getCurrentTime()<<"]: interceptBall: searching ball instead"<<endl;
    guiSend("searchBall", "orange");
    ACT->putCommandInQueue( searchBall() ); 
    return alignNeckWithBody( ); 
  }
  
  SoccerCommand soc(CMD_ILLEGAL);


  /*  if(MyWorldModel.isBallKickable()){
    cout << "[CoachablePlayer]: "<<WM->getCurrentTime()
	 <<" Warning: called interceptBall when ball is already kickable"<<endl;
	 }*/


  /* If the agent is the goalie
   * then he will do a special interception */
  if( MyWorldModel.isGoalie() ) {
  
    /* Intercept if the ball will be within penalty area in 10 cycles 
     * This prevents us from intercepting if the ball is leaving the penalty area 
     * ( eg, when we just kicked it away) */
    //if( MyWorldModel.FastestToBall() && WM->isInOwnPenaltyArea( WM->getBallPos() ) )
    // if( WM->isInOwnPenaltyArea( WM->getBallPos() ) && WM->isInOwnPenaltyArea(WM->predictPosAfterNrCycles( OBJECT_BALL, 45)){
    if( WM->isInOwnPenaltyArea( WM->getBallPos() ) && 
	WM->isInOwnPenaltyArea(WM->predictPosAfterNrCycles( OBJECT_BALL, 10) )){
	guiSend("intercept", "black", "green3");
	ACT->putCommandInQueue( soc = intercept( true ) );
	//cout << "["<<MyWorldModel.getCurrentTime()<<"]: goalie: intercepting"<<endl;
    }

    /* If ball is outside penalty area,
     * the goalie should goHome (position himself defensively)  */
    else{
      /* ACT->putCommandInQueue( soc = interceptScoringAttempt() );
      cout << "goalie: intercepting score attempt (blocking)"<<endl;
      */

      // NOTE: trilearn player does not use interceptScoringAttempt, and I can't 
      // get it to work well, so not using it for now. Re-attempt later? 
      //cout <<"["<<MyWorldModel.getCurrentTime()<<"]: goalie: going home, not intercepting"<<endl;
      ACT->putCommandInQueue( goHome() );      

    }
    /* jdavin: also should use interceptScoring if ball is far away (using 
     * outside the penalty area for now). we don't want to 
     * leave the goal too much in order to intercept */
  }

  /* If the agent is not the goalie, just intercept ball */
  else{
    guiSend("intercept", "black", "green3");
    ACT->putCommandInQueue( soc = intercept(false) );
  }

  return turnNeckToObject(OBJECT_BALL, soc);
}


/* This method commands the player to clear the ball from a given region.
   kick the ball toward to clear with the maximum possible speed */
SoccerCommand CoachablePlayer::clear(Region *reg)
{
  /* If the ball is not kickable, then intercept first */
  if(!MyWorldModel.isBallKickable())
    return interceptBall();

  vector <VecPosition> positions;     /* store positions of the opponent 
					 players */
  double maxEval = -1.0;                /* Evaluated value of the optimal
					 angle */
  int optimalAngle = -66;                /* Optimal angle to keep at */

  
  /* First retrieve the ball position */
  VecPosition ballPos = MyWorldModel.getBallPos();
  
  
  /* Now, we need to build a UNumSet of the opponents in the region to 
     find the optimal path to kick the ball through */
  for(int i = 1; i <= 11; i++) {
    
    /* For all opponents whose locations are known.... */
    if(MyWorldModel.playerConfidenceGood((UNum::unum_t) i, false)) {

      /* Retrieve the position of each opponent player */
      VecPosition oppPos = MyWorldModel.getGlobalPosition
	((UNum::unum_t) i, false);
    
      /* Calculate agent to opponent player angle */
      double angle = (oppPos - ballPos).getDirection();
        

      /* If the opponent player is in the region and at the right
	 angle interval -90 and 90, add to our list of
	 consideration */
      if(reg->isPtIn(oppPos, &MyWorldModel, myBindings) 
	 && angle > -90.0 && angle < 90.0)
	
	/* Push in the opponent player position */
	positions.push_back(oppPos);
    }
  }  
  
  /* We will use the clearing heuristic defined by the usefulness
     function: u(theta) = (sin(1.5 * theta + 45.0) + 1.0) / 2.0 */
  
  // jdavin: 
  // changed to u(theta) = (sin(1.5 * theta + 90.0) + 1.0) / 2.0
  // to center usefulness around theta=0. 

  /* Now, try every angle from -60.0 to 60.0 with the interval of 5.0 */
  for(int theta = -50; theta <= 50; theta += 5) {
    
    /* Total evaluation value = usefulness * sum of the success */
    double eval;

    /* First, convert to radian angle */
    //double angrad = (double) theta / PI;
    double angrad = (double) theta * PI / 180.;    

    /* Calculate the usefulness according to the usefulness function */
    double usefulness = (sin(1.5 * angrad + PI / 2.0) + 1.0) / 2.0;
    
    /* Our degree of success */
    double success = 1.0;

    /* Make the line given the agent position and the angle direction:
       this is the direction we want to clear to */
    Line l = Line::makeLineFromPositionAndAngle(ballPos, (double) theta);

    /* Now, go through our opponent position set, */
    for(unsigned int num = 0; num < positions.size(); num++)  {
     
      /* Calculate the degree of success for each opponent */

      /* Calculate the distance w between the clearing direction and
	 the opponent player position */
      double w = l.getDistanceWithPoint(positions[num]);
      
      /* Calculate the distance d between the player on the line and
	 the ball */
      double d = ballPos.getDistanceTo(l.getPointOnLineClosestTo
				       (positions[num]));

      /* We want to divide w by d. So check if d is 0. If so, check if w = 0 
	 as well. If so, the opponent player is at the ball, so we return 0 
	 as eval. Otherwise, the success rate remains the same */
      if(d == 0) {
	
	if(w == 0) {
	  success = 0.0;
	  break;
	}
	else
	  continue;
      }
      
      /* Otherwise, accumulate our result into success parameter */
      success *= (w / d);

    }

    /* also check if this optimalAngle would shoot the ball off the field within 20 meters */
    if( !WM->isInBoundsY( ballPos+VecPosition(20,theta,POLAR)) ){
      //cout <<"[clear]: angle "<<theta<<" would kick ball out of bounds, so reducing eval\n";
      usefulness = 0.;
    }

    /* Now, calculate our total evaluation value */
    eval = usefulness * success;

    /* Keep track of our optimal angle and the maximum eval value */
    if(eval > maxEval) {
      
      maxEval = eval;
      optimalAngle = theta;
    }

  }
  
  // jdavin: remove this later: 
  //if(optimalAngle == -66)
  //  cerr << MyWorldModel.getCurrentTime()<<": optimalAngle not selected!!!"<<endl;
  //assert(optimalAngle != -66);

  // compute the point that we should kick the ball to
  // We want the point to be at optimalAngle from the current ballPos and at the exit
  // point of the clear region
  VecPosition KickToPoint = ballPos+VecPosition(10,optimalAngle,POLAR); 
  // the default KickToPoint (above) is 10 units out from the ballPos.

  // get the intersection of a line from the ball at optimalAngle and the clear region
  Line angleLine = Line::makeLineFromTwoPoints( ballPos, 
						ballPos+VecPosition(200,optimalAngle,POLAR) );
  // the line extends 200 units from ballPos just to make sure it extends across the entire field

  vector<VecPosition> intersections = reg->getIntersection(angleLine, &MyWorldModel, myBindings);
  // since regions are convex, we can not have more than 2 line intersections.

  /*if(intersections.size() > 2){

    cout <<"Intersected angleLine: "<<angleLine<<" with region: "<<*reg<<" gave "<<
      intersections.size()<<" intersection"<<endl;

    assert(intersections.size() <= 2);
    }*/

  bool tmpISDefault = true; // remove this code
  if(intersections.size() == 1){
    // if the point is right of ball, use it
    if(intersections[0].isInFrontOf(ballPos)){
      KickToPoint = intersections[0];
      tmpISDefault = false;
    }
    // otherwise use the default KickToPoint. 
  }
  else if(intersections.size() == 2){
    VecPosition rightmost;
    // use the rightmost intersection point
    if(intersections[0].isInFrontOf(intersections[1]))
      rightmost = intersections[0];
    else // intersections[1] is rightmost
      rightmost = intersections[1];

    // make sure it's right of ball though
    if(rightmost.isInFrontOf(ballPos)){
      KickToPoint = rightmost;
      tmpISDefault = false;
    }
    // otherwise use default KickToPoint

  }
  else{ //(intersections.size() == 0)
    // the clear rule (from coach?) probably gave us a stupid clear region. 
    // ignore this case, just use default KickToPoint. 
     
  }


  char str[20];
  if(tmpISDefault)
    snprintf(str, 20, "clearDef(%d)", optimalAngle);
  else{
    snprintf(str, 20, "clearKickTo(%d)", optimalAngle);
    //cout <<MyWorldModel.getCurrentTime()<<":["<<MyWorldModel.getMyNum()<<"]: ballpos: "<<ballPos<<" clearing to: "<<KickToPoint<<" @"<<optimalAngle<<endl;

  }
  guiSend(str, "blue");
  /* Now, kick to the point such that the ball arrives at max speed (if possible)*/
  return kickTo(KickToPoint, SS->getBallSpeedMax());
}



/* This method commands the player to mark some opponent in UNUM_SET */
SoccerCommand CoachablePlayer::markUNumSet(const UNumSet &uset)
{
  SoccerCommand soc;
  /* Retrieve the UNum of the opponent */
  UNum::unum_t oppUNum = uset.getIndexAt(0).getUNum();
  
  
  //cout << "Player " << MyWorldModel.getMyNum() << " is marking opponent " << oppUNum << endl;

  /* First check, if the player confidence of the opponent to be marked is
     good. If not, we should search for the opponent */
  if( !MyWorldModel.playerConfidenceGood( oppUNum, false ) ) {
    
    ACT->putCommandInQueue( searchForPlayer( oppUNum, false ) );
    return alignNeckWithBody();
  }

  char str[15];
  snprintf(str, 15, "mark %d", oppUNum);
  guiSend(str, "blue", "red");

  /* Retrieve the location of the opponent to be marked */
  VecPosition oppPos = MyWorldModel.getGlobalPosition(oppUNum, false);

  /*  if(oppUNum == UNum::u1){
    cerr <<MyWorldModel.getCurrentTime()<<":["<<MyWorldModel.getMyNum().getUNum()<<"]: I'm marking "<<uset.getIndexAt(0)<<" who's at: "<<oppPos<<endl;
    }*/

  /* Now, we need to decide the marking mode depending on the location of
     the opponent */
    
  /* MARK BALL/MARK GOAL: The opponent can kick the ball */
  if( MyWorldModel.isBallKickableByPlayer( oppUNum, false ) ) {

    /* If the opponent is close to our own goal, defend our goal, MARK GOAL */
    if(oppPos.getDistanceTo(MyWorldModel.getPosOwnGoal()) < 30.0)
        ACT->putCommandInQueue( soc=mark(MyWorldModel.getObjectTFromUNum(oppUNum, false), 
				     2.0, MARK_GOAL));
    
    /* Otherwise, just do MARK BALL */
    else
      ACT->putCommandInQueue( soc=mark(MyWorldModel.getObjectTFromUNum(oppUNum, false), 2.0, 
		  MARK_BALL));
  }

  /* MARK BISECTOR: If the opponent can't kick the ball now, then do this */
  else
    ACT->putCommandInQueue( soc=mark(MyWorldModel.getObjectTFromUNum( oppUNum , false ), 2.0, 
		MARK_BISECTOR));

  // if we know where the ball is and we are close, try to look at ball
  if( MyWorldModel.ballConfidenceGood() ){
    return turnNeckToObject( OBJECT_BALL, soc );
  }
  else { /* ball confidence bad! */
    return scanWithNeck();  
  }
}

/* This method commands the player to mark the passing lane from the current 
   ball position to some opponent player in UNUM_SET. 
*/
SoccerCommand CoachablePlayer::markLaneUNumSet(const UNumSet &uset)
{
  SoccerCommand soc;
  /* Retrieve the UNum of the opponent */
  UNum::unum_t oppUNum = uset.getIndexAt(0).getUNum();


  /* First check, if the player confidence of the opponent to be marked is
     good. If not, we should search for the opponent */
  if( !MyWorldModel.playerConfidenceGood( oppUNum, false ) ) {
    
    ACT->putCommandInQueue( searchForPlayer( oppUNum, false ) );
    return alignNeckWithBody();
  }

 char str[15];
  snprintf(str, 15, "mark %d", oppUNum);
  guiSend(str, "blue", "red");
  /* Mark using MARK_BALL to block the passing lane */
  ACT->putCommandInQueue( soc = mark(MyWorldModel.getObjectTFromUNum( oppUNum, false), 2.0, 
	      MARK_BALL));  

  // if we know where the ball is and we are close, try to look at ball
  if( MyWorldModel.ballConfidenceGood() ){
    return turnNeckToObject( OBJECT_BALL, soc );
  }
  else { /* ball confidence bad! */
    return scanWithNeck();  
  }
}

/* This method commands the player to mark the passing lane from the current 
   ball position to a given region.
   This method has been declared obsolete since the reduction stage will
   reduce all markLaneRegion actions to markLaneUNumSet actions */
SoccerCommand CoachablePlayer::markLaneRegion(Region *)
{
  return SoccerCommand(CMD_ILLEGAL);
}
  
/* This method commands the player to tackle some opponent in UNUM_SET
   Due to reduction of actions, the UNUM_SET will contain only one UNUM */
SoccerCommand CoachablePlayer::tackle(const UNumSet &uset)
{
  /* Retrieve the UNum of the opponent */
  UNum::unum_t oppUNum = uset.getIndexAt(0).getUNum();


  VecPosition oppPos = WM->getGlobalPosition(SoccerTypes::getOpponentObjectFromIndex( oppUNum - 1 ));

  
  /* If the opponent is close to the ball, tackle */
  if(oppPos.getDistanceTo(MyWorldModel.getBallPos()) > 5.0)
    return interceptBall();
  else
    return intercept_tackle();

}



/* this is a fractional (0 to 1) number that defines the goalie position
 * along a line in penalty area. 
 * The x dimension of line is GOALIE_LINE*PENALTY_AREA_LENGTH  */
#define GOALIE_LINE 0.5  // trilearn: 0.7


/* This is from the trilearn goalie positioning code 
 * Returns a position on the line between the ball and our goal 
 * GOALIE_LINE determines how far across the goalie area will be selected for 
 * intersection with the ball->goal line. 
 */
VecPosition getGoalieBlockPosition(){
  VecPosition posIntersect;

  // define the top and bottom position of a rectangle in which keeper moves
  static const VecPosition posLeftTop( -PITCH_LENGTH/2.0 +
				       GOALIE_LINE*PENALTY_AREA_LENGTH, -PENALTY_AREA_WIDTH/4.0 );
  static const VecPosition posRightTop( -PITCH_LENGTH/2.0 +
					GOALIE_LINE*PENALTY_AREA_LENGTH, +PENALTY_AREA_WIDTH/4.0 );
  
  // define the borders of this rectangle using the two points.
  static Line  lineFront = Line::makeLineFromTwoPoints(posLeftTop, posRightTop);
  static Line  lineLeft  = Line::makeLineFromTwoPoints(VecPosition( -50.0, 
								    posLeftTop.getY()), posLeftTop );
  static Line  lineRight = Line::makeLineFromTwoPoints(VecPosition( -50.0, 
								    posRightTop.getY()),posRightTop );

  // make line between own goal and the ball
  VecPosition posMyGoal = ( MyWorldModel.TeamIsLeftSide() )
    ? SoccerTypes::getGlobalPositionFlag(OBJECT_GOAL_L, SIDE_LEFT )
    : SoccerTypes::getGlobalPositionFlag(OBJECT_GOAL_R, SIDE_RIGHT);
  Line lineBall = Line::makeLineFromTwoPoints( MyWorldModel.getBallPos(), posMyGoal);

  // determine where your front line intersects with the line from ball
  posIntersect = lineFront.getIntersection( lineBall );

  // outside rectangle, use line at side to get intersection
  if (posIntersect.isRightOf( posRightTop ) )
    posIntersect = lineRight.getIntersection( lineBall );
  else if (posIntersect.isLeftOf( posLeftTop )  )
    posIntersect = lineLeft.getIntersection( lineBall );

  if( posIntersect.getX() < -49.0 )
    posIntersect.setX( -49.0 );

  //  assert(posIntersect.getX() != 0 );
  if(posIntersect.getX() == 0 ) cerr << "goalie pos not selected!"<<endl;

  return posIntersect;
}


/* 
 * Return distance that current agent can travel in numcycles. 
 *
 */
double CoachablePlayer::DistanceForCycles(int numcycles){

  VecPosition oldpos = MyWorldModel.getAgentGlobalPosition();
  VecPosition newpos = WM->predictAgentPos( numcycles, SS->getMaxPower());
  // using the max dash power 

  /* cout << oldpos << newpos << " max dashpower: "<<SS->getMaxPower()<<endl; */
  return oldpos.getDistanceTo(newpos);
}

#define K 2 // constant used to add to number of cycles player is allowed to reach ball



/* 
 * Return the position that we want the goalie to move to. (Used by posRegion)
 * Determine optimal goalie position that will maximize likelihood of being 
 * able to intercept or catch ball. 
 * This is the point that equally blocks ball trajectory to each of the goal posts
 * We calculate the time for the ball to reach each post (ltime, rtime), 
 * and then draw circles around each post (radius = ltime or rtime) and the optimal
 * point is the rightmost intersection of the circles. 
 * Based on technique used by Cyberoos team, according to Tom Howard
 */
VecPosition CoachablePlayer::getGoalieTargetPos(){
  int ltime, rtime;
  // distance radius for left and right circles (left=top, right=bottom):
  double Lradius, Rradius; 
  ObjectT topflag, botflag;
  Circle topcircle, botcircle;
  VecPosition leftintersect, rightintersect, intersect;


  // ltime is the time for the ball to get to the left goalpost
  // and rtime is for right goal post
  if(MyWorldModel.TeamIsLeftSide()){ // we're on left so use left goal posts (our goal)
    topflag = OBJECT_FLAG_G_L_T; // right top goalpost
    botflag = OBJECT_FLAG_G_L_B; // right bottom goalpost
  }
  else{ // use left goal posts 
    topflag = OBJECT_FLAG_G_R_T; // left top goalpost
    botflag = OBJECT_FLAG_G_R_B; // left bottom goalpost
  }

  double distleft = WM->getGlobalPosition(topflag).getDistanceTo(WM->getBallPos());
  double distright = WM->getGlobalPosition(botflag).getDistanceTo(WM->getBallPos());

  // predict how many cycles it would take the ball to travel distleft,distright
  // at max speed. (I wrote this function in WorldModelPredict -jdavin)
  ltime = WM->predictCyclesForBallTravel(distleft);
  rtime = WM->predictCyclesForBallTravel(distright);

  //cout << "Top dist "<<distleft<< " Bottom dist "<<distright<<endl;
  //cout <<"Ball to left post in "<<ltime<<" and to right post in "<<rtime<<endl;

  //Lradius = DistanceForCycles( (int)(ltime + 1 ) );
  //Rradius = DistanceForCycles( (int)(rtime + 1 ) ); 

  Lradius = distleft  -6.; // * .85;
  Rradius = distright -6.;  //*.85;

  //cout <<"Player can travel "<<Lradius<<" from left and "<<Rradius<<" from right."<<endl;

  topcircle = Circle(WM->getGlobalPosition(topflag), Lradius);
  botcircle = Circle(WM->getGlobalPosition(botflag), Rradius);

  int num = topcircle.getIntersectionPoints ( botcircle, &leftintersect, &rightintersect );
  //cout << num << " intersections: right= "<<rightintersect<<" left="<<leftintersect<<endl;


  /* if num of intersections is 2 and time for ball to reach each post is <50, 
   * then use the circle intersection. */
  if(num == 2 && (distleft < 25. || distright < 25.)) {
    if(MyWorldModel.TeamIsLeftSide())
      intersect = rightintersect;
    else
      intersect = leftintersect;
    //cout << "Intersect at "<<intersect<<endl;

    // clip X coordinate if necessary (we don't want to leave penalty box)
    if(intersect.getX() > -PITCH_LENGTH/2.0 + PENALTY_AREA_LENGTH)
      intersect.setX(-PITCH_LENGTH/2.0 + PENALTY_AREA_LENGTH);

    // the y coord should never be outside the penalty area, but check anyway
    if(intersect.getY() < -PENALTY_AREA_WIDTH/2.0)
      intersect.setY(-PENALTY_AREA_WIDTH/2.0);
    if(intersect.getY() > PENALTY_AREA_WIDTH/2.0)
      intersect.setY(PENALTY_AREA_WIDTH/2.0);

    return intersect;
  }
  else // the circles don't intersect (or are too far out) 
    // so just position along line from goal to ball
    return getGoalieBlockPosition();

}





/* This method will command the player to position itself in the given
   region */
SoccerCommand CoachablePlayer::posRegion(Region *reg)
{
  /* Position to go */
  VecPosition p(0.0, 0.0);

  /* Get the current play mode */
  PlayMode pm = MyWorldModel.getPlayMode();


  /* if we're the goalie, determine the position point specially */
  /* note that this ignores the desired position and will therefore also
   * override the home position */
  if(MyWorldModel.isGoalie()){

    p = getGoalieTargetPos(); 

    // the actual moveTo and turnNeck commands will be sent later in this function

  }

  /* If the play mode is bko, or after either team scored a goal, */
  else if( pm == AfterGoal_Opp || pm == AfterGoal_Our ||
      pm == BeforeKickOff || pm == KickOff_Our || pm == KickOff_Opp ) {
    
    /* Get random pt */
    p = reg->getRandomPtIn(&MyWorldModel, myBindings);


    /* Shift the position onto our side if necessary */
    if(p.getX() >= 0.0)
      p.setX(-1.0);
  }
  else { /* Determine desired position using Spar */
    
    /* Retrieve the current position of the agent and pass it to the best point
       selection in the RegionIterator */
    RegionIterator *ri = reg->newIterator(4.0, &MyWorldModel, 
					  myBindings);
    
    /* Get the best point in the iterator */
    p = Spar::getBestPt(ri, MyWorldModel.getMyNum().getUNum(), 
			myBindings);

    delete ri;

    /* actionlog(100) << "offsideX is: " << MyWorldModel.getOffsideX() << spades::ende; */
    if (MyWorldModel.getOffsideX() > 100)
      {
	errorlog << "OffsideX is weird! " << MyWorldModel.getOffsideX()
		 << ": " << trilearn_wm->isConfidenceGood(OBJECT_BALL)
		 << " " << trilearn_wm->getBallPos() << "  ";
	for (int i=1; i<11; i++)
	  errorlog << i << "=" 
		   << MyWorldModel.playerConfidenceGood((rcss::clang::UNum::unum_t)i, false)
		   << ' ' << MyWorldModel.getGlobalPosition((rcss::clang::UNum::unum_t)i, false) << "  ";
	errorlog << spades::ende;
      }

    /* Adjust so that we are onsides */
    if(p.getX() >= MyWorldModel.getOffsideX() - 0.5)
      p.setX( MyWorldModel.getOffsideX() - 0.5 );
  }


  /* If the play mode is before kick off and it's kickoff for the opponent
     team, no players can't be inside the center circle */
  if( WM->isKickOffThem( ) && p.getDistanceTo( VecPosition(0,0) ) < 9.0 ){
    p = p.scaleTo(9.5);
  }
  
  /* We need to adjust the position so that the player is on the field */
  
  /* First check if the x-coordinate is on the field */
  if( p.getX() > PITCH_LENGTH / 2.0 )
    p.setX( PITCH_LENGTH / 2.0 );
  else if( p.getX() < -PITCH_LENGTH / 2.0 )
    p.setX( -PITCH_LENGTH / 2.0 );

  
  /* Then, check if the y-coordinate is on the field */
  if( p.getY() > PITCH_WIDTH / 2.0 )
    p.setY( PITCH_WIDTH / 2.0 );
  else if( p.getY() < -PITCH_WIDTH / 2.0 )
    p.setY( -PITCH_WIDTH / 2.0 );

 
  /* If we are near where we want to be (within 1m), */ 
  if(MyWorldModel.getAgentGlobalPosition().getDistanceTo(p) < 1.0) {

    /* If we don't have a good confidence in ball position, we should search */
    if( !MyWorldModel.ballConfidenceGood() ){
      guiSend("searchBall", "orange");
      ACT->putCommandInQueue( searchBall() );
    }
      
    /* If the playmode is not play-on, we should scan field */
    else if( pm != PlayOn )
      {
	//float dist_to_ball = WM->getBallPos().getDistanceTo(WM->getAgentGlobalPosition());
	/* if (dist_to_ball > 5 && dist_to_ball < 15) */
	//if (dist_to_ball < 15)
	//  ACT->putCommandInQueue( turnBodyToObject(OBJECT_BALL) );
	//else
	/* jdavin: commented out above because we need the kickoff player to still scan
	 * the field so that he knows where his teammates are so he can pass to them!
	 */
	ACT->putCommandInQueue( scanField() );
      }
    
    /* Otherwise, watch the ball */
    else{
      guiSend("posRegion turnBodyToObj");
      ACT->putCommandInQueue( turnBodyToObject(OBJECT_BALL) );
    }


    return alignNeckWithBody();
  }
  else { /* we're not in desired region, so need to move to it */

    /* Before KickOff, we have to teleport */
    if( pm == BeforeKickOff || pm == AfterGoal_Opp || pm == AfterGoal_Our )
      return teleportToPos(p);
    
    else {      
      SoccerCommand soc(CMD_ILLEGAL);
      
      /* Send moveToPos command */

      /* Since we can only turn neck from 90 to -90, if the ball is upfield of us,
       * we will lose sight of it by turning and moving in -x direction. 
       * So move backwards (-effort) instead. 
       */
      if(  MyWorldModel.getPlayMode() == PlayOn && MyWorldModel.ballConfidenceGood() && 
	   MyWorldModel.getAgentGlobalPosition().getX() < MyWorldModel.getBallPos().getX() 
	  && p.getX() < MyWorldModel.getAgentGlobalPosition().getX()){
	// ball is to right, so move backwards 
	ACT->putCommandInQueue(soc=moveToPos(p, PS->getPlayerWhenToTurnAngle(), 30., false));

	guiSend("moveTo");
      }
      else if( !MyWorldModel.ballConfidenceGood()
	       && p.getX() < MyWorldModel.getAgentGlobalPosition().getX()){
	// if we're moving left but don't know where ball is, find ball 
	// before we decide whether to move going backwards or forwards
	guiSend("searchBall", "orange");
	return searchBall();
      }
      else{ // move normally
	ACT->putCommandInQueue(soc=moveToPos(p, PS->getPlayerWhenToTurnAngle()));
	guiSend("moveTo");
      }

      // if we know where the ball is and we are close, try to look at ball
      if( MyWorldModel.ballConfidenceGood() /* &&
	  WM->getBallPos().getDistanceTo(WM->getAgentGlobalPosition()) < 20 */) 
	{
	  /*soc = turnNeckToPoint ( WM->getBallPos(), soc ); */
	  soc = turnNeckToObject( OBJECT_BALL, soc );
	}
      else { /* ball confidence bad! */
	  soc = scanWithNeck();
	  /* jdavin: soc = turnNeckToObject( OBJECT_BALL, soc ); */

	}


      return soc;
      
#ifdef OLD_CODE      
      /* If the ball confidence is not good, we will rotate the body while
	 moving by the agent's view cone */
      if( !MyWorldModel.ballConfidenceGood() )

	/* Turn the neck by the amount calculated by search ball function */
	return SoccerCommand(CMD_TURNNECK, getSearchBallAngle());

      /* Otherwise, align the neck with the body */
      else
	return alignNeckWithBody();
#endif
      
    }

  }
}

/* This method will tell the player to go to its home position */
SoccerCommand CoachablePlayer::goHome()
{
 
  guiSend("goHome", "grey20"); 

  /* Call posRegion on the home region */
  return posRegion( homeRegion );
}


/* This method will set the default position of the player to the given
   region.
*/
SoccerCommand CoachablePlayer::setHomeRegion(Region *reg)
{
  /* Receive the pointer to the new home region, and set it to our 
     player's home region */
  if(homeRegion != NULL)
    delete homeRegion;
  homeRegion = reg->deepCopy().release();

  guiSend("setHomeRegion");

  // We don't want this here because it conflicts with coach positioning rules.
  // instead, we rely on the default action to goHome in before kick off
#ifdef OLD_BAD_CODE
  
  /* If the play mode is before the kick-off, we have to do more than this */
  if( MyWorldModel.getPlayMode() == BeforeKickOff )   {
  
    /* Get the position for the agent */
    VecPosition pos = homeRegion->getRandomPtIn(&MyWorldModel, myBindings); 
    
      
    /* If the x coordinate is greater than 0 (if the position chosen is
	 on the opponent side of the field, adjust */
    if(pos.getX() >= 0.0)
      pos.setX(-2.0);


    /* If the play mode is before kick off and it's kickoff for the opponent
       team, no players can't be inside the center circle */
    if( WM->isKickOffThem( ) && 
        pos.getDistanceTo( VecPosition(0,0) ) < 9.0 )  
      pos.setX( -10.0 );


    /* We need to adjust the position so that the player is on the field */
      
    /* First check if the x-coordinate is on the field */
    if( pos.getX() > PITCH_LENGTH / 2.0 )
      pos.setX( PITCH_LENGTH / 2.0 );
    else if( pos.getX() < -PITCH_LENGTH / 2.0 )
      pos.setX( -PITCH_LENGTH / 2.0 );
      
      
    /* Then, check if the y-coordinate is on the field */
    if( pos.getY() > PITCH_WIDTH / 2.0 )
      pos.setY( PITCH_WIDTH / 2.0 );
    else if( pos.getY() < -PITCH_WIDTH / 2.0 )
      pos.setY( -PITCH_WIDTH / 2.0 );

    if (pos.getDistanceTo(MyWorldModel.getAgentGlobalPosition()) <= 2.0)
      {
        SoccerCommand soc(CMD_ILLEGAL);
        //we're already there, so look at the ball
        if( !MyWorldModel.ballConfidenceGood() )	
          /* ACT->putCommandInQueue(searchBall() ); */
          /* just turn to center (matches behavior of original UvATrilearn players) */
          soc = turnBodyToPoint( VecPosition( 0, 0 ), 0 );
        else
          soc = turnBodyToObject( OBJECT_BALL );
        ACT->putCommandInQueue(alignNeckWithBody());
        return soc;
      }
    else
      {
        /* Teleportation! */
        return teleportToPos(pos);
      }
  }
#endif
  
  /* If the play mode doesn't match before-kickoff, then do nothing */
  return SoccerCommand(CMD_ILLEGAL);
}


/* This method will set the offside-trap line for the player at the following 
   region */
SoccerCommand CoachablePlayer::setOffsideTrap(Region *reg)
{
  static int LastPlayerSearched = 1;

  
  /* Setting offside trap: just get the middle point of the region */
  OffsideTrap = (reg->getRandomPtIn(&MyWorldModel, myBindings)).getX();

  /* Get the agent's view of the current offside line trap */
  VecPosition oinfo = WM->getOurOffsideX();

  /* Get our confidence count and current offside line trap for our team */
  double confidenceCount = oinfo.getY();
  double curOffside = oinfo.getX();

  
  /* Get the current position of the agent */
  VecPosition p = MyWorldModel.getAgentGlobalPosition();
  
  //*logger << "OFFSIDES!!!!!!!!!!!!!!!!!!!!!!!!!\n";
  //*logger << "___________________________________\n";
  
  //for(int i = 1; i <= 11;i++) {
    
  //VecPosition temp = MyWorldModel.getGlobalPosition((UNum::unum_t) i, true);

    //*logger << "Positon for player " << i << " is " << temp.getX() << "\n";
  //}

  //*logger << "Player " << (int) (MyWorldModel.getMyNum().getUNum()) << "\n";
  //*logger << "Confidence count " << confidenceCount << "\n";
  //*logger << "Current offside line is at " << curOffside << "\n";
  //*logger << "We want it to be at " << OffsideTrap << "\n\n";
  

  /* If the offside line is nearly set, then, sit down and look at the ball */
  if(confidenceCount > 10.0 && fabs(curOffside - OffsideTrap) <= 0.5) {

    if( !MyWorldModel.ballConfidenceGood() ){	
      ACT->putCommandInQueue(searchBall() );
      guiSend("searchBall", "orange");
      return alignNeckWithBody( ); // search for it
    }
    else
      ACT->putCommandInQueue( turnBodyToObject( OBJECT_BALL ) );
    
    //*logger << "Close enough!!!!!!!!!!\n";
  }

  /* Otherwise, we need to tell the player to adjust the offside line */
  if(confidenceCount > 10.0 && curOffside < OffsideTrap) {
    
    /* If the agent thinks he is the moving offside line, then move */
    if(p.getX() < OffsideTrap) {

      p.setX(p.getX() + 4.0);
      
    }

    guiSend("setOffsideTrap", "white", "purple");
    return moveToPos(p, PS->getPlayerWhenToTurnAngle() );
  }
  else {

    if(LastPlayerSearched == (int) MyWorldModel.getMyNum().getUNum())
      LastPlayerSearched++;

    if(LastPlayerSearched == 12)
      LastPlayerSearched = 1;
    
    
    ACT->putCommandInQueue(searchForPlayer((UNum::unum_t) 
					   (LastPlayerSearched++), true));
    return alignNeckWithBody();
  }
 
  
}

/* 
 * Execute goalie actions depending on which play mode we're in 
 * (and other world conditions)
 *
 * Currently, this handles goalie catching of the ball, and 
 *  the goalie's positioning and kick-off methods when it has the ball.
 * This function will return false and not execute any action if the ball 
 *  is not catchable and we're in play_on mode. 
 *
 *
 * Return: true if we executed an action
 *         false otherwise. 
 */
bool CoachablePlayer::executeGoalieBehavior(){
  SoccerCommand soc(CMD_ILLEGAL);

  if(WM->getPlayMode() == PM_PLAY_ON || WM->isFreeKickThem() ||
     WM->isCornerKickThem()){
    if(WM->isBallCatchable() &&  
       MyWorldModel.ballConfidenceGood() ){
	
      ACT->putCommandInQueue( soc = catchBall() );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      //cout <<"["<<MyWorldModel.getCurrentTime()<<"]: Goalie catching ball" <<endl;
      guiSend("Catch","green","red");
      return true;
    }
    else
      return false;
  }
  else if(  WM->isFreeKickUs() || WM->isGoalKickUs() ){

    if(!MyWorldModel.ballConfidenceGood()){
      ACT->putCommandInQueue( searchBall() );
    }
    else if( WM->isBallKickable() ){
      if( WM->getTimeSinceLastCatch() == 25 && WM->isFreeKickUs() ) {
	// move to position with fewer opponents.

	// define the top and bottom position of a rectangle in which keeper moves
	static const VecPosition posLeftTop( -PITCH_LENGTH/2.0 +
					     0.7*PENALTY_AREA_LENGTH, -PENALTY_AREA_WIDTH/4.0 );
	static const VecPosition posRightTop( -PITCH_LENGTH/2.0 +
					      0.7*PENALTY_AREA_LENGTH, +PENALTY_AREA_WIDTH/4.0 );
	int OppRight = WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS, 
					       Circle(posRightTop, 15.0 ));
	int OppLeft = WM->getNrInSetInCircle( OBJECT_SET_OPPONENTS, 
					      Circle(posLeftTop,  15.0 ));
	int OurRight = WM->getNrInSetInCircle( OBJECT_SET_TEAMMATES, 
					       Circle(posRightTop, 15.0 ));
	int OurLeft = WM->getNrInSetInCircle( OBJECT_SET_TEAMMATES, 
					      Circle(posLeftTop,  15.0 ));
	int RightDiff = OurRight - OppRight;
	int LeftDiff  = OurLeft  - OppLeft; 
	// whichever is higher means we have more power on that side

	if( RightDiff > LeftDiff )
	  soc.makeCommand( CMD_MOVE, posRightTop.getX(),posRightTop.getY(),0.0);
	else // LeftDiff >= RightDiff
	  soc.makeCommand( CMD_MOVE, posLeftTop.getX(), posLeftTop.getY(), 0.0);

	guiSend("MoveTo");
	ACT->putCommandInQueue( soc );
      }
      else if( WM->getTimeSinceLastCatch() > 28 ){
	// get rid of ball by clearing from goal area

	//soc = kickTo( VecPosition(0, (MyWorldModel.getAgentGlobalPosition().getY())  *2.0), 2.0 );    
	// ACT->putCommandInQueue( soc );
	RegRec *goalarea = new RegRec(std::auto_ptr<Point>(new PointSimple(-PITCH_LENGTH/2.,-PENALTY_AREA_WIDTH/2.)),std::auto_ptr<Point>(new PointSimple(-PITCH_LENGTH/2.+PENALTY_AREA_LENGTH,PENALTY_AREA_WIDTH/2.)) );
	ACT->putCommandInQueue( clear(goalarea) );
      }
      else if( WM->getTimeSinceLastCatch() < 25 ){
	VecPosition posSide( 0.0, MyWorldModel.getAgentGlobalPosition().getY() ); 
	if( fabs( (posSide - MyWorldModel.getAgentGlobalPosition()).getDirection() 
		  - MyWorldModel.getAgentGlobalBodyAngle()) > 10 ){
	  soc = turnBodyToPoint( posSide );
	  guiSend("turnBodyToPoint");
	  ACT->putCommandInQueue( soc );
	}
	else
	  guiSend("turnNeckToObj");
	ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
      }
    }
    else if( WM->isGoalKickUs()  ){
      guiSend("intercept", "black", "green3");
      ACT->putCommandInQueue( soc = intercept( true ) );
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }
    else{
      guiSend("turnNeckToObj");
      ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
    }

    return true;
  } // end free-kick/goal-kick handling

  return false;
}


/* This method will take a coach language action and instruct the player
   to execute it */
void CoachablePlayer::executeCommand(const ActionBinding& actbinding)
{
  Action *act = actbinding.theaction;
  myBindings = actbinding.myBindings;
  SoccerCommand soc(CMD_ILLEGAL);
  actionsource = actbinding.sourceType;

  /*
  if( !MyWorldModel.ballConfidenceGood() )
    cout << "["<<MyWorldModel.getMyNum()<<"]: "<<MyWorldModel.getCurrentTime()
	 <<" I lost ball!"<<endl;

  cout << "["<<MyWorldModel.getMyNum()<<"]: "<<MyWorldModel.getCurrentTime()<<
    ": Ball conf = "<< WM->getConfidence(OBJECT_BALL)<<endl;
  */

  //SMURF: tmp logging
  /*
  cout << MyWorldModel.getCurrentTime() << ' '
       << MyWorldModel.getMyNum().getUNum() << ' '
       << MyWorldModel.ballConfidenceGood() << ' '
       << MyWorldModel.getAgentGlobalPosition().getDistanceTo(MyWorldModel.getBallPos())
       << endl;
  */

  /* check if we're missing cycles: */
  if(LastTime != -1 && MyWorldModel.getCurrentTime() - LastTime > 1){
    cerr <<"[executeCommand]: "<<MyWorldModel.getMyNum().getUNum()<<": time="<<MyWorldModel.getCurrentTime()<<" and LastTime executed="<<
      LastTime<<endl;

  }
  LastTime = MyWorldModel.getCurrentTime();


  gui->setPlayernum(MyWorldModel.getMyNum().getUNum());

  /*
  actionlog(40) << "My number is " << (int) MyWorldModel.getMyNum().getUNum()
		<< spades::ende;
  actionlog(40) << "My position is " << MyWorldModel.getAgentGlobalPosition()
		<< spades::ende;
  */

  /* if we're goalie, do special stuff 
   * We ignore coach advice and "do our own thing." */
  if( MyWorldModel.isGoalie() ){
    bool actionExecuted = executeGoalieBehavior();
    if(actionExecuted)
      return;
    // else continue in this function
    // the goalie will position itself when posRegion is called. 

  }



  // If nothing was picked from the Action Picker, then 
  // do default action
  if(act == NULL){
    
    /* Check the agent global position */
    VecPosition agentPos = MyWorldModel.getAgentGlobalPosition();
    
    PR(8, "No action found, executing default\n");
    
    if( MyWorldModel.getPlayMode() == BeforeKickOff ) {
      // This used to be more complicated, but goHome now handles the kickoff
      // stiff correctly
      ACT->putCommandInQueue( goHome() );      
      
    }
    
    /* Otherwise, go home */
    else {
      
      /* if we don't know where ball is, look for it. */
      if(!MyWorldModel.ballConfidenceGood()) {
	ACT->putCommandInQueue( searchBall() );
	guiSend("searchBall", "orange");
	ACT->putCommandInQueue(alignNeckWithBody( )); 
	return;
      }

      /* only go home if we have plenty of stamina */
      if(MyWorldModel.getAgentStamina().getStamina() > 2500. 
	 || MyWorldModel.getPlayMode() != PlayOn)
	ACT->putCommandInQueue( goHome() );
      else{ /* low on stamina. just rest and watch ball */
	//cout << "["<<MyWorldModel.getMyNum()<<"]: "<<"No action and stamina "<<
	//MyWorldModel.getAgentStamina().getStamina()<<" so I'm resting."<<endl;

	guiSend("turnBodyToObj");
	/* don't need to check ball confidence. already checked above */
	ACT->putCommandInQueue( turnBodyToObject( OBJECT_BALL ) );
      }
    }    
 
    return;     
  }



  /* ActionVisitor for the action */
  ActionVisitor avisitor(actbinding.myBindings);
  
  act->accept(avisitor);  /* extract action information */

  
  /* Now, fall into the switch statements depending on the action type */
  switch(avisitor.getActionType()) {
    
    /* Position the player in a given region */
  case ActionVisitor::POS:
    
    //    PR(8, "[executeCommand]: pos\n");

    ACT->putCommandInQueue(soc = posRegion((Region *) avisitor.getRegion())); 
    //ACT->putCommandInQueue(turnNeckToObject( OBJECT_BALL, soc ) );

    break;
    
    /* Set home region */
  case ActionVisitor::HOME:
    
    //PR(8, "[executeCommand]: home\n");
    
    ACT->putCommandInQueue(soc = setHomeRegion((Region *) avisitor.getRegion()));
    //ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
 
    break;
    
    /* Mark a player from a given UNumSet */
  case ActionVisitor::MARK:
    
    //PR(8, "[executeCommand]: markunumset\n");
    ACT->putCommandInQueue(markUNumSet(avisitor.getPlayers()));
    
    break;
    
    /* Mark a player from a given UNumSet to the passing lane to the ball */
  case ActionVisitor::MARKLINEPLAYER:
    
    //PR(8, "[executeCommand]: marklineplayer\n");
    ACT->putCommandInQueue(markLaneUNumSet(avisitor.getPlayers()));
        
    break;
    
    /* Mark a point in the region to the passing lane to the ball */
  case ActionVisitor::MARKLINEREG:
    
    //PR(8, "[executeCommand]: marklinereg\n");
    ACT->putCommandInQueue(markLaneRegion((Region *)avisitor.getRegion()));
    
    
    break;
    
    /* Set offsideline trap for the player */
  case ActionVisitor::OFFSIDESLINE:
    
    //PR(8, "[executeCommand]: offsidesline\n");
    ACT->putCommandInQueue(setOffsideTrap((Region *)avisitor.getRegion()));
    
    break;
    
    /* Passing to a player in a region */
  case ActionVisitor::PASSREG:
    
    //PR(8, "[executeCommand]: passreg\n");
    
    ACT->putCommandInQueue(passToRegion((Region *) avisitor.getRegion()));
 
    
    break;
    
    /* Passing to a player from a UNumSet */
  case ActionVisitor::PASSUNUM:
    
    //PR(8, "[executeCommand]: passunum\n");
    
    ACT->putCommandInQueue(soc = passToUNumSet(avisitor.getPlayers()));
    ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );
        
    break;
    
    /* Dribbling the ball to a region */
  case ActionVisitor::DRIBBLE:
    
    //PR(8, "[executeCommand]: dribble\n");
    
    ACT->putCommandInQueue(soc = dribbleToRegion((Region *)avisitor.getRegion()));
    ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );

    break;
    
    /* Clearing the ball outside the region */
  case ActionVisitor::CLEAR:
    
    //PR(8, "[executeCommand]: clear\n");   

    ACT->putCommandInQueue(soc = clear((Region *) avisitor.getRegion()));
    ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );

    break;
    
    /* Shoot the ball toward the goal */
  case ActionVisitor::SHOOT:
    
    //PR(8, "[executeCommand]: shoot\n");
    
    ACT->putCommandInQueue(soc = shoot());
    ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );

    break;
    
    /* Hold the ball in its place */
  case ActionVisitor::HOLD:
    
    //PR(8, "[executeCommand]: hold\n");
    
    ACT->putCommandInQueue(soc = hold());
    ACT->putCommandInQueue( turnNeckToObject( OBJECT_BALL, soc ) );

    break;
    
    /* Attempt interception */
  case ActionVisitor::INTERCEPT:
    
    //PR(8, "[executeCommand]: intercept\n");    
    ACT->putCommandInQueue(interceptBall());

    break;
    
    /* Attempt tackling from a given UNumSet*/
  case ActionVisitor::TACKLE:
    
    //PR(8, "[executeCommand]: tackle\n");
    ACT->putCommandInQueue(tackle(avisitor.getPlayers()));
    
    break;

  default:
    cerr << "ERROR in executeCommand: ActionVisitor type unhandled" <<endl;
    assert(0);
    break;
   
  }

}
