/*
 * CommPicker - picks communication message units to put into a say message.
 *   Used by CoachableAgentComm to set up a message to send. 
 *
 *
 * Creation Date: 12/18/03
 * Author: John Davin
 *
 */


#include "CommPicker.h"
#include "config.h"


CommPicker::CommPicker(CoachableAgentComm *agent){

  AgentComm = agent;

  //  utilityAccessor = new CommRandom();
  //utilityAccessor = new CommOurPosAndBallPos();

  //utilityAccessor = new CommWantPass(); //CommPassTo();
  //utilityAccessor = new CommNormal();
  utilityAccessor = new CommInfo();

}

CommPicker::~CommPicker(){
  if(utilityAccessor != NULL) delete utilityAccessor;
}

/*
 * Comparison function for sorting the availableMsgs vector
 */
bool CommPicker::SortMsgsByP(const MsgPpair &a, const MsgPpair &b){

  return a.p > b.p;
}


/*
 * Pick messages based on our heuristic. 
 * 'availableSize' is the number of characters we can use.
 *
 * Algorithm: availableMsgs is filled with all the messages types paired 
 *   with their utility values. We then filter out any messages that 
 *   are not applicable (predicates not satisfied). 
 * We then sort the messages by utility value (p) so that the highest priority
 * ones come first. 
 *  Then, while there are messages remaining to choose from: 
 *    If there is not enough room for this msg, we remove it and go to next. 
 *    Otherwise, we get a random number and add this message to the encoder
 *    if the random number falls within this msg type's probability (p)
 *    If so, we also remove the message since we have just used it. 
 *  This loop will run until all possible message units have been added to 
 *  the encoder.
 *
 */
void CommPicker::pickMsgs(unsigned int availableSize){
  unsigned int size = 0; // current size
  vector<MsgPpair> availableMsgs;

  // seed random number generator with microseconds
  struct timeval mytime;
  gettimeofday(&mytime, NULL);
  srand(mytime.tv_usec); // microseconds

  /*
  MySayMsgTypes *test;
  test = new MyOurPos();
  LOGOUT(50) << test<<endl;
  delete test;

  test = new MyOurPos();
  LOGOUT(50) << test<<endl;
  delete test;

  test = new MyOurPos();
  LOGOUT(50) << test<<endl;
  test->isSatisfied();
  ((SayMsgTypes*)test)->getName();
  delete test;
  */


  // initialize the storage with messages
  add(new MyOurPos(),       utilityAccessor->getOurPos(),       &availableMsgs);
  add(new MyBallPos(),      utilityAccessor->getBallPos(),      &availableMsgs);
  add(new MyBallVel(),      utilityAccessor->getBallVel(),      &availableMsgs);
  add(new MyWeHaveBall(),   utilityAccessor->getWeHaveBall(),   &availableMsgs);
  add(new MyOppHasBall(),   utilityAccessor->getOppHasBall(),   &availableMsgs);
  add(new MyPassToPlayer(), utilityAccessor->getPassToPlayer(), &availableMsgs);
  add(new MyPassToCoord(),  utilityAccessor->getPassToCoord(),  &availableMsgs);
  add(new MyWantPass(),     utilityAccessor->getWantPass(),     &availableMsgs);
  add(new MyOppPos(),       utilityAccessor->getOppPos(),       &availableMsgs);
  add(new MyTeammatePos(),  utilityAccessor->getTeammatePos(),  &availableMsgs);


  // filter out messages that are not applicable at this time.
  filter(&availableMsgs);

  const unsigned int totalMsgs = availableMsgs.size();

  if(totalMsgs == 0){
    cerr <<"[CommPicker]: Warning: Could not select any messages because "
	 <<"none match predicates, or they all have 0 probability values."<<endl;
    clearMsgs(&availableMsgs);
    return;
  }


  unsigned int msgsRemaining = totalMsgs;

  // first, pick any msgs that have p>=1. 
  for(unsigned int i=0; i<availableMsgs.size(); i++){
      unsigned int cost = ((SayMsgTypes*)availableMsgs[i].msg)->getCost();

      // if there is not enough room for this message, skip it
      if( cost + size > availableSize)
			 continue;
		if( availableMsgs[i].p >= 1.0){

			 availableMsgs[i].msg->addToEncoder(AgentComm->getEncoder());
			 size += cost;
			 LOGOUT(50) <<"Encoding "<<((SayMsgTypes*)availableMsgs[i].msg)->getName()<<
				  " because p>=1. size is now "<<size<<" and msgsRemaining="<<msgsRemaining<<endl;
			 // now that we used the message, remove it
			 availableMsgs[i].deleted = true;
			 msgsRemaining--;
		}
  }


  // normalize the probabilities
  normalizeP(&availableMsgs);

  // sort messages by probability: 
  // if one or more have the same probability, sort them randomly
  // sort the messages by p value
  //  std::sort(availableMsgs.begin(), availableMsgs.end(), &SortMsgsByP);

  // but we also have to randomize entries that have equal probabilities
  //randomizeEquals(&availableMsgs);

  LOGOUT(50) << "Filtered & normalized availableMsgs:"<<endl;
  for(unsigned int i=0; i<availableMsgs.size(); i++){
    LOGOUT(50) <<"["<<i<<"]: " 
	 <<((SayMsgTypes*)availableMsgs[i].msg)->getName()<<" "
	 <<availableMsgs[i].p<<endl;
  }


  while(msgsRemaining) {
    // while there are still message units remaining

    for(unsigned int i=0; i<totalMsgs; i++){
      if(availableMsgs[i].deleted) // this message has been removed.
			 continue;

      unsigned int cost = ((SayMsgTypes*)availableMsgs[i].msg)->getCost();

      // if there is not enough room for this message or its probability
      // of being picked is close to 0, remove it and go to next.
      if( cost + size > availableSize || availableMsgs[i].p < 0.01){
			 availableMsgs[i].deleted = true; // remove this message
			 msgsRemaining--;
			 normalizeP(&availableMsgs); // need to renormalize
			 continue;
      }

      // with probability of availableMsgs[i].p, add this message to the encoder.
      if( rand()/(RAND_MAX+1.0)  < availableMsgs[i].p){
			 // the random float will range from 0 to RAND_MAX/(RAND_MAX+1). 
			 // So if p=1, this will always be true, and if p=0 it will never be true.

	availableMsgs[i].msg->addToEncoder(AgentComm->getEncoder());
	size += cost;
	LOGOUT(50) <<"Encoding "<<((SayMsgTypes*)availableMsgs[i].msg)->getName()<<
	  ", size is now "<<size<<" and msgsRemaining="<<msgsRemaining<<endl;
	// now that we used the message, remove it
	availableMsgs[i].deleted = true;
	msgsRemaining--;
	normalizeP(&availableMsgs); // need to renormalize
      }

    }

  }

  LOGOUT(50) <<endl;
  //  LOGOUT(50) << "Done picking msg"<<endl;
  clearMsgs(&availableMsgs);
}



/*
 * Add (msg, p) to the storage
 *
 */
void CommPicker::add(MySayMsgTypes *msg, float p, vector<MsgPpair> *msgVector){

  msgVector->push_back(MsgPpair(msg, p));
}



/*
 * Filters out messages that are not applicable (predicates not satisfied).
 * (eg, PassToPlayer is not applicable if we are not passing 
 * to a player now). 
 * Also removes messages with probability (p) of < 0.01, since that is 
 * effectively zero and likely to never get picked. 
 */
void CommPicker::filter(vector<MsgPpair> *msgsVector){

  vector<MsgPpair>::iterator it=msgsVector->begin(); 

  while(it != msgsVector->end()){
    //LOGOUT(50) << "filtering "<<((SayMsgTypes*)(*it).msg)->getName()<<"("<<(*it).msg<<")"<<endl;

    if( ((MySayMsgTypes*)(*it).msg)->isSatisfied() != true
	|| (*it).p < 0.01 ){
      //LOGOUT(50) <<"erasing "<<((SayMsgTypes*)(*it).msg)->getName()<<" at "<<it<<endl;
      delete (*it).msg;
      msgsVector->erase(it);

      // set iterator back to beginning because we erased an entry 
      // (invalidates position of previous iterator)
      it = msgsVector->begin();
    }
    else
      it++;
  }

}

/*
 * Normalize the probabilities in the vector. 
 * Divide each p by the total sum
 * (so (.1,.1,.1) -> (.33,.33,.33)
 *
 * Only count messages that are not deleted!
 */
void CommPicker::normalizeP(vector<MsgPpair> *msgsVector){
  float total = 0.;

  // first we have to get the total sum
  for(unsigned int i=0; i<msgsVector->size(); i++){
		//if( (*msgsVector)[i].p < 1.0 )
		if( ! ((*msgsVector)[i].deleted) )
			 total += (*msgsVector)[i].p;
  }

  if(total == 0.) return;

  // now divide each p by the total
  for(unsigned int i=0; i<msgsVector->size(); i++){
		//if( (*msgsVector)[i].p < 1.0 )
		if( ! ((*msgsVector)[i].deleted) )
			 (*msgsVector)[i].p /= total;
  }


  LOGOUT(50) << "Normalized availableMsgs:"<<endl;
  for(unsigned int i=0; i<msgsVector->size(); i++){
    LOGOUT(50) <<"["<<i<<"]: " 
	 <<((SayMsgTypes*)(*msgsVector)[i].msg)->getName()<<" "
					<<(*msgsVector)[i].p<<" "<<(*msgsVector)[i].deleted<<endl;
  }

}


/*
 * Given the *sorted* vector of message+p pairs, randomize entries that 
 * have equal p's and return the new vector.
 */
void CommPicker::randomizeEquals(vector<MsgPpair> *msgsVector){
  float previous = -1.0;

  vector<MsgPpair> equals;
  vector<MsgPpair> vectorNew;

  for(unsigned int i=0; i<msgsVector->size(); i++){

    if((*msgsVector)[i].p == previous){

      equals.push_back((*msgsVector)[i]);
    }


    if((*msgsVector)[i].p != previous || i==msgsVector->size()-1){ // new p value

      // take all the entries from equals and randomly push them onto vectorNew
      while(!equals.empty()){

	// get a random number between 0 and equals.size()-1 (inclusive)
	unsigned int index = (unsigned int) (rand()/(RAND_MAX+1.0) * equals.size());
	assert(index < equals.size());

	vectorNew.push_back(equals[index]);
	equals.erase(equals.begin() + index);
      }

      equals.push_back((*msgsVector)[i]);


      // special case: if this is the last entry, and it was not equal to the 
      // previous p, then it needs to be added to the end of vectorNew.
      if(i==msgsVector->size()-1 && (*msgsVector)[i].p != previous)
	vectorNew.push_back((*msgsVector)[i]);

    }

    previous = (*msgsVector)[i].p;
  }

  // all done, so just set msgsVector equal to vectorNew
  *msgsVector = vectorNew;
}


/*
 * clear the storage vector and free the pointers (allocated in pickMsgs())
 *
 */
void CommPicker::clearMsgs(vector<MsgPpair> *msgsVector){

  for(unsigned int i=0; i<msgsVector->size(); i++){  
    delete (*msgsVector)[i].msg;
  }
  msgsVector->clear();

}






