/*
 * CommUtility - this is a simple set of classes that can be used to define 
 *   utility values for each of the message types. CommUtility is used by the
 *   CommPicker to determine what probability to assign each of the message 
 *   types when determining which ones to put in a message. 
 *
 *
 *
 * Creation Date: 12/23/03
 * Author: John Davin
 *
 */

#ifndef COMM_UTILITY_H
#define COMM_UTILITY_H


class CommUtility {
 public: 
  virtual float getOurPos() = 0;
  virtual float getBallPos() = 0;
  virtual float getBallVel() = 0;
  virtual float getWeHaveBall() = 0;
  virtual float getOppHasBall() = 0;
  virtual float getPassToPlayer() = 0;
  virtual float getPassToCoord() = 0;
  virtual float getWantPass() = 0;
  virtual float getOppPos() = 0;
  virtual float getTeammatePos() = 0;

};



/*
 * Utility functions should return values only in the range [0.0, 1.0].
 * A value of 0.0 means the message type will never be used, and 1.0 
 * means it will always be used (if there is space available in the msg).
 *
 * However, functions can return values >= 1.0. The only purpose to this
 * is if you have a couple message types that you want to guarantee are 
 * included, but one is more important than the other. Then you can return
 * a value like 1.1 for the more important one and it will be picked before the
 * 1.0 one. 
 * 
 */



/*
 * return a uniform probability distribution
 * (each p = 1/N where N = 10 because there are 10 types)
 */
class CommRandom : public CommUtility {
 public: 
  float getOurPos(){      return 0.1;  } 
  float getBallPos(){     return 0.1;  }
  float getBallVel(){     return 0.1;  }
  float getWeHaveBall(){  return 0.1;  }
  float getOppHasBall(){  return 0.1;  }
  float getPassToPlayer(){return 0.1;  }
  float getPassToCoord(){ return 0.1;  }
  float getWantPass(){    return 0.1;  }
  float getOppPos(){      return 0.1;  }
  float getTeammatePos(){ return 0.1;  }

};


/*
 * return 1.0 for OurPos and BallPos
 *
 */
class CommOurPosAndBallPos : public CommUtility {
 public: 
  float getOurPos(){      return 1.;   }
  float getBallPos(){     return 1.;   }
  float getBallVel(){     return 0.0;  }
  float getWeHaveBall(){  return 0.0;  }
  float getOppHasBall(){  return 0.0;  }
  float getPassToPlayer(){return 0.0;  }
  float getPassToCoord(){ return 0.0;  }
  float getWantPass(){    return 0.0;  }
  float getOppPos(){      return 0.0;  }
  float getTeammatePos(){ return 0.0;  }

};



/*
 * return optimal utility values for each message type
 * (functions may consider world model)
 */
class CommNormal : public CommUtility {
 public: 
  float getOurPos(){      return 0.5;  }
  float getBallPos(){     return 0.5;  }
  float getBallVel(){     return 0.3;  }
  float getWeHaveBall(){  return 0.2;  }
  float getOppHasBall(){  return 0.2;  }
  float getPassToPlayer(){return 1.1;  }
  float getPassToCoord(){ return 0.8;  }
  float getWantPass(){    return 1.0;  }
  float getOppPos(){      return 0.4;  }
  float getTeammatePos(){ return 0.4;  }

};


/*
 * for testing. 
 * equal to CommNormal but disables WantPass, PassTo*
 */
class CommInfo : public CommUtility {
 public: 
  float getOurPos(){      return 0.5;  }
  float getBallPos(){     return 0.5;  }
  float getBallVel(){     return 0.3;  }
  float getWeHaveBall(){  return 0.2;  }
  float getOppHasBall(){  return 0.2;  }
  float getPassToPlayer(){return 0.0;  }
  float getPassToCoord(){ return 0.0;  }
  float getWantPass(){    return 0.0;  }
  float getOppPos(){      return 0.4;  }
  float getTeammatePos(){ return 0.4;  }

};

/* For testing. 
 * Only activates PassToPlayer and PassToCoord msgs
 */
class CommPassTo : public CommUtility {
 public: 
  float getOurPos(){      return 0.0;  }
  float getBallPos(){     return 0.0;  }
  float getBallVel(){     return 0.0;  }
  float getWeHaveBall(){  return 0.0;  }
  float getOppHasBall(){  return 0.0;  }
  float getPassToPlayer(){return 0.1;  }
  float getPassToCoord(){ return 0.1;  }
  float getWantPass(){    return 0.0;  }
  float getOppPos(){      return 0.0;  }
  float getTeammatePos(){ return 0.0;  }

};


/* For testing. 
 * Only activates WantPass msg
 */
class CommWantPass : public CommUtility {
 public: 
  float getOurPos(){      return 0.0;  }
  float getBallPos(){     return 0.0;  }
  float getBallVel(){     return 0.0;  }
  float getWeHaveBall(){  return 0.0;  }
  float getOppHasBall(){  return 0.0;  }
  float getPassToPlayer(){return 0.0;  }
  float getPassToCoord(){ return 0.0;  }
  float getWantPass(){    return 1.0;  }
  float getOppPos(){      return 0.0;  }
  float getTeammatePos(){ return 0.0;  }

};





#endif
