/*
 * FocusStrategy - this is a simple set of classes that can be used to define 
 * different strategies for choosing what player to focus on for receiving 
 * communication messages. 
 * The main purpose of these classes is just for testing different focus strategies.
 *  When not testing, FocusNormal will be our only actual strategy. 
 * 
 * Creation Date: 1/1/04
 * Author: John Davin
 */


#include "FocusStrategy.h"

// Server setting for the range that a player can hear
#define HEAR_RANGE (WM->getServerSettings()->getAudioCutDist() - 1.0)
// the  -1.0 is to give us a margin of error. ie, if we change our attention
// slightly before our current listen to player goes out of range, we'll insure 
// that we switch in time. 


/***** class FocusNormal : public FocusStrategy ****/

/*
 * Focus on ball owner, but if ball owner is far away from ball (need good conf 
 *  on ball and owner to determine this), then focus on player closest to ball.
 * If we are ball owner, focus on closest player?
 *
 */
unsigned short FocusNormal::getDesiredFocus(){
  VecPosition mypos = WM->getAgentGlobalPosition();

  // am I ballowner?
  ObjectT ballowner = WM->getBallOwner();
  if(ballowner == WM->getAgentObjectType()){ // <I am ball owner>

    // Is there a player upfield of us? (closer to right out-of-bounds line)
    VecPosition rightmost = mypos;
    int rightmostplayer = -1;
    for(int i=0; i<11; i++){
      ObjectT player = SoccerTypes::getTeammateObjectFromIndex(i);

      // only check players that we're confident of.
      if( ! WM->isConfidenceGood(player))
	continue;

      VecPosition pos = WM->getGlobalPosition(player);

      // if this player is out of hearing range, skip him
      if(mypos.getDistanceTo(pos) > HEAR_RANGE)
	continue;


      if(pos.isInFrontOf(rightmost)){
	rightmost = pos;
	rightmostplayer = i+1;
      }
    }

    if(rightmostplayer != -1){ // <<there's a player upfield of us>>
      Log.log(66, "[FocusNormal]: I am bowner. focusing on %d", rightmostplayer);
      return rightmostplayer;
    }
    else { // <<no player upfield of us>>
      Log.log(66, "[FocusNormal]: I am bowner. focusing on %d", 0);
      return 0; // turn off focusing. (randomly get messages).
    }

  }
  else { // <Not ball owner>


    VecPosition bownerPos = WM->getGlobalPosition(ballowner);
    VecPosition ballpos   = WM->getGlobalPosition(OBJECT_BALL);
    double distToBall = bownerPos.getDistanceTo(ballpos);

    // REMOVE THIS
    assert(SoccerTypes::isKnownPlayer(ballowner) 
	   || WM->isConfidenceGood(ballowner) == false);

    // do we know where ballowner is && is he close to ball (<15) && within hear range?
    if(SoccerTypes::isTeammate(ballowner) && WM->isConfidenceGood(ballowner) 
       && distToBall < 15.0 && mypos.getDistanceTo(bownerPos) <= HEAR_RANGE ){ 
      // <<ballowner close to ball && within hear range>>

      Log.log(66, "[FocusNormal]: ballowner=%d, agentobj=%d", ballowner, WM->getAgentObjectType());
      Log.log(66, "[FocusNormal]: focusing on bowner (%d)",
		      SoccerTypes::getIndex(ballowner)+1);
      return SoccerTypes::getIndex(ballowner)+1;
    }
    else{ // <<ballowner not close to ball or we don't know where he is>>

      // focus on closest known player to ball

      double closestdist = 1000.;
      int closestplayer = -1;
      for(int i=0; i<11; i++){
	ObjectT player = SoccerTypes::getTeammateObjectFromIndex(i);

	// only check players that we're confident of.
	if( ! WM->isConfidenceGood(player) )
	  continue;


	VecPosition playerpos = WM->getGlobalPosition(player);

	// if this player is out of hearing range, skip him
	if(mypos.getDistanceTo(playerpos) > HEAR_RANGE)
	  continue;

	double distToBall = playerpos.getDistanceTo(ballpos);

	if( distToBall < closestdist ){
	  closestdist = distToBall;
	  closestplayer = i+1;
	}
      }

      if(closestplayer != -1){ // we found the closest player to ball
	Log.log(66, "[FocusNormal]: focusing on closest player to ball (%d)",closestplayer);
	return closestplayer;
      }
      else{ // couldn't find closest player to ball so use random focus.
	Log.log(66, "[FocusNormal]: couldn't find closest to ball. focusing on 0");
	return 0;
      }

    }


  } // end <Not ball owner> branch.

} // end function




/***** class FocusTest : public FocusStrategy  *****/

/*
 * Focus on player 2. If we are player 2, focus on player 3.
 */
unsigned short FocusTest::getDesiredFocus(){

  if(WM->getPlayerNumber() == 2)
    return 3;
  else
    return 2;

  /*  if(WM->getCurrentCycle() == 50 && WM->getPlayerNumber() != 3) // just to test switching
      return 3;
  */

}



/*** class FocusRandom : public FocusStrategy ****/

unsigned short FocusRandom::getDesiredFocus(){

  // if we always return 0, the player will not set a focus player,
  // and the soccer server will thereby randomly select a say message 
  // from the available messages within hearing distance. 
  return 0;
  
}






