/*
 * MySayMsgTypes.h - Defines subclasses of the SayMsgTypes in order to 
 *  define our process() functions (which can add data to the world model)
 *
 * Creation Date: 12/4/03
 * Author: John Davin
 *
 *
 */

#ifndef MY_SAY_MSG_TYPES_H
#define MY_SAY_MSG_TYPES_H


#include "saymsglib/SayMsgTypes.h"
#include "saymsglib/SayMsgEncoder.h"

#include "WorldModel.h"
#include "config.h"

extern WorldModel *trilearn_wm; // from PlayerInterface.C
#define WM trilearn_wm


/*
 * Note on cycle times: 
 * All cycle times (number of cycles since saw ball for example) are increased by 1
 * when sending them to the WorldModelCommUpdate.  This is because say messages are 
 * generally not received until the cycle *after* they were sent. So although the player
 * may have seen the ball in cycle 68 (resulting in cycleno=0), we do not receive his message
 * until cycle 69, which means the information is now 1 cycle old. 
 *
 */



class MySayMsgTypes : virtual SayMsgTypes {
 public:
  virtual bool addToEncoder(SayMsgEncoder *MyEncoder) = 0;

  virtual bool isSatisfied() = 0;

};



/*
 * OurPos(X,Y): 
 * float X - our position X value 
 * float Y - our position Y value
 * Cost = 3
 */
class MyOurPos : public OurPos, public MySayMsgTypes  {

 public:   
  void process(){
    /* add X and Y positions of sending player to our world model */
    
    LOGOUT(49) << "MyOurPos: playernum="<<SenderPlayerNum<<" X="<<X<<" Y="<<Y<<endl;

    WM->updateTeammatePos(SenderPlayerNum, X, Y, 1); 
    // 1 as number of cycles because player is always 100% confident of his own position
    // (and this say message is 1 cycle old)

  }

  bool addToEncoder(SayMsgEncoder *MyEncoder){
    VecPosition ourpos = WM->getAgentGlobalPosition();
    float X = ourpos.getX();
    float Y = ourpos.getY();

    DataTypes::X::bound(&X);
    DataTypes::Y::bound(&Y);
    
    bool rtn = MyEncoder->add( new OurPos(X,Y) ); // cost 3
    if(!rtn)
      cerr <<"[MyOurPos]: WARNING: add failed!"<<endl;
    return rtn;
  }

  bool isSatisfied(){
    return true;
  }

};

/*
 * BallPos(X,Y,cycleno): 
 * float X - ball position X value 
 * float Y - ball position Y value
 * unsigned short cycleno - number of cycles since saw ball
 * Cost = 4
 */
class MyBallPos : public BallPos, public MySayMsgTypes {

 public: 
  void process(){
    
    LOGOUT(49) << "MyBallPos: playernum="<<SenderPlayerNum<<" reports ball at ("<<X
	 <<","<<Y<<") with confidence "<<cycleno<<" cycles."<<endl;

    WM->updateBallPos(X,Y,cycleno+1);
  }


  bool addToEncoder(SayMsgEncoder *MyEncoder){
    VecPosition ballpos = WM->getBallPos();
    Time lastseen = WM->getTimeLastSeen( OBJECT_BALL);
    int cycles = WM->getCurrentTime().getTimeDifference(lastseen);
    float X = ballpos.getX();
    float Y = ballpos.getY();

    //    LOGOUT(49) << ballpos.getX() << " " << ballpos.getY() << " " 
    //	 << WM->getCurrentTime() << "-" <<lastseen <<"="<<cycles<< endl;

    DataTypes::X::bound(&X);
    DataTypes::Y::bound(&Y);
    DataTypes::Cycleno::bound(&cycles);

    bool rtn = MyEncoder->add( new BallPos(X,Y,cycles) ); // cost 4
    if(!rtn)
      cerr <<"[MyBallPos]: WARNING: add failed!"<<endl;
    return rtn;
  }

  bool isSatisfied(){
    return WM->isConfidenceGood(OBJECT_BALL);
  }
};

/*
 * BallVel(dX, dY, cycles): 
 * float dX - ball velocity X value 
 * float dY - ball velocity Y value
 * unsigned short cycleno - number of cycles since saw ball
 * Cost = 4
 */
class MyBallVel : public BallVel, public MySayMsgTypes {

 public: 
  void process(){
    LOGOUT(49) << "MyBallVel: playernum="<<SenderPlayerNum<<" reports ball at velocity ("
	 << dX<<","<<dY<<") with cycles="<<cycleno<<endl;

    WM->updateBallVel(dX,dY,cycleno+1);
  }

  bool addToEncoder(SayMsgEncoder *MyEncoder){
    VecPosition ballvel = WM->getBallVelocity();
    Time lastseen = WM->getTimeLastSeen( OBJECT_BALL);
    int cycles = WM->getCurrentTime().getTimeDifference(lastseen);
    float dX = ballvel.getX();
    float dY = ballvel.getY();

    //    LOGOUT(49) << "vel: "<<dX << " " << dY << " " << WM->getCurrentTime() << "-" 
    // <<lastseen <<"="<<cycles<< endl;

    DataTypes::dX::bound(&dX);
    DataTypes::dY::bound(&dY);
    DataTypes::Cycleno::bound(&cycles);

    bool rtn = MyEncoder->add( new BallVel(dX,dY,cycles) );
    if(!rtn)
      cerr <<"[MyBallVel]: WARNING: add failed!"<<endl;
    return rtn;
  }

  bool isSatisfied(){
    return WM->isConfidenceGood(OBJECT_BALL);
  }
};


/*
 * WeHaveBall(playernum): 
 * unsigned short playernum - number of player who has ball
 */
class MyWeHaveBall : public WeHaveBall, public MySayMsgTypes {

 public: 
  
  /*
   * If we receive this message, the Sending player thinks he knows who the ball owner is.
   * Since this message type has no form of confidence value, we have to assume that 
   * the sender will only send this message when he is pretty confident that he knows 
   * the correct ball owner. 
   * Therefore, we will simply set our world model ball owner to the info we are given. 
   * This is only actually useful when we not know the correct ball owner - 
   *  this likely will only happen when the player has lost sight of the ball, or the ball
   *  is out of visible range. 
   * Note: our world model attempts to update the ball owner every cycle by checking 
   * if any players are close to the ball. So if a saymsg gives us inaccurate data about 
   * who is ball owner, then our ball owner variable will be set incorrectly until 
   * we see a new player gain possession of the ball. 
   */
  void process(){

    if(playernum == 0)
      LOGOUT(49) << "MyWeHaveBall: playernum="<<SenderPlayerNum
	   <<" reports that we have ball (playernum=0)."<<endl;
    else
      LOGOUT(49) << "MyWeHaveBall: playernum="<<SenderPlayerNum<<" reports our player "
	 <<playernum<<" has ball."<<endl;

    WM->updateBallOwner(playernum, OUR_TEAM);

  }

  bool addToEncoder(SayMsgEncoder *MyEncoder){
    // We assume that this function is only called when we own the ball

    int playernum = SoccerTypes::getIndex(WM->getBallOwner());
    if(playernum < 0){ // we don't know the player number
      cerr << "[addWeHaveBall]: Warning: ball owner # is unknown"<<endl;
      playernum = 0;
    }
    else
      playernum++; // getIndex returns player 1 = 0.

    DataTypes::Playernum::bound(&playernum);

    bool rtn = MyEncoder->add( new WeHaveBall(playernum) );
    if(!rtn)
      cerr <<"[CoachableAgentComm:addWeHaveBall]: WARNING: add failed!"<<endl;
    return rtn;
  }

  /* 
   *  true if ball confidence is good and we own the ball 
   */
  bool isSatisfied(){
    return WM->isConfidenceGood(OBJECT_BALL) && SoccerTypes::isTeammate(WM->getBallOwner());
  }

};


/*
 * OppHasBall member variables: 
 * unsigned short playernum - number of player who has ball
 */
class MyOppHasBall : public OppHasBall, public MySayMsgTypes {

 public: 
  void process(){
    if(playernum == 0)
      LOGOUT(49) << "MyOppHasBall: playernum="<<SenderPlayerNum
	   <<" reports opp team has ball (playernum=0)"<<endl;
    else
      LOGOUT(49) << "MyOppHasBall: playernum="<<SenderPlayerNum<<" reports opp player "
	   <<playernum<<" has ball."<<endl;

    WM->updateBallOwner(playernum, OPPONENT_TEAM);
  }

  bool addToEncoder(SayMsgEncoder *MyEncoder){
    // We assume that this function is only called when opponent team owns the ball

    int playernum = SoccerTypes::getIndex(WM->getBallOwner());
    if(playernum < 0){ // we don't know the player number
      cerr << "[addOppHasBall]: Warning: ball owner # is unknown!"<<endl;
      playernum = 0;
    }
    else
      playernum++; // getIndex returns player 1 = 0.

    DataTypes::Playernum::bound(&playernum);

    bool rtn = MyEncoder->add( new OppHasBall(playernum) );
    if(!rtn)
      cerr <<"[CoachableAgentComm:addOppHasBall]: WARNING: add failed!"<<endl;
    return rtn;
  }

  /* 
   *  true if ball confidence is good and opponent team owns the ball 
   */
  bool isSatisfied(){
    return WM->isConfidenceGood(OBJECT_BALL) && SoccerTypes::isOpponent(WM->getBallOwner());
  }
};

/*
 * PassToPlayer member variables: 
 * unsigned short playernum - number of player who has ball
 */
class MyPassToPlayer : public PassToPlayer, public MySayMsgTypes {

 public: 
  void process(){    
    LOGOUT(49) << "MyPassToPlayer: playernum="<<SenderPlayerNum<<
      " is passing to player "<<playernum<<endl;

    // if the sender is not passing to us, then we don't care about this message
    // (we're not going to do anything special when the sender is passing to our teammate)
    if(playernum != WM->getPlayerNumber()){
      return;
    }

    WM->updatePassToPlayer(SenderPlayerNum);
    // updatePassToPlayer should *only* be called when a player is passing to us
    // (ie, playernum == WM->getPlayerNumber()). We currently do not support 
    // using PassToPlayer info that specifies a teammate passing to a 
    // teammate who is not us
  }

  bool addToEncoder(SayMsgEncoder *MyEncoder){
    DataTypes::Playernum_t playernum;

    // REMOVE THIS later. Our predicate requirement for this function should be 
    // that WM->getPassToTime() != -1 and our evaluation should be based on whether
    // WM->getPassToTime() == WM->getCurrentTime(). 
    assert( WM->getPassingToTime() != -1 );

    playernum = WM->getPassingToPlayer(); // returns the player we're currently passing to

    DataTypes::Playernum::bound(&((int)playernum));

    bool rtn = MyEncoder->add( new PassToPlayer(playernum) );
    if(!rtn)
      cerr <<"[CoachableAgentComm:addPassToPlayer]: WARNING: add failed!"<<endl;
    return rtn;
  }

  /* 
   *  true if we executed a passto action within the last 5 cycles
   */
  bool isSatisfied(){
    return (WM->getPassingToTime() != -1 && 
            WM->getPassingToTime() >= WM->getCurrentTime() - 5 );
  }
};

/*
 * PassToCoord member variables: 
 * float X - X coordinate of position we're passing to.
 * float Y - Y coordinate of position we're passing to. 
 */
class MyPassToCoord : public PassToCoord, public MySayMsgTypes {

 public: 
  void process(){
    LOGOUT(49) << "MyPassToCoord: playernum="<<SenderPlayerNum<<" is passing to ("
	 <<X<<","<<Y<<")"<<endl;

    WM->updatePassToCoord(SenderPlayerNum,X,Y);
  }

  bool addToEncoder(SayMsgEncoder *MyEncoder){
    DataTypes::X_t X;
    DataTypes::Y_t Y;

    // REMOVE THIS LATER.
    assert( WM->getPassingToTime() != -1 );

    VecPosition passto = WM->getPassingToCoord();
    X = passto.getX();
    Y = passto.getY();

    DataTypes::X::bound(&X);
    DataTypes::Y::bound(&Y);

    bool rtn = MyEncoder->add( new PassToCoord(X,Y) );
    if(!rtn)
      cerr <<"[CoachableAgentComm:addPassToCoord]: WARNING: add failed!"<<endl;
    return rtn;
  }

  /* 
   *  true if we executed a passto action within the last 5 cycles
   */
  bool isSatisfied(){
    return (WM->getPassingToTime() != -1 && 
            WM->getPassingToTime() >= WM->getCurrentTime() - 5 );
  }
};


/*
 * WantPass member variables:
 * none
 */
class MyWantPass : public WantPass, public MySayMsgTypes {

 public: 
  void process(){    
    LOGOUT(49) << "MyWantPass: playernum="<<SenderPlayerNum<<" wants pass."<<endl;

    WM->updateWantPassPlayer(SenderPlayerNum);
  }

  bool addToEncoder(SayMsgEncoder *MyEncoder){
    bool rtn = MyEncoder->add( new WantPass() );
    if(!rtn)
      cerr <<"[CoachableAgentComm:addWantPass]: WARNING: add failed!"<<endl;
    return rtn;
  }

  /* 
   * true if we're positioned well to receive a pass
   * Predicates: I'm not ball owner && I'm close to opponent goal 
   *  && I have a good shot on goal
   */
  bool isSatisfied(){

    /* return false if we're already the ball owner */
    if(WM->getBallOwner() == WM->getAgentObjectType())
      return false;

    /* Return false if we're not close to the goal */
#ifndef NEAR_OPPGOAL
#define NEAR_OPPGOAL   25.0
#endif
  
    VecPosition oppgoalPos = WM->getPosOpponentGoal();
    VecPosition agentPos = WM->getAgentGlobalPosition();
  
    if(agentPos.getDistanceTo(oppgoalPos) > NEAR_OPPGOAL)
      return false;


    /*** now, determine if we have a good shot on the goal ****/

#define BLOCK_DIST 3.0

    vector <VecPosition> oppPositions; // positions of the opponent players

    for(ObjectT opp = OBJECT_OPPONENT_1; opp <= OBJECT_OPPONENT_11; opp = (ObjectT)(opp+1)) {
      // For all opponents whose locations are known....
      if(WM->isConfidenceGood(opp)) {
	oppPositions.push_back(WM->getGlobalPosition(opp));
      }
    }

    // try three lines to the goal: a line to the top goal corner, goal center,
    // and bottom goal corner. 

    for(float y = -6.; y <= 6.; y += 3.) {
      oppgoalPos.setY(y);

      /* Make a line between us and the goal */
      Line lineTOgoal = Line::makeLineFromTwoPoints(agentPos, oppgoalPos);

      bool opponentsBlocking = false;

      for(unsigned int i = 0; i < oppPositions.size(); i++) {
    
	// get the distance between the opponent and our shoot line
	double dist = lineTOgoal.getDistanceWithPoint(oppPositions[i]);
	
	if(dist < BLOCK_DIST)
	  opponentsBlocking = true;
      }

      if(!opponentsBlocking)
	return true;
    }

    // didn't find a goal shot line with no opponents blocking
    return false;
  }
};


/*
 * OppPos member variables: 
 * float X - opponent position X value 
 * float Y - opponent position Y value
 * unsigned short cycleno - number of cycles since saw player
 * unsigned short playernum - # of player at pos X,Y
 */
class MyOppPos : public OppPos, public MySayMsgTypes {

 public: 
  void process(){    
    LOGOUT(49) << "MyOppPos: playernum="<<SenderPlayerNum<<" reports opponent #"<<playernum
	 <<" at ("<<X<<","<<Y<<") with confidence "<<cycleno<<" cycles."<<endl;

    WM->updateOpponentPos(playernum, X, Y, cycleno+1);
  }

  /* Helper to addToEncoder. 
   * Returns playernum of a player that we would like to give position info on
   * Returns playernum in range [1,11].
   */
  DataTypes::Playernum_t choosePlayerNum(){

    // for now just return the first player we're confident of his position
    for(int i=0; i<11; i++){
      if( WM->getCurrentTime().getTimeDifference(
	  WM->getTimeLastSeen(SoccerTypes::getOpponentObjectFromIndex(i))) <= 6)
	return i+1;
    }

    cerr <<"[MyOpponentPos]: not confident of any opponent pos! "
	 <<"addToEncoder violating predicate.\n"<<endl;
    return 9;
  }

  bool addToEncoder(SayMsgEncoder *MyEncoder){
    DataTypes::X_t X;
    DataTypes::Y_t Y;
    int cycles;

    DataTypes::Playernum_t playernum = choosePlayerNum();

    assert(playernum >=1 && playernum <= 11);
    ObjectT playerobj = SoccerTypes::getOpponentObjectFromIndex(playernum - 1);

    VecPosition pos = WM->getGlobalPosition(playerobj);
    X = pos.getX();
    Y = pos.getY();

    Time lastseen = WM->getTimeLastSeen( playerobj );
    cycles = WM->getCurrentTime().getTimeDifference(lastseen);

    DataTypes::Playernum::bound(&((int)playernum));
    DataTypes::X::bound(&X);
    DataTypes::Y::bound(&Y);
    DataTypes::Cycleno::bound(&cycles);

    bool rtn = MyEncoder->add( new OppPos(playernum, X,Y,cycles) );
    if(!rtn)
      cerr <<"[CoachableAgentComm:addOppPos]: WARNING: add failed!"<<endl;
    return rtn;
  }

  /* 
   *  true if we're confident of the location of at least one opponent
   */
  bool isSatisfied(){
    for(int i=0; i<11; i++){
      if( WM->getCurrentTime().getTimeDifference(
	  WM->getTimeLastSeen(SoccerTypes::getOpponentObjectFromIndex(i))) <= 6)
	return true;

    }
    return false;
  }
};



/*
 * TeammatePos member variables: 
 * float X - teammate position X value 
 * float Y - teammate position Y value
 * unsigned short cycleno - number of cycles since saw player
 * unsigned short playernum - # of player at pos X,Y
 */
class MyTeammatePos : public TeammatePos, public MySayMsgTypes {

 public: 
  void process(){
    LOGOUT(49) << "MyTeammatePos: playernum="<<SenderPlayerNum<<" reports teammate #"<<playernum
	 <<" at ("<<X<<","<<Y<<") with confidence "<<cycleno<<" cycles."<<endl;
    if(playernum != WM->getPlayerNumber())
      WM->updateTeammatePos(playernum, X, Y, cycleno+1); 
  }

  /* Helper to addToEncoder. 
   * Returns playernum of a player that we would like to give position info on
   * Returns playernum in range [1,11].
   */
  DataTypes::Playernum_t choosePlayerNum(){

    // for now just return the first player we're confident of his position
    for(int i=0; i<11; i++){
      if( WM->getCurrentTime().getTimeDifference(
	  WM->getTimeLastSeen(SoccerTypes::getTeammateObjectFromIndex(i))) <= 6)
	return i+1;
    }

    cerr <<"[MyTeammatePos]: not confident of any teammate pos! "
	 <<"addToEncoder violating predicate.\n"<<endl;
    return 9;
  }

  bool addToEncoder(SayMsgEncoder *MyEncoder){
    DataTypes::X_t X;
    DataTypes::Y_t Y;
    int cycles;

    DataTypes::Playernum_t playernum = choosePlayerNum();
    assert(playernum >=1 && playernum <= 11);
    ObjectT playerobj = SoccerTypes::getTeammateObjectFromIndex(playernum - 1);

    VecPosition pos = WM->getGlobalPosition(playerobj);
    X = pos.getX();
    Y = pos.getY();

    Time lastseen = WM->getTimeLastSeen( playerobj );
    cycles = WM->getCurrentTime().getTimeDifference(lastseen);

    DataTypes::Playernum::bound(&((int)playernum));
    DataTypes::X::bound(&X);
    DataTypes::Y::bound(&Y);
    DataTypes::Cycleno::bound(&cycles);

    bool rtn = MyEncoder->add( new TeammatePos(playernum, X,Y,cycles) );
    if(!rtn)
      cerr <<"[CoachableAgentComm:addTeammatePos]: WARNING: add failed!"<<endl;
    return rtn;
  }

  /* 
   *  true if we're confident of the location of at least one teammate
   */
  bool isSatisfied(){

    for(int i=0; i<11; i++){
      if( WM->getCurrentTime().getTimeDifference(
	  WM->getTimeLastSeen(SoccerTypes::getTeammateObjectFromIndex(i))) <= 6)
	return true;

    }
    return false;
  }
};



#endif
