/* Dongryeol Lee
 *
 * RegionVisitor.h
 * 
 * Header file for RegionVisitor Class
 */

#include "Geometry.h"
#include "visitor.h"
#include "region.h"
#include <stdio.h>
#include <math.h>
#include <vector>
#include "VarBindings.h"


#ifndef REGIONVISITOR
#define REGIONVISITOR

using namespace rcss::clang;
using namespace std;

class RegionVisitor: public Region::Visitor
{
 public:
  /* Define my region enum type */
  enum myRegionType {QUAD, ARC, POINT, TRI, REC};

  RegionVisitor(){ init(); }

  double getRegionArea() const;
  
  double getRegionXMax() const;

  double getRegionXMin() const;
  
  double getRegionYMax() const;

  double getRegionYMin() const;

  void updateBoundary(vector <VecPosition> v);

  // functions added for variable handling: 
  RegionVisitor(const VarBindings& b)
    : myBindings(b)
    {
      init();
    }

  VarBindings getBindings() const {return myBindings;}

  bool hasVariables() const;
  //end


  void visit(RegNull *reg);

  void visit(RegQuad *reg);

  void visit(RegArc *reg);

  void visit(RegUnion *reg);

  void visit(RegNamed *reg);

  void visit(RegPoint *reg);

  void visit(RegTri *reg);

  void visit(RegRec *reg);

  int isPointInsideRegion(const VecPosition& p) const;

  int isPointInsideRegArc(const VecPosition& center, 
			  double startRad, 
			  double endRad, 
			  double startAng, 
			  double spanAng,
			  const VecPosition& p) const;

  int isPointInsideRegTri(const VecPosition& p0, 
			  const VecPosition& p1, 
			  const VecPosition& p2, 
			  const VecPosition& p) const;

  int isPointInsideRegRec(const VecPosition& p0, const VecPosition& p1, 
			  const VecPosition& p) const;
  
  double calculateSignedTriangleArea(const VecPosition& v0, 
				     const VecPosition& v1, 
				     const VecPosition& v2) const;

  VecPosition selectRandomPointInsideArc(const VecPosition &center, 
					 double startRad, 
					 double endRad,
					 double startAng, 
					 double spanAng) const;
  
  VecPosition selectRandomPointInsideTri(const VecPosition &p0, 
					 const VecPosition &p1, 
					 const VecPosition &p2) const;
  
  VecPosition selectRandomPointInsideRec(const VecPosition &p0, 
					 const VecPosition &p1) const;

  VecPosition selectRandomPointInsideRegion() const;
  
  VecPosition selectRandomPointOutsideRegion() const;

  vector <VecPosition> intersectArc(const VecPosition &center, 
				    double startRad, 
				    double endRad, 
				    double startAng, 
				    double spandAng, 
				    const Line &l) const;

  vector <VecPosition> intersectTri(const VecPosition &p0, 
				    const VecPosition &p1, 
				    const VecPosition &p2, 
				    const Line &l) const;

  vector <VecPosition> intersectRec(const VecPosition &p0, 
				    const VecPosition &p1, 
				    const Line &l) const;

  VecPosition getPointClosestToObjectWithinSet(vector <VecPosition> solutions,
					       const VecPosition& objectPos) const;

  VecPosition getPointClosestFromRegArcToObject(const VecPosition& center, 
						double startRad, 
						double endRad, 
						double startAng, 
						double spanAng, 
						const VecPosition& objectPos) const;

  VecPosition getPointClosestFromRegTriToObject(const VecPosition &v0, 
						const VecPosition &v1, 
						const VecPosition &v2, 
						const VecPosition &objectPos) const;

  VecPosition getPointClosestFromRegRecToObject(const VecPosition &v0, 
						const VecPosition &v1, 
						const VecPosition &objectPos) const;

  VecPosition getPointClosestFromRegionToObject(const VecPosition &objectPos) const;

 private:
  double regionArea;   /* Total area of the region visited */

  /* Keep track of the region visited */
  vector <myRegionType> regionsVisited;
  
  /* Keep track of the vertices of the regions */
  vector <VecPosition> points;

  double region_ymax;        /* denotes the approximated maximum value of the y-coordinate */
  
  double region_ymin;        /* approximated minimum value of the y-coordinate */
  
  double region_xmax;        /* approximated maximum value of the x-coordinate */

  double region_xmin;        /* approximated minimum value of the x-coordinate */

  VarBindings myBindings;  /* holds variable bindings for this region */

  bool found_variables; /* flag. set to true if region visitor encounters clang variables */

  void init(); /* called by the RegionVisitor constructors */
			 
};

#endif
