/*
 * A visitor class to handle Rule types. 
 *
 * John Davin
 */


#include "RuleVisitor.h"
#include "CondVisitor.h"
#include "DirVisitor.h"
#include "msgstore.h"
#include "msgrulestore.h"
#include <vector>

//#define IF_RULEDEBUG if(trilearn_wm->getCurrentTime() == 50)
#define IF_RULEDEBUG if(0)

extern MsgRuleStore defrulestorage; // declared in msgsorterdefine.h

// extern the world model from PlayerInterface.C
// needed only for getting the simulation time for the matched actions logging
extern WorldModel *trilearn_wm; 

/*
 * Evaluate the rule's condition, and if it's true, use a DirVisitor
 * to evaluate the directive (which will add actions to the ActionPicker)
 */
void RuleVisitor::visit( SimpleRule *rule){
  Dir *curdir;
  Cond *thecond;

  IF_RULEDEBUG cout << "Visiting SimpleRule: "<<*rule<<endl;
  //PR(4, "[RuleVisitor:SimpleRule]: Got a SimpleRule\n");
  thecond = (Cond*) rule->getCond();

  if(Log.isInLogLevel(4)){ 
    *logger <<endl; rule->print(*logger); *logger <<endl;}

  //PR(4, "[RuleVisitor:SimpleRule]: Eval'ing condition\n");
  // Even if force_cond_true, we do the evaluation so that bindings (maybe) get
  // set up correctly
  CondVisitor condvis(&myBindings);
  thecond->accept(condvis);
  if( !force_cond_true && ! condvis.getEval() ){ // if cond is false, we're done
    PR(4, "[RuleVisitor:SimpleRule]: Cond was false: ");
    if(Log.isInLogLevel(4)){ 
      thecond->print(*logger); *logger << endl;}
    
    IF_RULEDEBUG cout << "cond false: "<<trilearn_wm->getCurrentTime() << ": "<< rulename
		      <<" "<<*rule<<endl;
    return;
    }
  
  PR(4, "[RuleVisitor:SimpleRule]: Cond was true: ");
  if(Log.isInLogLevel(4)){ 
    thecond->print(*logger); *logger << endl;}

  /* if rule logging is turned on, log our rule name or rule definition */
  if(rulelogging){

    if(rulelogging == 1){ // log names
      rulelog << trilearn_wm->getCurrentTime() << ": "<<rulename<<endl; 
    }
    else{ // rulelogging == 2
      rulelog << trilearn_wm->getCurrentTime() << ": "<< *rule<<endl;
    }
    //cout << "rulelog: "<<trilearn_wm->getCurrentTime() << ": "<< rulename
    //	 <<" "<<*rule<<endl;
  }

  /*  PR(8, "[RuleVisitor:SimpleRule]: myBindings are now: \n");
  myBindings.PrintAll();*/

  // the list of Dir's (directives) is in a Storage iterator: 
  SimpleRule::HasManyDirs::Storage dirs = rule->getDirs();

  for( SimpleRule::HasManyDirs::Storage::iterator i = dirs.begin(); 
       i != dirs.end(); i++ ){
    // visit the Dir
    curdir = *i;
    DirVisitor dirvisitor(myBindings, sourceType); 
    curdir->accept(dirvisitor);

  }

}

/*
 * A cond rule has a condition, but no directives. 
 *
 */
void RuleVisitor::visit( CondRule *){

  /*
  PR(7, "[RuleVisitor]: Got a CondRule\n");
  PR(9, "[RuleVisitor:CondRule]: Warning: What is a CondRule for? It doesn't have any directives\n");
  */


}


/*
 * Evaluate the nested rule's condition and if it's true, 
 * recursively visit/evaluate the rules in the rule list. 
 * 
 */
void RuleVisitor::visit( NestedRule *rule){
  Rule *currule;
  Cond *thecond;
  
  IF_RULEDEBUG cout <<"visiting NestedRule: "<<*rule<<endl;
  //PR(7, "[RuleVisitor]: Got a NestedRule\n");

  thecond = (Cond*) rule->getCond();

  //PR(7, "[RuleVisitor:NestedRule]: Eval'ing condition\n");

  // Even if force_cond_true, we do the evaluation so that bindings (maybe) get
  // set up correctly
  CondVisitor condvis(&myBindings);
  thecond->accept(condvis);
  if( !force_cond_true && ! condvis.getEval() ){ // if cond is false, we're done
    IF_RULEDEBUG cout << "NestedRule: cond false in rule: "<<rulename<<": "<<*rule<<endl;
    return;
  }
  // get the bindings from this condition and add them to our own, to be passed down to our nested rules.
  //myBindings.add(condvis.getBindingSet());


  //PR(7, "[RuleVisitor:NestedRule]: Cond evaluated to true, now evaluating the nested rules\n");

  NestedRule::HasManyRules::Storage rules = rule->getRules();

  for( NestedRule::HasManyRules::Storage::iterator i = rules.begin(); 
       i != rules.end(); i++ ){

    currule = *i;

    // visit each rule with a new rule visitor
    // the visitor will add applicable directives to the directive_list
    RuleVisitor rulevisitor(myBindings, sourceType, rulename, rulelogging, force_cond_true);
    currule->accept(rulevisitor);

  }

}


/*
 * A list of rules. 
 * Recursively visit (evaluate) each rule in the list 
 */
void RuleVisitor::visit( IDListRule *rule)
{
  RuleID rulename;
  Rule *currule;
  
  IF_RULEDEBUG cout <<"visiting IDListRule: "<<*rule<<endl;
  //PR(7, "[RuleVisitor]: Got a IDListRule: %x\n", rule);

  RuleIDList rules = rule->getIDList();

  if(rules.all()){ // RuleIDList is empty so it should include *all* rules. 
    // put all rules in our list: 

    for(MsgRuleStore::MsgIterator i = defrulestorage.getIterBegin(); 
	i!=defrulestorage.getIterEnd(); i++){
      currule = (*i).second.ptr;

      if (currule == rule)
        {
          // If we recurse on outself, we'll just end up going forever
          continue;
        }
      
      if (currule == NULL)
        {
          cerr << "[RuleVisitor]: In iterating over all rules, I got a NULL rule!" << endl;
          continue;
        }
      
      // visit each rule with a new rule visitor
      // the visitor will add applicable actions to the action list
      RuleVisitor rulevisitor(myBindings, sourceType, (*i).second.rulename, rulelogging, force_cond_true);
      currule->accept(rulevisitor);
    }

    return; // done - evaluated all rules. 
  }

  // evaluate the rules in the id list


  /*
  const std::vector <Rule *> ruleList = rule->getRuleList();

  if(rules.size() != ruleList.size()){
    cerr << "Warning: IDListRule may not have enough rule pointers set "<<
      rules.size()<<" "<<ruleList.size()<<endl;
  }

  int size = ruleList.size();

  for(int i = 0; i < size; i++) {

    RuleVisitor rulevisitor(myBindings, sourceType, "pair name with rulelist", rulelogging);

    (ruleList[i])->accept(rulevisitor);
  }
  */

  std::list< RuleID >::iterator iter;
  int idx;
  for( iter = rules.begin(), idx = 0; 
       iter != rules.end();
       ++iter, ++idx ) {

    if ((signed)rule->getRuleList().size() <= idx || rule->getRuleList()[idx] == NULL)
      {
        cerr << "Warning: IDListRule lacks or has invalid rule pointer set "
             << idx << " " << rule->getRuleList().size() << endl;
        rulename = *iter;
        RulePair pair;
        bool found = defrulestorage.lookup(rulename, &pair);

        if(!found)   {
          PR(9, "[RuleVisitor]: WARNING!: IDListRule specified an unknown rule name: %s\n", rulename.c_str());
          cerr << "[RuleVisitor]: WARNING!: IDListRule specified an unknown rule name: "
               << rulename << endl;
      
          continue; // or would it be better to just fail (return)?
        }
        
        currule = pair.ptr;
      }
    else
      {
        currule = rule->getRuleList()[idx];
      }
    
    // visit each rule with a new rule visitor
    // the visitor will add applicable actions to the action list
    RuleVisitor rulevisitor(myBindings, sourceType, rulename, rulelogging, force_cond_true);

    currule->accept(rulevisitor);

  }

}

