/*
Copyright (c) 2000-2002, Jelle Kok, University of Amsterdam
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution. 

3. Neither the name of the University of Amsterdam nor the names of its 
contributors may be used to endorse or promote products derived from this 
software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*! \file WorldModel.h
<pre>
<b>File:</b>          WorldModel.h
<b>Project:</b>       Robocup Soccer Simulation Team: UvA Trilearn
<b>Authors:</b>       Jelle Kok
<b>Created:</b>       12/02/2001
<b>Last Revision:</b> $ID$
<b>Contents:</b>      class declarations of WorldModel. This class contains
               methods that give information about the current and future
               state of the world (soccer field).
<hr size=2>
<h2><b>Changes</b></h2>
<b>Date</b>             <b>Author</b>          <b>Comment</b>
12/02/2001       Jelle Kok       Initial version created
</pre>
*/

#ifndef _WORLD_MODEL_
#define _WORLD_MODEL_

#include "Objects.h"        // needed for PlayerObject
#include "PlayerSettings.h" // needed for getPlayerDistTolerance
#include "UvaLogger.h"         // needed for Log
#include "Formations.h"     // needed for getStrategicPosition (prediction)

#include <pthread.h>        // needed for pthread_mutex
#include "clangutil.h"      // for UNum

extern Logger Log;          // defined in Logger.C
class CoachableAgentComm; // forward declaration for setAgentComm()


/*****************************************************************************/
/********************** CLASS WORLDMODEL *************************************/
/*****************************************************************************/

/*! The Class WorlModel contains all the RoboCup information that is available
    on the field. It contains information about the players, ball, flags and
    lines. Furthermore it contains methods to extract useful information.
    The (large amount of) attributes can be separated into different groups:
    - Environmental information: specific information about the soccer server
    - Match information: general information about the current state of a match
    - Object information: all the objects on the soccer field
    - Action information: actions that the agent has performed

    The methods can also be divided into different groups:
    - Retrieval methods: directly retrieving information of objects
    - Update methods: update world based on new sensory information
    - Prediction methods: predict future states based on past perceptions
    - High-Level methods: deriving high-level conclusions from basic world state
*/
class WorldModel
{
private:

  /****************************************************************************/
  /*************************** ATTRIBUTES *************************************/
  /****************************************************************************/

  ////////////////////////// ENVIRONMENTAL INFORMATION /////////////////////////

  ServerSettings *SS;                     /*!< Reference to all server params */
  PlayerSettings *PS;                     /*!< Reference to all client params */
  HeteroPlayerSettings pt[MAX_HETERO_PLAYERS]; /*!< info hetero player types  */
  Formations     *formations;             /*!< Reference to formation used    */
  CoachableAgentComm *AgentComm;          /*!< for processing comm msgs       */

  ////////////////////////// CURRENT MATCH INFORMATION /////////////////////////

  // time information
  Time          timeLastSeeMessage;      /*!< server time of last see msg     */
  Time          timeLastSenseMessage;    /*!< server time of last sense msg   */
  Time          timeLastFullstateMessage;/*!< server time of last fullstate msg*/
  bool          bNewInfo;                /*!< indicates new info from server  */
  Time          timeLastCatch;           /*!< time of last catch by goalie    */
  Time          timeLastRefMessage;      /*!< time of last referee message    */

  // player information
  char          strTeamName[MAX_TEAM_NAME_LENGTH]; /*!< Team name             */
  int           iPlayerNumber;           /*!< player number in soccerserver   */
  rcss::clang::UNum iPlayerUNum;         /*!< the UNum for player number      */
  SideT         sideSide;                /*!< side where the agent started    */

  // match information
  PlayModeT     playMode;                /*!< current play mode in the game   */
  int           iGoalDiff;               /*!< goal difference                 */
  int           ourScore;                /*!< goals scored for our team       */
  int           oppScore;                /*!< goals scored by opponent team   */

  // Ball owner is the last player to have come within a specified range of the ball
  ObjectT BallOwner; // the player id of the current ball owner
  // OBJECT_PLAYER_UNKNOWN if bowner is unknown, OBJECT_TEAMMATE_UNKNOWN if bowner
  // is our team but we don't know which player (same for opponent team). 

  ////////////////////////// OBJECTS ///////////////////////////////////////////

  // dynamic objects
  BallObject    Ball;                    /*!< information of the ball         */
  AgentObject   agentObject;             /*!< information of the agent itself */
  PlayerObject  Teammates[MAX_TEAMMATES];/*!< information of all teammates    */
  PlayerObject  Opponents[MAX_OPPONENTS];/*!< information of all opponents    */
  PlayerObject  UnknownPlayers[MAX_TEAMMATES+MAX_OPPONENTS];
                                        /*!< info unknown players are stored
                                             here til mapped to known player */
  int           iNrUnknownPlayers;       /*!< number of unknown players       */

  // fixed objects
  FixedObject   Flags[MAX_FLAGS];        /*!< all flag information            */
  FixedObject   Lines[MAX_LINES];        /*!< all line information            */

  //////////////////////// OBJECTS FOR FULLSTATE INFORMATION ///////////////////

  BallObject    FullState_Ball;
  PlayerObject  FullState_Teammates[MAX_TEAMMATES];
  PlayerObject  FullState_Opponents[MAX_OPPONENTS];

  ////////////////////////// LOCALIZATION INFORMATION //////////////////////////

  static const int iNrParticlesAgent = 100; /*!<nr of particles used to store
                                                agent position                */
  static const int iNrParticlesBall  = 100; /*!<nr of particles used to store
                                                ball position and velocity    */
  VecPosition   particlesPosAgent[iNrParticlesAgent]; /*!< particles to store
                                                         agent position       */
  VecPosition   particlesPosBall[iNrParticlesBall];   /*!< particles to store
                                                         ball position        */
  VecPosition   particlesVelBall[iNrParticlesBall];   /*!< particles to store
                                                         ball velocity        */
  ////////////////////////// PREVIOUS ACTION INFORMATION ///////////////////////

  // arrays needed to keep track of actually performed actions.
  SoccerCommand queuedCommands[CMD_MAX_COMMANDS];   /*!< all performed commands,
                                                        set by ActHandler     */
  bool          performedCommands[CMD_MAX_COMMANDS];/*!< commands performed in
                                                        last cycle, index is
                                                        CommandT              */
  int           iCommandCounters[CMD_MAX_COMMANDS]; /*!< counters for all
                                                        performed commands    */


  ////////////////////////// VARIOUS ///////////////////////////////////////////

  // attributes only applicable to the coach
  Time          timeCheckBall;           /*!< time bsCheckBall applies to     */
  BallStatusT   bsCheckBall;             /*!< state of the ball               */

  // synchronization
  pthread_mutex_t mutex_newInfo;        /*!< mutex to protect bNewInfo        */
  pthread_cond_t  cond_newInfo;         /*!< cond variable for bNewInfo       */
  pthread_mutex_t mutex_AgentComm;      /*!< mutex to protect AgentComm funct */

  // communication
  char          m_strPlayerMsg[MAX_MSG];/*!< message communicated by player   */
  int           m_timeMsgRecvd;          /*!< cycle contained in message       */ 
  int           m_iMessageSender;       /*!< player who send message          */
  bool          m_msgNotProcessed;      /*!< true when new msg not processed yet*/

  // other
  bool          m_bWasCollision;        /*!< Indicates whether it is collision*/  
  bool          m_bPerformedKick;       /*!< Indicates whether ball was kicked*/  
  
  // Added for agent communication: 
  ObjectT       passingToPlayer;        /*!< Tells who we're passing to       */
  VecPosition   passingToCoord;         /*!< Coordinate we're passing to      */
  Time          timePassingToPlayer;    /*!< time passing to vars were updated*/

  ObjectT       wantPassPlayer;         /*!< teammate that wants a pass       */
  Time          wantPassTime;           /*!< time wantPassPlayer last updated */

  ObjectT       PassToPlayerSender;     /*!< player who is passing to us      */
  Time          PassToPlayerTime;       /*!< time PassToPlayer received       */

  ObjectT       PassToCoordSender;      /*!< player who is passing to PassToCoord*/
  VecPosition   PassToCoord;            /*!< coordinate player is passing to  */
  Time          PassToCoordTime;        /*!< time PassToCoord received        */

public:

  // statistics
  int           iNrHoles;                /*!< nr of holes recorded            */
  int           iNrOpponentsSeen;        /*!< total nr of opponents seen      */
  int           iNrTeammatesSeen;        /*!< total nr of teammates seen      */

  // last received messages
  char          strLastSeeMessage  [MAX_MSG];  /*!< Last see message          */
  char          strLastSenseMessage[MAX_MSG];  /*!< Last sense_body message   */
  char          strLastHearMessage [MAX_MSG];  /*!< Last hear message         */
  char          strLastFullstateMessage[MAX_MSG];/*!< Last fullstate message  */

  /****************************************************************************/
  /*************************** OPERATIONS *************************************/
  /****************************************************************************/

  ////////////////////////// DIRECT RETRIEVAL (WorldModel.C) ///////////////////

private:
  // private methods
  Object*       getObjectPtrFromType       ( ObjectT        o           );
  Object*       getFullstateObjectFromType ( ObjectT        o           );
                
public:         
                
  // get and set methods of attributes in WorldModel itself

  ServerSettings *getServerSettings();
  PlayerSettings *getPlayerSettings() { return PS; }
  void          setAgentComm               (CoachableAgentComm *comm    );

  bool isGoalie() const;

  void          setTimeLastCatch           ( Time           time        );
  int           getTimeSinceLastCatch      (                            );
  bool          setTimeLastRefereeMessage  ( Time           time        );
  Time          getTimeLastRefereeMessage  (                            );
  Time          getCurrentTime             (                            );
  int           getCurrentCycle            (                            );
  bool          isTimeStopped              (                            );
  bool          isLastMessageSee           (                            ) const;
  Time          getTimeLastSeeGlobalMessage(                            ) const;
  bool          setTimeLastSeeGlobalMessage( Time           time        );
  Time          getTimeLastSeeMessage      (                            ) const;
  bool          setTimeLastSeeMessage      ( Time           time        );
  Time          getTimeLastSenseMessage    (                            ) const;
  bool          setTimeLastSenseMessage    ( Time           time        );
  Time          getTimeLastFullstateMessage(                            ) const;
  bool          setTimeLastFullstateMessage( Time           time        );

  int           getPlayerNumber            (                            ) const;
  rcss::clang::UNum getPlayerUNum          (                            ) const;
  bool          setPlayerNumber            ( int            i           );
  SideT         getSide                    (                            ) const;
  bool          setSide                    ( SideT          s           );
  const char*   getTeamName                (                            ) const;
  bool          setTeamName                ( char           *str        );
  PlayModeT     getPlayMode                (                            ) const;
  bool          setPlayMode                ( PlayModeT      pm          );
  int           getGoalDiff                (                            ) const;
  int           addOneToGoalDiff           (                            );
  int           subtractOneFromGoalDiff    (                            );
  int           getOurScore                (                            ) const;
  int           getOppScore                (                            ) const;
  int           getNrOfCommands            ( CommandT       c           ) const;
  bool          setNrOfCommands            ( CommandT       c,
                                             int            i           );
  Time          getTimeCheckBall           (                            ) const;
  bool          setTimeCheckBall           ( Time           time        );
  BallStatusT   getCheckBallStatus         (                            ) const;
  bool          setCheckBallStatus         ( BallStatusT    bs          );

  ObjectT       getBallOwner               (                            ) const;


  // iterate over a specific object set
  ObjectT       iterateObjectStart         ( int            &iIndex,
                                             ObjectSetT     g,
                                             double         dConf = -1.0      );
  ObjectT       iterateObjectNext          ( int            &iIndex,
                                             ObjectSetT     g,
                                             double         dConf = -1.0      );
  void          iterateObjectDone          ( int            &iIndex           );
              
  // get and set methods for agent information
  ObjectT       getAgentObjectType         (                            ) const;
  bool          setAgentObjectType         ( ObjectT        o           );
  AngDeg        getAgentBodyAngleRelToNeck (                            ) const;
  AngDeg        getAgentGlobalNeckAngle    (                            ) const;
  AngDeg        getAgentGlobalBodyAngle    (                            );
  Stamina       getAgentStamina            (                            ) const;
  TiredNessT    getAgentTiredNess          (                            ) const;  
  double        getAgentEffort             (                            ) const;
  VecPosition   getAgentGlobalVelocity     (                            ) const;
  double        getAgentSpeed              (                            ) const;
  VecPosition   getAgentGlobalPosition     (                            ) const;
  ViewAngleT    getAgentViewAngle          (                            ) const;
  ViewQualityT  getAgentViewQuality        (                            ) const;
  double        getAgentViewFrequency      ( ViewAngleT     va = VA_ILLEGAL,
                                             ViewQualityT   vq = VQ_ILLEGAL   );
  int           getFocusPlayerNum          (                            ) const;
  AngDeg        getAgentViewAngleValue     ( ViewAngleT va              ) const;
  AngDeg        getAgentHalfViewAngleValue ( ViewAngleT va              ) const;
              
  // get methods for ball information
  VecPosition   getBallPos                 (                                  );
  double        getBallSpeed               (                                  );
  AngDeg        getBallDirection           (                                  );
  VecPosition   getBallVelocity            (                                  );
              
  // get method for different information about a specific object
  Time          getTimeGlobalPosition      ( ObjectT        o                 );
  VecPosition   getGlobalPosition          ( ObjectT        o                 );
  VecPosition   getFullstatePosition       ( ObjectT        o                 );
  Time          getTimeGlobalVelocity      ( ObjectT        o                 );
  VecPosition   getGlobalVelocity          ( ObjectT        o                 );
  double        getRelativeDistance        ( ObjectT        o                 );
  VecPosition   getRelativePosition        ( ObjectT        o                 );
  AngDeg        getRelativeAngle           ( ObjectT        o,
                                             bool           bWithBody = false );
  Time          getTimeGlobalAngles        ( ObjectT        o                 );
  AngDeg        getGlobalBodyAngle         ( ObjectT        o                 );
  AngDeg        getGlobalNeckAngle         ( ObjectT        o                 );
  AngDeg        getGlobalAngle             ( ObjectT        o                 );
  double        getConfidence              ( ObjectT        o                 );
  bool          isKnownPlayer              ( ObjectT        o                 );
  ObjectT       getOppGoalieType           (                                  );
  ObjectT       getOwnGoalieType           (                                  );
  Time          getTimeLastSeen            ( ObjectT        o                 );
  Time          getTimeChangeInformation   ( ObjectT        o                 );
               
  // set methods for objects
  bool          setIsKnownPlayer           ( ObjectT        o,
                                             bool           isKnownPlayer     );
  bool          setTimeLastSeen            ( ObjectT        o,
                                             Time           time              );

               
  // get method for information about goals
  VecPosition   getPosOpponentGoal         (                                  );
  VecPosition   getPosOwnGoal              (                                  );
  double        getRelDistanceOpponentGoal (                                  );
  AngDeg        getRelAngleOpponentGoal    (                                  );
               
               
  // get methods about fixed specifications (heterogeneous players)
  HeteroPlayerSettings getInfoHeteroPlayer ( int iIndex                       );
               
  // get method about previous commands
  bool          isQueuedActionPerformed    (                                  );
               
  // methodst hat return truth values about current play mode
  bool          isFreeKickUs               (  PlayModeT pm = PM_ILLEGAL       );
  bool          isFreeKickThem             (  PlayModeT pm = PM_ILLEGAL       );
  bool          isCornerKickUs             (  PlayModeT pm = PM_ILLEGAL       );
  bool          isCornerKickThem           (  PlayModeT pm = PM_ILLEGAL       );
  bool          isOffsideUs                (  PlayModeT pm = PM_ILLEGAL       ); 
  bool          isOffsideThem              (  PlayModeT pm = PM_ILLEGAL       ); 
  bool          isKickInUs                 (  PlayModeT pm = PM_ILLEGAL       ); 
  bool          isKickInThem               (  PlayModeT pm = PM_ILLEGAL       ); 
  bool          isFreeKickFaultUs          (  PlayModeT pm = PM_ILLEGAL       ); 
  bool          isFreeKickFaultThem        (  PlayModeT pm = PM_ILLEGAL       ); 
  bool          isKickOffUs                (  PlayModeT pm = PM_ILLEGAL       );
  bool          isKickOffThem              (  PlayModeT pm = PM_ILLEGAL       );
  bool          isBackPassUs               (  PlayModeT pm = PM_ILLEGAL       );
  bool          isBackPassThem             (  PlayModeT pm = PM_ILLEGAL       );
  bool          isGoalKickUs               (  PlayModeT pm = PM_ILLEGAL       );
  bool          isGoalKickThem             (  PlayModeT pm = PM_ILLEGAL       );
  bool          isBeforeKickOff            (  PlayModeT pm = PM_ILLEGAL       ); 
  bool          isDeadBallUs               (  PlayModeT pm = PM_ILLEGAL       ); 
  bool          isDeadBallThem             (  PlayModeT pm = PM_ILLEGAL       ); 

  ////////////////////////// UPDATE METHODS (WorldModelUpdate.C) ////////////////

  // processing new information about the objects (filling World Model).
  void          processSeeGlobalInfo       ( ObjectT        o,
                                             Time           time,
                                             VecPosition    pos,
                                             VecPosition    vel,
                                             AngDeg         angBody,
                                             AngDeg         angNeck           );
  bool          processNewAgentInfo        ( ViewQualityT   vq,
                                             ViewAngleT     va,
                                             double         dStamina,
                                             double         dEffort,
                                             double         dSpeed,
                                             AngDeg         angSpeed,
                                             AngDeg         angHeadAngle,
					     char           focusteam,
					     int            focusplayer       );
 void           processNewObjectInfo       ( ObjectT        o,
                                             Time           time,
                                             double         dDist,
                                             int            iDir,
                                             double         dDistChange,
                                             double         dDirChange,
                                             AngDeg         angRelBodyAng,
                                             AngDeg         angRelNeckAng,
                                             bool           isGoalie,
                                             ObjectT        objMin,
                                             ObjectT        objMax            );
 bool           processPerfectHearInfo     ( ObjectT        o,
                                             VecPosition    pos,
                                             double         dConf,
                                             bool           bIsGoalie=0       );
 bool           processPerfectHearInfoBall ( VecPosition    pos,
                                             VecPosition    vel,
                                             double         dConf             );
 bool           processUnsureHearInfo      ( ObjectT        o,
                                             VecPosition    pos,
                                             double         dConf             );
 bool           processNewHeteroPlayer     ( int            iIndex,
                                             double         dPlayerSpeedMax,
                                             double         dStaminaIncMax,
                                             double         dPlayerDecay,
                                             double         dInertiaMoment,
                                             double         dDashPowerRate,
                                             double         dPlayerSize,
                                             double         dKickableMargin,
                                             double         dKickRand,
                                             double         dExtraStamina,
                                             double         dEffortMax,
                                             double         dEffortMin        );
  void          processCatchedBall         ( RefereeMessageT rm,
                                             Time           time              );
  void          processQueuedCommands      ( SoccerCommand  commands[],
                                             int            iCommands         );
  bool          storePlayerMessage         ( int            iPlayer, 
                                             char           *strMsg, 
                                             int            iCycle            ); 
  bool          processPlayerMessage       (                                  );                                                                                          
               
  // update methods (defined in WorldModelUpdate.C)
  bool          updateAll                  (                                  );

  void          updateBallOwner            (                                  );

               
private:       
  // update with new visual information
  void          processLastSeeMessage      (                                  );
  bool          updateAfterSeeMessage      (                                  );
  bool          updateAfterFullstateMessage(                                  );
  bool          updateAgentObjectAfterSee  (                                  );
  bool          updateDynamicObjectAfterSee( ObjectT        o                 );
               
  // update with prediction for new cycle
  bool          updateAfterSenseMessage    (                                  );
  bool          updateAgentAndBallAfterSense(                                 );
  bool          updateBallAfterKick        ( double         dPower,
                                             AngDeg         ang               );
  bool          updateDynamicObjectForNextCycle( ObjectT    o,
                                             int            iCycle            );
  bool          updateBallForCollision     ( VecPosition    posAgent          );
               
  // update from relative to global and vice versa
  bool          updateRelativeFromGlobal   (                                  );
  bool          updateObjectRelativeFromGlobal( ObjectT     o                 );
               
  // methods to determine agent state from perceived information
  bool          calculateStateAgent        ( VecPosition    *posGlobal,
                                             VecPosition    *velGlobal,
                                             AngDeg         *angGlobal        );
  void          initParticlesAgent         ( AngDeg         angGlobal         );
  void          initParticlesAgent         ( VecPosition    posInitital       );
  int           checkParticlesAgent        ( AngDeg         angGlobalNeck     );
  void          updateParticlesAgent       ( VecPosition    vel,
                                             bool           bAfterSense       );
  VecPosition   averageParticles           ( VecPosition    posArray[],
                                             int            iLength           );
  void          resampleParticlesAgent     ( int            iLeft             );
  bool          calculateStateAgent2       ( VecPosition    *posGlobal,
                                             VecPosition    *velGlobal,
                                             AngDeg         *angGlobal        );
  VecPosition   calculatePosAgentWith2Flags( ObjectT        objFlag1,
                                             ObjectT        objFlag2          );
  AngDeg        calculateAngleAgentWithPos ( VecPosition    pos               );
               
               
  // methods to determine ball state from perceived information
  bool          calculateStateBall         ( VecPosition    *posGlobal,
                                             VecPosition    *velGlobal        );
  void          initParticlesBall         (  VecPosition    posArray[],
                                             VecPosition    velArray[],
                                             int            iLength           );
  void          checkParticlesBall        (  VecPosition    posArray[],
                                             VecPosition    velArray[],
                                             int            iLength,
                                             int            *iNrLeft          );
  void          updateParticlesBall       (  VecPosition    posArray[],
                                             VecPosition    velArray[],
                                             int            iLength,
                                             double         dPower,
                                             AngDeg         ang               );
  void          resampleParticlesBall      ( VecPosition    posArray[],
                                             VecPosition    velArray[],
                                             int            iLength,
                                             int            iLeft             );
  ObjectT       getMaxRangeUnknownPlayer   ( ObjectT        obj, 
                                             char*          strMsg            );
  VecPosition   calculateVelocityDynamicObject( ObjectT     o                 );
  bool          calculateStateBall2        ( VecPosition    *posGlobal,
                                             VecPosition    *velGlobal        );


  // methods to determine player state from perceived information
  bool          calculateStatePlayer       ( ObjectT        o,
                                             VecPosition    *posGlobal,
                                             VecPosition    *velGlobal        );

  // methods to calculate noise range (quantized noise)
  bool          getMinMaxDistQuantizeValue ( double         dInput,
                                             double         *dMin,
                                             double         *dMax,
                                             double         x1,
                                             double         x2                );
  bool          getMinMaxDirChange         ( double         dOutput,
                                             double         *dMin,
                                             double         *dMax,
                                             double         x1                );
  bool          getMinMaxDistChange        ( double         dOutput,
                                             double         dDist,
                                             double         *dMin,
                                             double         *dMax,
                                             double         x1,
                                             double         xDist1,
                                             double         xDist2            );
  double        invQuantizeMin             ( double         dOutput,
                                             double         dQuantizeStep     );
  double        invQuantizeMax             ( double         dOutput,
                                             double         dQuantizeStep     );


public:
  // various update methods
  void          mapUnknownPlayers          ( Time           time              );
  bool          updateSSToHeteroPlayerType ( int            i                 );
  bool          resetTimeObjects           (                                  );
  void          removeGhosts               (                                  );


  ////////////////////////// PREDICTIONS (WorldModelPredict.C) /////////////////

  // predictions of worldmodel about future states of different objects
  bool          predictStateAfterCommand   ( SoccerCommand  com,
                                             VecPosition    *pos,
                                             VecPosition    *vel,
                                             AngDeg         *angGlobalBody,
                                             AngDeg         *angGlobalNeck,
                                             Stamina        *sta = NULL       );
  bool          predictAgentStateAfterCommand(SoccerCommand com,
                                             VecPosition    *pos,
                                             VecPosition    *vel,
                                             AngDeg         *angBody,
                                             AngDeg         *angNeck,
                                             Stamina        *sta              );
  VecPosition   predictAgentPosAfterCommand( SoccerCommand  com               );
  void          predictStateAfterDash      ( double         dActualPower,
                                             VecPosition    *pos,
                                             VecPosition    *vel,
                                             Stamina        *sta,
                                             double         dDirection        );
  void          predictStateAfterTurn      ( AngDeg         dSendAngle,
                                             VecPosition    *pos,
                                             VecPosition    *vel,
                                             AngDeg         *angBody,
                                             AngDeg         *angNeck,
                                             Stamina        *sta = NULL       );
  VecPosition   predictPosAfterNrCycles    ( ObjectT        o,
                                             int            iCycles,
                                             int            iDashPower = 100,
                                             VecPosition    *vel = NULL       );
  VecPosition   predictAgentPos            ( int            iCycles,
                                             int            iDashPower = 0    );
  int           predictNrCyclesToPoint     ( ObjectT        o,
                                             VecPosition    posTo,
                                             AngDeg         ang               );
  int           predictNrCyclesToObject    ( ObjectT        objFrom,
                                             ObjectT        objTo             );

  int           predictCyclesForBallTravel ( double dist );
  void          predictStaminaAfterDash    ( double         dPower,
                                             Stamina        *sta              );
  bool          isCollisionAfterDash       ( SoccerCommand  soc               );  


  ////////////////////////// HIGH-LEVEL METHODS (WorldModelHighLevel.C) ////////

  // methods that return the number of players in a certain area of the field
  int           getNrInSetInRectangle      ( ObjectSetT     objectSet,
                                             Rectangle      *rect = NULL      );
  int           getNrInSetInCircle         ( ObjectSetT     objectSet,
                                             Circle         c                 );
  int           getNrInSetInCone           ( ObjectSetT     objectSet,
                                             double         dWidth,
                                             VecPosition    start,
                                             VecPosition    end               );

  // method that return the closest or fastest player to a certain pos or object
  ObjectT       getClosestInSetTo          ( ObjectSetT     objectSet,
                                             ObjectT        o ,
                                             double         *dDist = NULL,
                                             double         dConfThr = -1.0   );
  ObjectT       getClosestInSetTo          ( ObjectSetT     objectSet,
                                             VecPosition    pos,
                                             double         *dDist = NULL,
                                             double         dConfThr = -1.0   );
  ObjectT       getClosestInSetTo          ( ObjectSetT     objectSet,
                                             Line           l,
                                             VecPosition    pos1,
                                             VecPosition    pos2,
                                             double         *dDistToLine = NULL,
                                             double         *dDistPos1ToP= NULL);
  ObjectT       getClosestRelativeInSet    ( ObjectSetT     set,
                                             double         *dDist = NULL     );
  ObjectT       getSecondClosestInSetTo    ( ObjectSetT     objectSet,
                                             ObjectT        o,
                                             double         *dDist = NULL,
                                             double         dConfThr = -1.0   );
  ObjectT       getSecondClosestRelativeInSet( ObjectSetT   set,
                                             double         *dDist = NULL     );
  ObjectT       getFastestInSetTo          ( ObjectSetT     objectSet,
                                             ObjectT        o,
                                             int            *iCycles = NULL   );
  ObjectT       getFastestInSetTo          ( ObjectSetT     objectSet,
                                             VecPosition    pos,
                                             VecPosition    vel,
                                             double         dDecay,
                                             int            *iCycles = NULL   );
  ObjectT       getFurthestInSetTo         ( ObjectSetT     objectSet,
                                             ObjectT        o ,
                                             double         *dDist = NULL,
                                             double         dConfThr = -1.0   );
  ObjectT       getFurthestRelativeInSet   ( ObjectSetT     set,
                                             double         *dDist = NULL     );
  double        getMaxTraveledDistance     ( ObjectT        o                 );

  // various methods dealing with sets of objects
  ObjectT       getFirstEmptySpotInSet     ( ObjectSetT     objectSet,
                                             int            iUnknownPlayer=-1 );

  // boolean methods that indicate whether an object fulfills a  constraint
  bool          isVisible                  ( ObjectT        o                 );
  bool          isBallKickable             (                                  );
  bool          isBallCatchable            (                                  );
  bool          isBallHeadingToGoal        (                                  );
  bool          isBallInOurPossesion       (                                  );
  bool          isBallInOwnPenaltyArea     (                                  );
  bool          isInOwnPenaltyArea         ( VecPosition    pos               );
  bool          isInTheirPenaltyArea       ( VecPosition    pos               );
  bool          isConfidenceGood           ( ObjectT                          );
  bool          isConfidenceVeryGood       ( ObjectT                          );
  bool          isOnside                   ( ObjectT                          );
  bool          isOpponentAtAngle          ( AngDeg         ang,
                                             double         dDist             );
             
  // various methods that supply specific information about the field
  Time          getTimeFromConfidence      ( double         dConf             );
  double        getOffsideX                ( bool           bIncludeComm= true);
  VecPosition   getOurOffsideX             ( bool           bIncludeComm= true);

  VecPosition   getOuterPositionInField    ( VecPosition    pos,
                                             AngDeg         ang,
                                             double         dDist = 3.0,
                                             bool           bWithPenalty = true);
  AngDeg        getDirectionOfWidestAngle  ( VecPosition    posOrg,
                                             AngDeg         angMin,
                                             AngDeg         angMax,
                                             AngDeg         *ang,
                                             double         dDist             );
  // strategic positioning
  VecPosition   getStrategicPosition      ( int            iPlayer = -1       );  
  
  // methods  for computing the actual argument for a soccer command.
  double        getActualKickPowerRate     (                                  );
  double        getKickPowerForSpeed       ( double         dDesiredSpeed     );
  double        getKickSpeedToTravel       ( double         dDistance,
                                             double         dEndSpeed         );
  double        getFirstSpeedFromEndSpeed  ( double         dEndSpeed,
                                             double         dCycles           );
  double        getFirstSpeedFromDist      ( double         dDist,
                                             double         dCycles           );
  double        getEndSpeedFromFirstSpeed  ( double         dFirstSpeed,
                                             double         dCycles           );
  AngDeg        getAngleForTurn            ( AngDeg         angDesiredAngle,
                                             double         dSpeed            );
  AngDeg        getActualTurnAngle         ( AngDeg         angActualAngle,
                                             double         dSpeed            );
  double        getPowerForDash            ( VecPosition    posRelTo,
                                             AngDeg         angBody,
                                             VecPosition    vel,
                                             double         dEffort           );


  ////////////////////////// VARIOUS (WorldModel.C) ////////////////////////////

  // constructor
  WorldModel                               ( ServerSettings *ss,
                                             PlayerSettings *ps,
                                             Formations     *fs               );
 
  // print information about WorldModel
  void          show                       ( ostream        &os = cout        );
  void          show                       ( ObjectSetT     set,
                                             ostream        &os = cout        );
  void          showQueuedCommands         ( ostream        &os = cout        );
  void          show                       ( ObjectT        o,
                                            ostream        &os = cout        );

  // methods that deal with timing information
  bool          waitForNewInformation      (                                  );

  // methods that deal with debugging
  void          logObjectInformation       ( int            iLogLevel,
                                             ObjectT        o                 );

  bool          isInBoundsX(VecPosition X);
  bool          isInBoundsY(VecPosition X);


  //////// Coachable Agent Comm update functions (WorldModelCommUpdate.C) ////////////

  bool updateTeammatePos(int playernum, float X, float Y, unsigned short cycles);

  bool updateOpponentPos(int playernum, float X, float Y, unsigned short cycles);

  // helper function to the above two functions
  bool updatePlayerPos(int playernum, float X, float Y, unsigned short cycles, TeamT team);

  bool updateBallPos(float X, float Y, unsigned short cycles);

  bool updateBallVel(float dX, float dY, unsigned short cycles);

  bool updateBallOwner(unsigned short playernum, TeamT team);

  bool updateWantPassPlayer(unsigned short playernum);

  ObjectT getWantPassPlayer();
  Time getWantPassTime();

  // methods for setting/accessing info on who is passing to us
  bool        updatePassToPlayer(int passernum);
  ObjectT     getPassToPlayerSender() const;
  Time        getPassToPlayerTime() const;

  bool        updatePassToCoord(int passernum, float X, float Y);
  VecPosition getPassToCoord() const;
  ObjectT     getPassToCoordSender() const;
  Time        getPassToCoordTime() const;



  ////////// Functions to aid agent communication (in WorldModel.C) ///////////////

  /**** These methods access/set the variables that tell who we are passing to ****/
  void           setPassingToPlayer(ObjectT player, VecPosition coord);

  unsigned short getPassingToPlayer() const;

  VecPosition    getPassingToCoord() const;

  Time           getPassingToTime() const;

  int            getPassCount();

};

#endif
