/*
 * An implementation of the WorldModelInterface abstract functions. 
 * This class accesses the trilearn world model state. 
 *
 * John Davin
 */

#ifndef WORLD_MODEL_ACCESSOR_H
#define WORLD_MODEL_ACCESSOR_H

#include <iostream>
#include "Geometry.h"
#include "SoccerTypes.h"
#include "WorldModelInterface.h"
#include "WorldModel.h" // the trilearn WorldModel


extern WorldModel *trilearn_wm; // this comes from PlayerInterface.C


class WorldModelAccessor : public WorldModelInterface {

 public:

  VecPosition getGlobalPosition(rcss::clang::UNum::unum_t num, bool ourside);

  rcss::clang::PlayMode getPlayMode();

  rcss::geom::Vector2D getBallPos(){
    return trilearn_wm->getBallPos().convertToVector2D();

  }

  /* THe player ALWAYS uses left hand coordinates, even on the right side */
  bool useRightSideCoordinates() { return false; }

  /*
   * return true if confidence of ball location is good
   */
  bool ballConfidenceGood(){
    return trilearn_wm->isConfidenceGood(OBJECT_BALL);
  }

  /* Returns true if the player with the Unum's position is confident
     by the agent */
  bool playerConfidenceGood(rcss::clang::UNum::unum_t playerUNum, 
			    bool ourSide); 


  /* Returns true if the player with the UNum's position is visible by
     the agent */
  bool playerVisibilityGood(rcss::clang::UNum::unum_t playerUNum,
			    bool ourSide);


  /* Returns true if the ball is kickable by the agent */
  bool isBallKickable() {
    return trilearn_wm->isBallKickable();
  }

  /* Returns true if the ball is kickable by the player specified by the unum
     and its team */
  bool isBallKickableByPlayer(rcss::clang::UNum::unum_t u, bool ourSide) {
    
    /* First check the ball confidence, and if it's low, then no judgment
       can be made regarding this */
    if( !ballConfidenceGood())
      return false;

    /* Next, check the confidence value of the opponent position, and if it's
       not good, then don't proceed */
    if( !playerConfidenceGood(u, ourSide))
      return false;


    /* Get the player position */
    VecPosition playerPos = getGlobalPosition(u, ourSide);
    
    /* Get the ball location */
    VecPosition ballPos = getBallPos();

    /* Just determine player-to-ball distance < threshold */
    return playerPos.getDistanceTo(ballPos)<(trilearn_wm->getServerSettings())
      ->getMaximalKickDist();
  }
    
  VecPosition getPosOpponentGoal() {
    return trilearn_wm->getPosOpponentGoal();
  }

  VecPosition getPosOwnGoal() {
    return trilearn_wm->getPosOwnGoal();
  }

  VecPosition getPosOpponentGoalie() {
    return trilearn_wm->getGlobalPosition(OBJECT_OPPONENT_GOALIE);
  }

  VecPosition getPosOurGoalie() {
    return trilearn_wm->getGlobalPosition(OBJECT_TEAMMATE_GOALIE);
  }

  rcss::clang::UNum::unum_t getUNumFromObjectT(ObjectT trilearnobj);
  int getCurrentTime(){
    return (trilearn_wm->getCurrentTime()).getTime();
  }
  
  Time getTimeLastSeeMessage() {
    return trilearn_wm->getTimeLastSeeMessage();
  }
    
  bool TeamIsLeftSide(){

    return (trilearn_wm->getSide()==SIDE_LEFT ? true : false);
  }


  /*
   * Effect: Place a unum_t into player corresponding to the last player to touch the ball.
   * Return: true if the ball owner is on our team, false if opponent team.
   */
  bool getBallOwner(rcss::clang::UNum::unum_t *player);

  int getOurGoal(){

    return trilearn_wm->getOurScore();
  }

  int getOppGoal(){

    return trilearn_wm->getOppScore();
  }

  /*
   * This returns ourgoal - oppgoal. So positive means you're winning, negative losing.
   */
  int getGoalDiff(){
    return trilearn_wm->getGoalDiff();

  }

  rcss::clang::UNum getMyNum(){
    return trilearn_wm->getPlayerUNum();

  }

  /* Get agent view angle */
  ViewAngleT getAgentViewAngle() {
    return trilearn_wm->getAgentViewAngle();
  }

  double getAgentSpeed() {
    return trilearn_wm->getAgentSpeed();
  }

  double getAgentGlobalBodyAngle() {
    return trilearn_wm->getAgentGlobalBodyAngle();
  }

  VecPosition getAgentGlobalPosition() {
    return trilearn_wm->getAgentGlobalPosition();
  }

  AngDeg getAngleForTurn(AngDeg angDesiredAngle, double dSpeed) {
    return trilearn_wm->getAngleForTurn(angDesiredAngle, dSpeed);
  }

  double getBallSpeed() {
    return trilearn_wm->getBallSpeed();
  }


  /* Tells whether the agent is a goalie or not */
  bool isGoalie() {
    return trilearn_wm->isGoalie();
  }

  VecPosition predictBallPosAfterNrCycles(int ncycles) {
    return trilearn_wm->predictPosAfterNrCycles(OBJECT_BALL, ncycles);
  }

  ObjectT getObjectTFromUNum(rcss::clang::UNum::unum_t, bool ourside);

  double getOffsideX() {
    return trilearn_wm->getOffsideX();
  }

  bool updateAll() {
    return trilearn_wm->updateAll();
  }

  VecPosition predictPosAfterNrCycles(rcss::clang::UNum::unum_t u,
				      bool ourSide,
				      int iCycles);

  int predictNrCyclesToObject( ObjectT objFrom, ObjectT objTo ){
    return trilearn_wm->predictNrCyclesToObject( objFrom, objTo );
  }


  Stamina getAgentStamina( ) const;

  TiredNessT getAgentTiredNess( ) const;
  
  bool matchGoalie(const rcss::clang::UNum::unum_t &u, bool ourSide);

  bool FastestToBall() const;

};


#endif // WORLD_MODEL_ACCESSOR_H

