/*
Copyright (c) 2000-2002, Jelle Kok, University of Amsterdam
All rights reserved.

Redistribution and use in source and binary forms, with or without 
modification, are permitted provided that the following conditions are met:

1. Redistributions of source code must retain the above copyright notice, this 
list of conditions and the following disclaimer. 

2. Redistributions in binary form must reproduce the above copyright notice, 
this list of conditions and the following disclaimer in the documentation 
and/or other materials provided with the distribution. 

3. Neither the name of the University of Amsterdam nor the names of its 
contributors may be used to endorse or promote products derived from this 
software without specific prior written permission. 

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" 
AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE 
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE 
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE 
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR 
SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER 
CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, 
OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE 
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/
/*! \file WorldModelUpdate.C
<pre>
<b>File:</b>          WorldModelUpdate.C
<b>Project:</b>       Robocup Soccer Simulation Team: UvA Trilearn
<b>Authors:</b>       Jelle Kok
<b>Created:</b>       12/02/2001
<b>Last Revision:</b> $ID$
<b>Contents:</b>      class definitions of WorldModel. This class contains
               methods that process and update the information in the world
               model.
<hr size=2>
<h2><b>Changes</b></h2>
<b>Date</b>             <b>Author</b>          <b>Comment</b>
12/02/2001       Jelle Kok       Initial version created
</pre>
*/

#include "WorldModel.h"
#include "CoachableAgentComm.h" // for processPlayerMessage()
#include "Parse.h"
#include <stdio.h>      // needed for sprintf
#include <list>         // needed for list

using namespace std;

/******************************************************************************/
/********************** CLASS WORLDMODEL **************************************/
/******************************************************************************/


/*! This method is called when new visual information about objects on
    the field is received from the see_global message (only the coach
    receives these messages). This method updates the information of the
    Object that corresponds to the ObjectT that is supplied as the first
    argument.
    \param o objectType that should be updated
    \param time time of the corresponding values
    \param pos global position of this object
    \param vel global velocity of this object
    \param angBody body angle of this object
    \param angNeck neck angle of this object */
void WorldModel::processSeeGlobalInfo( ObjectT o, Time time,
      VecPosition pos, VecPosition vel, AngDeg angBody, AngDeg angNeck)
{
  DynamicObject * dobj;
  PlayerObject * pobj;

  if( o == OBJECT_ILLEGAL )
    return;
  if( SoccerTypes::isPlayer( o ) )
  {
    pobj = (PlayerObject*)getObjectPtrFromType( o );
    pobj->setTimeLastSeen( time );
    pobj->setGlobalPosition( pos, time );
    pobj->setGlobalVelocity( vel, time );
    pobj->setGlobalBodyAngle( angBody, time );
    pobj->setGlobalNeckAngle( VecPosition::normalizeAngle(angBody+angNeck),
                                                                     time );
    pobj->setIsKnownPlayer( true );

    //    if(pobj->getType() == OBJECT_OPPONENT_1 && pos.getX() < 40){
    //  cerr << "Setting opponent1 to pos "<<pos<<endl<<endl;
    //}


  }
  else if( SoccerTypes::isBall( o ) )
  {
    dobj = (DynamicObject*)getObjectPtrFromType( o );
    dobj->setTimeLastSeen( time );
    dobj->setGlobalPosition( pos, time );
    dobj->setGlobalVelocity( vel, time );
  }
}

/*! This method is called when new visual information about the agent
    is received. It updates the information of the AgentObject stored in
    the WorldModel.
    \param vq new ViewQuality for the agent
    \param va new ViewAngle for the agent
    \param dStamina new stamina for the agent
    \param dEffort new effort for the agent
    \param dSpeed magnitude of the velocity
    \param angSpeed angle of the speed velocity (relative to neck)
    \param angHeadAngle new relative angle between body and neck
    \return bool indicating whether update was succesful */
bool WorldModel::processNewAgentInfo( ViewQualityT vq, ViewAngleT va,
     double dStamina, double dEffort, double dSpeed, AngDeg angSpeed,
     AngDeg angHeadAngle, char focusteam, int focusplayer )
{
  Stamina sta = agentObject.getStamina();

  sta.setStamina                   ( dStamina                              );
  sta.setEffort                    ( dEffort                               );
  agentObject.setStamina           ( sta                                   );
  agentObject.setViewQuality       ( vq                                    );
  agentObject.setViewAngle         ( va                                    );
  agentObject.setSpeedRelToNeck    ( VecPosition( dSpeed, angSpeed, POLAR) );
  agentObject.setBodyAngleRelToNeck( angHeadAngle                          );

  if(focusplayer == -1) // don't need to set focusPlayer. 
    return true;
  if(focusplayer == 0){ // attentionto is off
    agentObject.setFocusPlayer( OBJECT_ILLEGAL );
    return true;
  }

  if( getSide() == SIDE_LEFT && focusteam != 'l'  || 
      getSide() == SIDE_RIGHT && focusteam != 'r'){
    cerr << "[WorldModel]: we only support focusing on our teammates!"<<endl;
    if(getSide() == SIDE_LEFT)
      cout << "get side is left and focusteam="<<focusteam<<endl;
    return false;
  }

  agentObject.setFocusPlayer( SoccerTypes::getTeammateObjectFromIndex(focusplayer-1) );

  return true;
}

/*! This method is called when new visual information about other objects on
    the field is received. It updates the information of the Object that
    corresponds to the ObjectT that is supplied as the first argument. If
    some of the arguments are not received with the visual information, they
    are passed with the value 'UnknownDoubleValue'. Note that the objects are
    only updated with the (relative) information passed as arguments. To make
    sure all global information in the objects is synchronized with the relative
    information, the method updateAll() has to be called after all (relative)
    object information is updated.
    \param o objectType that should be updated
    \param time time of the corresponding values
    \param dDist distance from agent to this object o
    \param iDir angle from agent with this object o (relative to neck)
    \param dDistChange distance change between agent and this object o
    \param dDirChange angle change between agent and this object o
    \param angRelBodyAng angle between neck of agent and body of object o
    \param angRelNeckAng angle between neck of agent and neck of object o
    \param isGoalie bool indicating whether object is a goalie
    \return bool indicating whether update was succesful */
void WorldModel::processNewObjectInfo( ObjectT o, Time time,
      double dDist, int iDir, double dDistChange, double dDirChange,
      AngDeg angRelBodyAng,   AngDeg angRelNeckAng, bool isGoalie,
      ObjectT objMin, ObjectT objMax )
{
  if( dDist == UnknownDoubleValue || o == OBJECT_ILLEGAL )
    return; // no sense to update when only angle is known.

  if( SoccerTypes::isFlag( o ) )
  {
    Flags[SoccerTypes::getIndex(o)].setRelativePosition(dDist,(double)iDir,time);
    Flags[SoccerTypes::getIndex(o)].setTimeLastSeen    ( time                );
    Flags[SoccerTypes::getIndex(o)].setType            ( o                   );
  }
  else if( SoccerTypes::isPlayer( o ) || SoccerTypes::isBall( o ) )
  {
    DynamicObject *d = NULL;

    // if we do not have all information, update UnknownPlayer array
    if( !( SoccerTypes::isKnownPlayer( o ) || SoccerTypes::isBall( o ) ) )
    {
      UnknownPlayers[iNrUnknownPlayers].setIsKnownPlayer( false );
      UnknownPlayers[iNrUnknownPlayers].setPossibleRange( objMin, objMax );
      d = &UnknownPlayers[iNrUnknownPlayers];
      iNrUnknownPlayers++;
      Log.log( 459, "range %f: %d (%d)-%d (%d)",
         dDist, objMin, SoccerTypes::getIndex(objMin)+1,
         objMax, SoccerTypes::getIndex(objMax)+1 );
    }
    else // else update the known object (teammate, opponent, ball)
    {
      d = (DynamicObject*)getObjectPtrFromType( o );
      if( SoccerTypes::isPlayer( o ) )
        ((PlayerObject*)d)->setIsKnownPlayer( true );
    }

    if( d != NULL )  // if object was known
    {
      // set all values for this dynamicobject
      d->setRelativePosition( dDist, (double)iDir, time );
      if( dDistChange != UnknownDoubleValue )
        d->setRelativeDistanceChange( dDistChange, time );
      if( dDirChange != UnknownDoubleValue )
        d->setRelativeAngleChange( dDirChange, time );
      if( angRelBodyAng != UnknownAngleValue )
        ((PlayerObject*)d)->setRelativeBodyAngle( angRelBodyAng, time );
      if( angRelNeckAng != UnknownAngleValue )
        ((PlayerObject*)d)->setRelativeNeckAngle( angRelNeckAng, time );
      if( isGoalie == true && SoccerTypes::isPlayer( o ))
        ((PlayerObject*)d)->setIsGoalie( true );
      else if( SoccerTypes::isPlayer( o ))
        ((PlayerObject*)d)->setIsGoalie( false );
      d->setType( o );
      d->setTimeLastSeen( time );

      // check if there wasn't an unknown player located in the worldmodel
      // that corresponded to the same player
      if( SoccerTypes::isPlayer( o ) && SoccerTypes::isKnownPlayer( o )  )
      {
        int        iIndex;
        ObjectT    objMin = OBJECT_ILLEGAL;
        double     dMinDist = 1000.0, dTmp;
        ObjectSetT set = SoccerTypes::isOpponent( o )
                             ? OBJECT_SET_OPPONENTS
                             : OBJECT_SET_TEAMMATES;

        for( ObjectT obj = iterateObjectStart( iIndex, set );
             obj != OBJECT_ILLEGAL;
             obj = iterateObjectNext ( iIndex, set ) )
        {
          if( obj != getAgentObjectType() && isKnownPlayer( obj ) == false )
          {
            dTmp=(getRelativePosition(obj)-d->getRelativePosition()).getMagnitude();
            if( dTmp < dMinDist )
            {
              objMin   = obj;
              dMinDist = dTmp;
            }
          }
        }
        iterateObjectDone( iIndex );
        if( objMin != OBJECT_ILLEGAL &&
            dMinDist < PS->getPlayerDistTolerance() )
        {
          PlayerObject *pob = (PlayerObject*) getObjectPtrFromType( objMin );
          pob->setTimeLastSeen( -1 ); // delete the unknown player
        }
      }
    }
  }
  else if( SoccerTypes::isLine( o ) )
  {
    // angle returned is angle of neck angle with line, convert to angle
    // of neck with orthogonal to line
    iDir = ( iDir < 0 ) ? (90 + iDir ) : - (90 - iDir );
    Lines[SoccerTypes::getIndex(o)].setRelativePosition(dDist,(double)iDir,time);
    Lines[SoccerTypes::getIndex(o)].setTimeLastSeen( time );
    Lines[SoccerTypes::getIndex(o)].setType( o );
  }
}

/*! This method stores a message communicated by an other player in the world
    model. It is not processed yet. This is done by the updateAll method.
    \param iPlayer player who communicated
    \param strMsg string that was communicated
    \param iDiff difference between cycles message was communicated and current
    time
    \return bool indicating whether all values were stored successfully. */
bool WorldModel::storePlayerMessage( int iPlayer, char *strMsg, int iCycle )
{
  strcpy( m_strPlayerMsg, strMsg );
  m_timeMsgRecvd = iCycle;
  m_iMessageSender = iPlayer;
  m_msgNotProcessed = true;
  return true;
}


/*! This method processes a communication message from a teammate. */
/* NOTE: This function can be called by the sense thread in ActHandler::sendCommands().
 * Therefore, we must protect the AgentComm function call with a mutex to insure that 
 * we do not have both threads in that function at the same time
 */
bool WorldModel::processPlayerMessage( ){

  if( m_msgNotProcessed == true ){
    m_msgNotProcessed = false;

    pthread_mutex_lock  ( &mutex_AgentComm );
    AgentComm->handleReceivedMsg(m_strPlayerMsg, m_iMessageSender);
    pthread_mutex_unlock( &mutex_AgentComm );

    return true;
  }
  else
    return false;
}

/*! This method processes a communication message from a teammate. */
/* OLD TRILEARN FUNCTION
bool WorldModel::processPlayerMessage( )
{
  static char strMessage[MAX_MSG];                       // location for message
  int         iDiff = getCurrentCycle() - m_timeMsgRecvd; // time difference
  double      dDiff = (double)iDiff/100.0;               // conf difference
  char        *strMsg;
  strcpy( strMessage, m_strPlayerMsg );
  strMsg = strMessage;                                   // pointer to work with

  // translate ball position and velocity back to initial range.
  double dBallX =    (int)(strMsg[0]-'0')*10 + (int)(strMsg[1] - '0' ) - 52.5;
  double dBallY =    (int)(strMsg[2]-'0')*10 + (int)(strMsg[3] - '0' ) - 34.0;
  double dBallVelX = (int)(strMsg[4]-'0')    + (int)(strMsg[5] - '0' )/10.0 - 2.7;
  double dBallVelY = (int)(strMsg[6]-'0')    + (int)(strMsg[7] - '0' )/10.0 - 2.7;

  VecPosition pos( dBallX, dBallY );
  VecPosition vel( dBallVelX, dBallVelY );

  // if ball not seen or felt for three cycles, update ball information
  if( getTimeChangeInformation( OBJECT_BALL ) < getCurrentTime() - 3  &&
      getRelativeDistance( OBJECT_BALL ) > SS->getVisibleDistance() )
   processPerfectHearInfoBall( pos, vel, 1.00 - dDiff )    ;

  return true;
}
*/

/*! This method is called when new information about the ball is heard.
    It updates the information of the ball only when the confidence is higher
    than the information already available in the WorldModel.
    \param posGlobal global position of the ball
    \param vel global velocity of the ball
    \param dConf confidence in the ball information
    \return true when information is updated, false otherwise */
bool WorldModel::processPerfectHearInfoBall( VecPosition posGlobal,
                             VecPosition vel, double dConf )
{
  if( Ball.getConfidence( getCurrentTime() ) < dConf )
  {
    Time time = getTimeFromConfidence( dConf );
    Ball.setGlobalPosition( posGlobal, time );
    Ball.setGlobalVelocity( vel,       time );
    Ball.setTimeLastSeen  (            time );
    updateObjectRelativeFromGlobal( OBJECT_BALL );
    return true;
  }
  return false;
}

/*! This method is called when new information about an object is heard. But
    only when the player who said the message was completely sure about the
    object type (it would not give information about a player it does not know
    the player number of, then it would call processUnsureHearInfo).
    It updates the information of the specified object only when the confidence
    is higher than the information already available in the WorldModel.
    \param o object type to which this information relates to
    \param posGlobal global position of this object type.
    \param dConf confidence in the object information.
    \param bIsGoalie boolean indicating whether 'o' is a goalie
    \return true when information is updated, false otherwise */
bool WorldModel::processPerfectHearInfo( ObjectT o, VecPosition posGlobal,
                                         double dConf, bool bIsGoalie )
{
  if( SoccerTypes::isBall( o ) || o == getAgentObjectType() )
    return false; // ball should be called with processPerfectHearInfoBall
  else if( !SoccerTypes::isKnownPlayer( o ) )
    return processUnsureHearInfo( o, posGlobal, dConf );

  PlayerObject *object = (PlayerObject *)getObjectPtrFromType( o );
  if( object == NULL )
    return false;

  Time time = getTimeFromConfidence( dConf ) ;

  // if we are not sure about the exact player number of this player in
  // the world model (getIsKnownPlayer() == false) we overwrite the
  // information of this player since the player who said this information
  // is sure about it (otherwise processUnsureHearInfo would be called instead
  // of processPERFECTHearInfo)
  if( object->getConfidence( getCurrentTime() ) < dConf ||
      object->getIsKnownPlayer() == false  )
  {
    object->setGlobalPosition     ( posGlobal         , time );
    object->setTimeLastSeen       ( time                     );
    object->setGlobalVelocity     ( VecPosition( 0, 0), time );
    object->setIsKnownPlayer      ( true                     );
    object->setIsGoalie           ( bIsGoalie                );
    updateObjectRelativeFromGlobal( o                        );
    return true;
  }
  return false;
}

/*! This method is called when new information about an object is heard. But
    only when the player who said the message was not sure about the
    object type. It is tried to map the given information to an object already
    in the WorldModel. If we cannot find such a player, we add the player info
    at the position of the first player who we don't have information about.
    The information of the specified object is only updated when the confidence
    is higher than the information already available in the WorldModel.
    \param o object type to which this information relates to
    \param pos global position of this object type.
    \param dConf confidence in the object information.
    \return true when information is updated, false otherwise */
bool WorldModel::processUnsureHearInfo( ObjectT o, VecPosition pos,
                                               double dConf )
{
  double     dMinDist;        // used to find closest player to pos
  ObjectT    objInitial = o;

  if( o != OBJECT_TEAMMATE_UNKNOWN && o != OBJECT_OPPONENT_UNKNOWN )
    return false;

  // if o is a teammate find closest teammate to pos and store distance
  if( SoccerTypes::isTeammate( o ) )
    o = getClosestInSetTo( OBJECT_SET_TEAMMATES, pos, &dMinDist);
  else if( SoccerTypes::isOpponent( o ) )  // if o is an opponent, do the same
    o = getClosestInSetTo( OBJECT_SET_OPPONENTS, pos, &dMinDist);

  if( o == getAgentObjectType() &&
     pos.getDistanceTo(getAgentGlobalPosition())<PS->getPlayerDistTolerance())
    return false;  // do not update my own position, localization is better

  // if opponent or teammate was found and distance lies in tolerance distance
  //  update this opponent or teammate with the specified information.
  // else put the information in the first player position of which we have
  //  no information.
  else if( SoccerTypes::isKnownPlayer(o) &&
           dMinDist < PS->getPlayerDistTolerance())
  {
    processPerfectHearInfo( o, pos, dConf );
    return true;
  }

  if( objInitial == OBJECT_TEAMMATE_UNKNOWN )
    o = getFirstEmptySpotInSet( OBJECT_SET_TEAMMATES );
  else if( objInitial == OBJECT_OPPONENT_UNKNOWN )
    o = getFirstEmptySpotInSet( OBJECT_SET_OPPONENTS );
  else
    return false ;  // in case of OBJECT_PLAYER_UNKNOWN

  if( o != OBJECT_ILLEGAL )   // can be the case that there is no empty spot
  {
    processPerfectHearInfo( o, pos, dConf );
    setIsKnownPlayer( o, false );
  }
  return true;
}

/*! This methods fills the heterogeneous player array (stored as 'pt') with the
    specified information. This information is directly parsed from the
    player_type message received from the server. This method is therefore
    only called from the SenseHandler. The information in this array can later
    be used (by the coach) to determine the best heterogeneous player for a
    specific position on the field and to update the parameters in the
    ServerSettings when the player type of the agent is changed.
    \param iIndex index of the heterogeneous player as indicated by the server
    \param dPlayerSpeedMax maximum speed of the player
    \param dStaminaIncMax maximum increase of stamina for the player
    \param dPlayerDecay amount of velocity decay of the player
    \param dIntertiaMoment indication how fast is turned when moving
    \param dDashPowerRate how is power in dash command converted into speed
    \param dPlayerSize what is the size of the player
    \param dKickableMarng in which area can player still kick the ball
    \param dKickRand how much noise is added to kick command
    \param dExtraStamina how much extra stamina does player receive
    \param dEffortMax what is the maximum effort (effective dash)
    \param dEffortMin what is the minimum effort
    \return bool indicating whether update was succesfull. */
bool WorldModel::processNewHeteroPlayer( int iIndex,     double dPlayerSpeedMax,
            double dStaminaIncMax, double dPlayerDecay,  double dInertiaMoment,
            double dDashPowerRate, double dPlayerSize,   double dKickableMargin,
            double dKickRand,      double dExtraStamina, double dEffortMax,
            double dEffortMin )
{
   pt[iIndex].dPlayerSpeedMax = dPlayerSpeedMax;
   pt[iIndex].dStaminaIncMax  = dStaminaIncMax;
   pt[iIndex].dPlayerDecay    = dPlayerDecay;
   pt[iIndex].dInertiaMoment  = dInertiaMoment;
   pt[iIndex].dDashPowerRate  = dDashPowerRate;
   pt[iIndex].dPlayerSize     = dPlayerSize;
   pt[iIndex].dKickableMargin = dKickableMargin;
   pt[iIndex].dKickRand       = dKickRand;
   pt[iIndex].dExtraStamina   = dExtraStamina;
   pt[iIndex].dEffortMax      = dEffortMax;
   pt[iIndex].dEffortMin      = dEffortMin;
   return true;
}

/*! This method sets the catch cycle time for the goalie. This is the cycle time
    that the ball is catched by the goalie. This information is said by the
    referee by indicating the side that catched the ball. This method checks
    whether our side catched the ball. If this is the case, the time is set
    and the goalie cannot catch the ball for a certain cycles
    (see ServerSettings).
    \param rm referee message to indicated side of goalie that catched ball
    \param iTime time the ball was catched.  */
void WorldModel::processCatchedBall( RefereeMessageT rm, Time time )
{
  if( rm == REFC_GOALIE_CATCH_BALL_LEFT && sideSide == SIDE_LEFT )
    timeLastCatch = time;
  else if( rm == REFC_GOALIE_CATCH_BALL_RIGHT && sideSide == SIDE_RIGHT )
    timeLastCatch = time;
  Ball.setGlobalVelocity( VecPosition(0,0), getCurrentTime() );
}

/*! This method sets the performed commands by the agent object. Using this
    information, the future world states can be calculated when an update is
    performed. In this method a timestamp of the current time cycle is added
    to the commands structure for later usage.
    This method is called by the ActHandler when it has sent these commands to
    the server.
    \param commands all the commands that were sent in this cycle
    \param iCommands number of commands that were sent in this cycle. */
void WorldModel::processQueuedCommands( SoccerCommand commands[],int iCommands )
{
  if( iCommands > MAX_COMMANDS )
  {
    cerr << "(WorldModel::setQueuedCommands) queuedCommands array cannot "
         << "contain so many commands!\n";
    return;
  }

  // put all sent commands to the array which stores queued commands.
  for( int i = 0 ; i < iCommands ; i ++ )
  {
    commands[i].time                             = getCurrentTime();
    queuedCommands[(int)commands[i].commandType] = commands[i];
  }
}

/*! This is called to update the WorldModel. It determines the last
    message (see or sense) and updates the worldmodel accordingly. When see
    information is received ("perfect" information) all objects are updated
    with this received information. After sense message the information of
    the object is calculated for the new cycle using the last visual
    information.
    \return bool to indicate whether update succeeded. */
bool WorldModel::updateAll( )
{
  bool        bReturn            = false, bUpdateAfterSee = false;
  bool        bUpdateAfterSense  = false;
  static Time timeLastHoleRecorded;
  static Time timeBeginInterval;
  static Time timePlayersCounted;
  static int  iNrHolesLastTime   = 0;
  static Time timeLastSenseUpdate;
  static Time timeLastSeeUpdate;

  // check if last update of agent was not more than one cycle ago
  if( agentObject.getTimeGlobalPosition() < getCurrentTime() - 1  )
    Log.log( 3, "(WorldModel::updateAfterSenseMessage) missed a sense??" );

  if( isQueuedActionPerformed() == false &&
      timeLastHoleRecorded != getCurrentTime() )
  {
    Log.log( 2, "HOLE recorded" );
    timeLastHoleRecorded = getCurrentTime();
    iNrHoles++;
  }
  
  // call update method depending on last received message
  if( getTimeLastSeeMessage() > timeLastSeeUpdate )
    bUpdateAfterSee = true;
  if( getTimeLastSenseMessage() > timeLastSenseUpdate )
    bUpdateAfterSense = true;

  // rare situation: can occur that see arrives between sense and calling
  // updateAll or sense arrives between see and calling updateAll.
  if( bUpdateAfterSee && bUpdateAfterSense )
  {
    Log.log( 3, "!!! Two updates !!! " );
    Log.log( 3, "see: %d sense: %d", getTimeLastSeeMessage().getTime(),
       getTimeLastSenseMessage().getTime() );
    if( getTimeLastSeeMessage( ) == getTimeLastSenseMessage() )
    {
      bReturn  = updateAfterSenseMessage( );
      bReturn &= updateAfterSeeMessage( );
    }
    else if( getTimeLastSeeMessage( ) < getTimeLastSenseMessage() )
    {
      bReturn  = updateAfterSeeMessage( );
      bReturn &= updateAfterSenseMessage( );
      updateRelativeFromGlobal();
    }

    timeLastSenseUpdate = getTimeLastSenseMessage();
    timeLastSeeUpdate   = getTimeLastSeeMessage();
  }
  else if( bUpdateAfterSee )                        // process see message
  {
    bReturn           = updateAfterSeeMessage( );
    timeLastSeeUpdate = getTimeLastSeeMessage();
  }
  else if( bUpdateAfterSense )                      // process sense message
  {
    bReturn             = updateAfterSenseMessage( );
    timeLastSenseUpdate = getTimeLastSenseMessage();
    updateRelativeFromGlobal();
  }

  // also process fullstate message if applicable
  if( getTimeLastFullstateMessage() == getTimeLastSenseMessage() ){
    bReturn &= updateAfterFullstateMessage();
  }

  processPlayerMessage();

  // check player positions and set ball owner accordingly. 
  updateBallOwner();


  // determine number of holes in last time interval and act accordingly
  int    iTimeDiff = getCurrentTime() - timeBeginInterval;
  double dHolePerc = (double)(iNrHoles - iNrHolesLastTime)/iTimeDiff*100;
  if( ! isLastMessageSee( ) && iTimeDiff % 400 == 0 && dHolePerc > 1.0 &&
      PS->getFractionWaitSeeEnd() > 0.70 )
  {
    PS->setFractionWaitSeeEnd( PS->getFractionWaitSeeEnd() - 0.05 );
    timeBeginInterval = getCurrentTime();
    cerr << "Lowered send time to " << PS->getFractionWaitSeeEnd() <<
            " for player number "   << getPlayerNumber()           <<
            "; nr of holes is "<< dHolePerc << "%" << endl;
    iNrHolesLastTime   = iNrHoles;
  }

  // store some statistics about number of players seen each cycle
  if( bUpdateAfterSense == true  && ! isTimeStopped() &&
      getCurrentTime() != timePlayersCounted )
  {
    iNrTeammatesSeen += getNrInSetInRectangle( OBJECT_SET_TEAMMATES );
    iNrOpponentsSeen += getNrInSetInRectangle( OBJECT_SET_OPPONENTS );
    timePlayersCounted = getCurrentTime();
  }

  // log specific information when log level is set
  if( Log.isInLogLevel( 456 ) )
    logObjectInformation( 456, getAgentObjectType() );

  if( bUpdateAfterSee == true )
    Log.logWithTime(3, "  finished update_all see; start determining action" );
  if( bUpdateAfterSense == true )
    Log.logWithTime(3, "  finished update_all sense; start determining action");

  if( Log.isInLogLevel( 459 ) )
  {
    Log.log( 459, "%s%s", strLastSeeMessage, strLastSenseMessage );
    show( OBJECT_BALL, Log.getOutputStream() );
    show( OBJECT_SET_PLAYERS, Log.getOutputStream() );
  }
  if( ( Log.isInLogLevel( 101 ) && getRelativeDistance( OBJECT_BALL ) < 2.0 ) )
    show( OBJECT_BALL, Log.getOutputStream() );
  if( Log.isInLogLevel( 556 ) )
  {
    Log.log( 556, "%s", strLastSeeMessage );
    show( OBJECT_BALL, Log.getOutputStream() );
  }

  return bReturn;
}

/*! Update the ball owner. Check each player to determine which is closest to ball 
 */
void WorldModel::updateBallOwner()
{
  VecPosition pos;
  VecPosition ballpos;
  //char tmpbuf[50];
  double mindist = 1000000;
  double dist;
  double ballrange = SS->getKickableMargin() + SS->getPlayerSize() + SS->getBallSize() + 1.0;
  // the + 1.0 is a ball range margin, ie, the distance from kickable region that we want 
  // to use for ball ownership. 

  ballpos = Ball.getGlobalPosition();

  // first check this agent itself
  pos =  agentObject.getGlobalPosition();
  dist = ballpos.getDistanceTo(pos);
  
  /*
  Log.log(15, "This Player is at (%f, %f) and ball is at (%f, %f). Dist=%f, ballrange=%f\n", pos.getX(), pos.getY(), ballpos.getX(), ballpos.getY(), dist, ballrange);
  */

  // this is a hack to not give ball to players who are still in initial starting pos. 
  if(pos.getX() <= 0.1 && pos.getX() >= -0.1 &&  pos.getY() <= 0.1 && pos.getY() >= -0.1)
    return;


  // if the player is within ball range
  if(dist < ballrange){
    mindist = dist;
    // set the World Model parameters: 
    BallOwner = agentObject.getType();
  }

  // iterate through the teammates
  for( int i = 0; i < MAX_TEAMMATES; i++ ){

    if(! isConfidenceGood(Teammates[i].getType()))
      continue;

    pos =  Teammates[i].getGlobalPosition();
    dist = ballpos.getDistanceTo(pos);

    /*
    Log.log(15, "Teammates num %d is at (%f, %f) and ball is at (%f, %f). Dist=%f\n", 
	    i+1, pos.getX(), pos.getY(), ballpos.getX(), ballpos.getY(), dist);
    */
	    
    // if the player is within ball range
    if(dist < ballrange){
      // then if the player is closer than any previous players, store its name
      if(dist < mindist){
	mindist = dist;
	// set the World Model parameters: 
	BallOwner = Teammates[i].getType();
      }
    }
  }

  // iterate through the opponents
  for( int i = 0; i < MAX_OPPONENTS; i++ ){

    if(! isConfidenceGood(Opponents[i].getType()))
      continue;

    pos =  Opponents[i].getGlobalPosition();
    dist = ballpos.getDistanceTo(pos);

    Log.log(15, "Opponents num %d is at (%f, %f) and ball is at (%f, %f). Dist=%f\n", 
	    i+1, pos.getX(), pos.getY(), ballpos.getX(), ballpos.getY(), dist);
    // if the player is within ball range
    if(dist < ballrange){
      // now if the player is closer than any previous players, store its name
      if(dist < mindist){
	mindist = dist;
	// set the World Model parameters: 
	BallOwner = Opponents[i].getType();
      }
    }
  }

  if(BallOwner == OBJECT_ILLEGAL){ 
    // then probably play hasn't started yet, just give ownership to everyone on our team
    BallOwner = OBJECT_PLAYER_UNKNOWN;
  }


  /*
  if(SoccerTypes::isTeammate(BallOwner))
    Log.log(10, "Player %s on our team is now ball owner\n", 
	 SoccerTypes::getObjectStr(tmpbuf, BallOwner, "our"));
  else
    Log.log(10, "Player %s on opp team is now ball owner\n", 
	   SoccerTypes::getObjectStr(tmpbuf, BallOwner, "opp"));
  */

}



/******************************************************************************/
/*************** WORLDMODEL: SEE RELATED UPDATES ******************************/
/******************************************************************************/

/*! This method processes the last received see message and updates all the
    values in the world model accordingly. */
void  WorldModel::processLastSeeMessage( )
{
  ObjectT o;
  double  dDist, dDistChange,    dDirChange,     dTemp;
  int     iDir,  iBodyFacingDir, iHeadFacingDir, iTime = UnknownTime;
  Time    time = getTimeLastSenseMessage();
  bool    isGoalie;
  static char strTmp[MAX_MSG];
  char   *strMsg = strLastSeeMessage ;
  iTime = Parse::parseFirstInt( &strMsg );         // get the time

  ObjectT objTeamMin = OBJECT_ILLEGAL;    // players are listed in order of
  ObjectT objMax     = OBJECT_ILLEGAL;    // of initialization, can use this
  ObjectT objOppMin  = OBJECT_ILLEGAL;    // info to determine object type
  ObjectT objMin     = OBJECT_ILLEGAL;    // of unknown players

  while( *strMsg != ')' )                          // " ((objname.." or ")"
  {
    dDist = dDistChange    = dDirChange     = dTemp = UnknownDoubleValue;
    iDir  = iBodyFacingDir = iHeadFacingDir =         UnknownIntValue;

    // jdavin: 
    //    cout << strMsg<<endl;

    strMsg += 2;          // get the object name

    // get the object name from the first part of the string
    o = SoccerTypes::getObjectFromStr( &strMsg, &isGoalie, getTeamName() );

    if( o == OBJECT_ILLEGAL )
    {
      Log.log( 4, "Illegal object in: ", strLastSeeMessage );
      Log.log( 4, "rest of message: ", strMsg );
    }
    else if( SoccerTypes::isKnownPlayer( o ) )    // we know the player
    {
      if( SoccerTypes::isTeammate( o ) )          // set the minimum object
        objTeamMin = o;                           // all objects in the msg
      else                                        // hereafter, should have
        objOppMin  = o;                           // a higher index
    }
    else if( SoccerTypes::isPlayer( o ) )         // and thus an unknown player
    {
      if( SoccerTypes::isTeammate( o ) )          // raise minimum object with 1
      {
        if( objTeamMin == OBJECT_ILLEGAL )
          objTeamMin = OBJECT_TEAMMATE_1;
        else
          objTeamMin = (ObjectT)((int)objTeamMin + 1);
      }
      else if( SoccerTypes::isOpponent( o ) )
      {
        if( objOppMin == OBJECT_ILLEGAL )
          objOppMin = OBJECT_OPPONENT_1;
        else
          objOppMin = (ObjectT)((int)objOppMin + 1);
      }
      strcpy( strTmp, strMsg );                  // get the maximum object by
      objMax = getMaxRangeUnknownPlayer( o, strTmp ); // looking to next object
    }

    dTemp = Parse::parseFirstDouble( &strMsg );   // parse first value
    if ( *strMsg == ')' )                         // if it was the only value
      iDir = (int)dTemp;                          // it was the direction
    else
    {
      dDist = dTemp;                              // else it was distance
      iDir = Parse::parseFirstInt( &strMsg );     // and have to get direction
      if( *strMsg != ')' )                        // stil not finished
      {                                           // get dist and dir change
        dDistChange = Parse::parseFirstDouble( &strMsg );
        dDirChange  = Parse::parseFirstDouble( &strMsg );
      }
      if( *strMsg != ')' )                        // and stil not finished
      {                                           // get also facing dir.
        iBodyFacingDir = Parse::parseFirstInt( &strMsg );
        iHeadFacingDir = Parse::parseFirstInt( &strMsg );
      }
    }
    // go to end object information
    // skip ending bracket of object information.
    Parse::gotoFirstOccurenceOf( ')', &strMsg );
    strMsg++;

    if( SoccerTypes::isPlayer( o ) )
    {
      objMin = (SoccerTypes::isTeammate(o) == true ) ? objTeamMin : objOppMin ;
      if( objMin != OBJECT_ILLEGAL && objMin == objMax )
        o = objMin;
    }

    // process the parsed information (unread values are Unknown...)
    processNewObjectInfo( o, time, dDist, iDir, dDistChange,
           dDirChange, (AngDeg)iBodyFacingDir, (AngDeg)iHeadFacingDir,
           isGoalie, objMin, objMax );
  }
  mapUnknownPlayers( time );       // map players with unknown nr
}

/*! This method is called to update the WorldModel after a see message.
    The update methods for the agent, all teammates, all opponents and the
    ball are called. But only when the last see message of these objects equals
    the time of the last received see message.
    \return bool to indicate whether update succeeded. */
bool WorldModel::updateAfterSeeMessage( )
{
  processLastSeeMessage( );
   // update the agent (global position and angle using flags and lines)
  if( getCurrentTime().getTime() != -1 )
    updateAgentObjectAfterSee( );

  // walk past all players on the field an when new information was perceived
  // (and put in the relative attributes) update this dynamic object. When it
  // was not seen, convert its global position (this is an estimate from the
  // sense message) to a relative position
  double dConfThr = PS->getPlayerConfThr();
  int    iIndex;

  for( ObjectT o = iterateObjectStart( iIndex, OBJECT_SET_PLAYERS, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, OBJECT_SET_PLAYERS, dConfThr ) )
  {
    if( getTimeLastSeen( o ) == getTimeLastSeeMessage() &&
        o != getAgentObjectType() )
      updateDynamicObjectAfterSee   ( o );
    else
      updateObjectRelativeFromGlobal( o );
  }
  iterateObjectDone( iIndex );

  // if ball was seen update him, otherwise make estimated global relative
  if( getTimeLastSeen( OBJECT_BALL ) == getTimeLastSeeMessage() )
    updateDynamicObjectAfterSee   ( OBJECT_BALL );
  else
    updateObjectRelativeFromGlobal( OBJECT_BALL );

  // delete objects from worldmodel that should have been seen, but aren't
  removeGhosts();

  return true;
}


/*! This method is called to update the WorldModel after a fullstate message.
 *  sample fullstate message: 
  (fullstate 136 (pmode play_on) (vmode high normal) (count 0 36 138 0 1 169 0 135) (arm (movable 0) (expires 0) (target 0 0) (count 0)) (score 0 0) ((b) -33.75 16.6875 0 0) ((p l 1 g) -38.148 9.94482 -0.0434769 0.306863 -83.3767 90 (stamina 1330 1 1)))
    \return bool to indicate whether update succeeded. */
bool WorldModel::updateAfterFullstateMessage( ){
  ObjectT o;
  Time    time = getTimeLastFullstateMessage();
  char   *strMsg = strLastFullstateMessage;
  int iTime;
  bool isGoalie;
  double posX, posY, velX, velY, angBody, angNeck;
  PlayerObject *pobj;

  //  cout << "processing "<<strMsg<<endl;

  iTime = Parse::parseFirstInt( &strMsg );         // get the time
  strMsg += 3; // skip space and (
  Parse::gotoFirstOccurenceOf( '(', &strMsg );   // skip pmode 
  // at (vmode 
  strMsg++; // skip (
  Parse::gotoFirstOccurenceOf( '(', &strMsg );   // skip vmode

  // now at (count ...
  strMsg++;
  Parse::gotoFirstOccurenceOf( '(', &strMsg );   // skip (count ...
  //cout << "before (arm : "<<strMsg<<endl;
  // now at (arm ...
  // skip arm info: 
  Parse::gotoFirstOccurenceOf( ')', &strMsg ); strMsg++; // now at " (expires..."
  Parse::gotoFirstOccurenceOf( ')', &strMsg ); strMsg++; // now at " (target ..."
  Parse::gotoFirstOccurenceOf( ')', &strMsg ); strMsg++; // now at " (count ..."
  Parse::gotoFirstOccurenceOf( ')', &strMsg ); strMsg += 3; // now at " (score ..."
  // done skipping arm info

  // skip (score ...)
  Parse::gotoFirstOccurenceOf( ')', &strMsg ); strMsg++; // now at " ((b) ..."
  //cout << "before ((b) : "<<strMsg<<endl;

  // parse ball position and speed
  posX = Parse::parseFirstDouble( &strMsg );
  posY = Parse::parseFirstDouble( &strMsg );
  velX = Parse::parseFirstDouble( &strMsg );
  velY = Parse::parseFirstDouble( &strMsg );

  FullState_Ball.setTimeLastSeen( time );
  FullState_Ball.setGlobalPosition( VecPosition(posX, posY), time );
  FullState_Ball.setGlobalVelocity( VecPosition(velX, velY), time );

  strMsg += 2; // now at (<players info> )

  while( *strMsg != ')' && *strMsg != '\0' ){

    strMsg++;
    // get the object type at the current position in the string
    o = SoccerTypes::getObjectFromFullstateStr( &strMsg, &isGoalie, getSide() );
    pobj = (PlayerObject*)getFullstateObjectFromType( o );
    assert( pobj != NULL);

    posX    = Parse::parseFirstDouble( &strMsg );
    posY    = Parse::parseFirstDouble( &strMsg );
    velX    = Parse::parseFirstDouble( &strMsg );
    velY    = Parse::parseFirstDouble( &strMsg );
    angBody = Parse::parseFirstDouble( &strMsg );
    angNeck = Parse::parseFirstDouble( &strMsg );

    pobj->setTimeLastSeen( time );
    pobj->setGlobalPosition( VecPosition(posX, posY), time );
    pobj->setGlobalVelocity( VecPosition(velX, velY), time );
    pobj->setGlobalBodyAngle( angBody, time );
    pobj->setGlobalNeckAngle(VecPosition::normalizeAngle(angBody+angNeck),time);
    pobj->setIsKnownPlayer( true );

    // skip  (stamina <stamina> <effort> <recovery>))
    Parse::gotoFirstOccurenceOf( ')', &strMsg ); strMsg += 3;
    // now past end of this player (iterate for next player)

  }
  
  //cout << "remaining: "<< strMsg <<endl;

  return true;
}

/*! This method updates an agent after a see message.
    The global position and the global neck angle are calculated using the
    visual information using the method calculateStateAgent.
    \return bool to indicate whether update succeeded. */
bool WorldModel::updateAgentObjectAfterSee(  )
{
  VecPosition posGlobal, velGlobal;
  AngDeg      angGlobal;

  // calculate the state of the agent
  calculateStateAgent( &posGlobal, &velGlobal, &angGlobal );

  // and set the needed attributes
  agentObject.setTimeLastSeen         ( getTimeLastSeeMessage() );
  // store difference with predicted global position to compensate for error
  // in global position when global velocity is calculated for other objects.
  agentObject.setPositionDifference(posGlobal- agentObject.getGlobalPosition());
  agentObject.setGlobalPosition       ( posGlobal, getTimeLastSeeMessage() );
  agentObject.setGlobalPositionLastSee( posGlobal, getTimeLastSeeMessage() );
  agentObject.setGlobalNeckAngle      ( angGlobal );
  agentObject.setGlobalVelocity       ( velGlobal, getTimeLastSeeMessage() );

  return true;
}


/*! This method updates a dynamic object after a see message.
    The global position and the velocity are calculated using the
    visual information. For the ball the method 'calculateStateBall'
    is called, for players the method 'calculateStatePlayer' is called.
    \param o object that should be updated after a see message */
bool WorldModel::updateDynamicObjectAfterSee( ObjectT o )
{
  VecPosition posGlobal, velGlobal;

  if( o == OBJECT_BALL )
  {
    calculateStateBall           ( &posGlobal, &velGlobal              );
    Ball.setGlobalVelocity       ( velGlobal,  getTimeLastSeeMessage() );
    Ball.setGlobalPosition       ( posGlobal,  getTimeLastSeeMessage() );
    Ball.setGlobalPositionLastSee( posGlobal,  getTimeLastSeeMessage() );
    return true;
  }
  else if( SoccerTypes::isKnownPlayer( o ) )
  {
    calculateStatePlayer( o, &posGlobal, &velGlobal );

    PlayerObject *pob = (PlayerObject*) getObjectPtrFromType( o );

    pob->setGlobalVelocity       ( velGlobal, getTimeLastSeeMessage() );
    pob->setGlobalPosition       ( posGlobal, getTimeLastSeeMessage() );
    pob->setGlobalPositionLastSee( posGlobal, getTimeLastSeeMessage() );

    if( pob->getTimeRelativeAngles() == getTimeLastSeeMessage() )
    {
      AngDeg ang = getAgentGlobalNeckAngle() + pob->getRelativeBodyAngle();
      ang = VecPosition::normalizeAngle( ang );
      pob->setGlobalBodyAngle( ang, getTimeLastSeeMessage() );
      ang = getAgentGlobalNeckAngle() + pob->getRelativeNeckAngle();
      ang = VecPosition::normalizeAngle( ang );
      pob->setGlobalNeckAngle( ang, getTimeLastSeeMessage() );
    }
    return true;
  }

  return false;
}


/******************************************************************************/
/******************** SENSE RELATED UPDATES ***********************************/
/******************************************************************************/

/*! This method is called to update the WorldModel after a sense message.
    The current information for the agent, all teammates, all opponents and the
    ball are calculated using the associated methods. This is done by updating
    the objects till the time of the information matches the current time of
    the sense message. After all global
    positions are determined, the relative information is updated using the new
    global information using the method determineRelativeFromGlobal().
    \return bool to indicate whether update succeeded. */
bool WorldModel::updateAfterSenseMessage( )
{
  // update agent information
  updateAgentAndBallAfterSense( );

  // update all global information of players that have a good confidence
  double dConfThr = PS->getPlayerConfThr();
  int    iIndex;

  for( ObjectT o = iterateObjectStart( iIndex, OBJECT_SET_PLAYERS, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, OBJECT_SET_PLAYERS, dConfThr ) )
  {
    if( o != getAgentObjectType() && getTimeGlobalPosition(o) > 0 )
    {
      updateDynamicObjectForNextCycle( o,
                                 getCurrentTime() - getTimeGlobalPosition(o) );
    }
  }
  iterateObjectDone( iIndex);

  return true;
}

/*! This method updates the agent and ball information after a sense message.
    This is done using the current known information and the action that is
    performed in the last server cycle as indicated by the attribute
    'PerformedCommands'
    \return bool to indicate whether update succeeded. */
bool WorldModel::updateAgentAndBallAfterSense( )
{
  // get info from commands from previous cycle (if timestopped current cycle)
  bool        bBallUpdated   = false;
  VecPosition pos            = getAgentGlobalPosition();
  AngDeg      angGlobalNeck  = getAgentGlobalNeckAngle();
  AngDeg      angGlobalBody  = getAgentGlobalBodyAngle();
  Stamina     sta            = getAgentStamina();
  VecPosition velNeck        = agentObject.getSpeedRelToNeck(); // !!
  VecPosition vel            = getAgentGlobalVelocity();
  Time        time           = getCurrentTime() - 1 ;

  // first update information based on performed actions without using
  // the received velocity information.
  for( int i = 0; i < MAX_COMMANDS; i ++  )
  {
    if( performedCommands[i] == true && // sense msg indicates we executed this
        ( queuedCommands[i].time.getTime() == time.getTime() ||      // no hole
          queuedCommands[i].time.getTime() == time.getTime() - 1 ) )// hole
    {
      switch( queuedCommands[i].commandType )
      {
        case CMD_TURN:
        case CMD_TURNNECK:
        case CMD_DASH:
          predictStateAfterCommand( queuedCommands[i], &pos, &vel,
                                    &angGlobalBody, &angGlobalNeck, &sta );
          break;
        case CMD_KICK:
          updateBallAfterKick( queuedCommands[i].dPower,
                               queuedCommands[i].dAngle);
          bBallUpdated     = true;         // should not be updated later on.
          queuedCommands[i].time.updateTime( -1 );
          break;
        case CMD_MOVE:
          pos.setVecPosition( queuedCommands[i].dX, queuedCommands[i].dY );
          initParticlesAgent( pos );
          break;
        default:
          break;
      }
    }
  }

  // update the ball when not done yet
  if( !bBallUpdated )
  {
    m_bPerformedKick = false;
    updateDynamicObjectForNextCycle( OBJECT_BALL, 1 );
  }
  else
    m_bPerformedKick = true;

  // now use velocity information to check whether a collision occured
  velNeck = velNeck.rotate( angGlobalNeck ); // make relative info global
  double dDistBall = pos.getDistanceTo( getBallPos() );
  if( ( dDistBall<SS->getPlayerSize()+SS->getBallSize() && // ball too close AND
        velNeck.getMagnitude( ) < EPSILON                  // vel. gives no info
       ) 
     ||                                                    // OR
     ( dDistBall < SS->getMaximalKickDist() + 0.5  &&      // ball close
      velNeck.getMagnitude( ) > EPSILON &&                 // can compare vel
      sign( vel.getX() ) != sign( velNeck.getX() ) &&      // both signs have
      sign( vel.getY() ) != sign( velNeck.getY() ) &&      // changed and is
      velNeck.getMagnitude() < 0.15 * vel.getMagnitude() ) ) // lessened a lot
  {
    m_bWasCollision = true;
    double dDistPlayer;
    getClosestInSetTo(
         OBJECT_SET_PLAYERS, getAgentObjectType(), &dDistPlayer );
    if( dDistBall < dDistPlayer )
    {
      updateBallForCollision( pos );
      Log.log( 101, "COLLISION WITH BALL %f, vel(%f,%f) velNeck(%f,%f)",
         dDistBall, vel.getX(), vel.getY(), velNeck.getX(), velNeck.getY() );
    }
    else
      Log.log( 101, "COLLISION WITH PLAYER vel(%f,%f) velNeck(%f,%f)",
         vel.getX(), vel.getY(), velNeck.getX(), velNeck.getY() );

    // new vel. agent is the one received from sense message
    vel = velNeck;
  }
  else
  {
    if( dDistBall < SS->getVisibleDistance() )
      Log.log( 101, "No collision: dist: %f, velNeck (%f,%f), vel (%f,%f)",
        dDistBall, velNeck.getX(), velNeck.getY(), vel.getX(), vel.getY() );
    m_bWasCollision = false;
    // use better vel. estimate in sense message to update information
    pos            = getAgentGlobalPosition();
    angGlobalNeck  = getAgentGlobalNeckAngle();
    angGlobalBody  = getAgentGlobalBodyAngle();
    sta            = getAgentStamina();
    vel            = agentObject.getSpeedRelToNeck();

    // calculate velocity at end of previous cycle using velocity from
    // current cycle. Although we do not know direction yet (this is
    // relative to neck which is not yet known), we can use the
    // magnitude to determine traveled  distance (speed) of the agent
    // After neck angle is estimated, we can rotate velocity vector
    // to get actual velocity.
    vel.setMagnitude( vel.getMagnitude()/SS->getPlayerDecay() );

    for( int i = 0; i < MAX_COMMANDS; i ++  )
    {
      if( performedCommands[i] == true &&
          ( queuedCommands[i].time.getTime() == time.getTime() ||
            queuedCommands[i].time.getTime() == time.getTime() - 1 ) )
      {
        switch( queuedCommands[i].commandType )
        {
          case CMD_TURN:
          case CMD_TURNNECK:
            predictStateAfterCommand( queuedCommands[i], &pos, &vel,
                               &angGlobalBody, &angGlobalNeck, &sta );
            break;
          case CMD_MOVE:
            pos.setVecPosition( queuedCommands[i].dX, queuedCommands[i].dY );
            initParticlesAgent( pos );
            break;
          case CMD_DASH:     // no action needed, since resulting
                             // vel. is already available from sense
                             // this vel can be used to update pos
            break;
         default:
            break;
        }
      }
      queuedCommands[i].time.updateTime( -1 );            // processed
    }
    // reset pos and vel information to previous cycle (since can be
    // changed in predictStateAfterCommand)
    vel = agentObject.getSpeedRelToNeck();
    pos = getAgentGlobalPosition();
    vel.setMagnitude( vel.getMagnitude()/SS->getPlayerDecay() );
    vel = vel.rotate( angGlobalNeck ); // rotate velocity using information about neck

    // predict global position using the calculated vel at the end of
    // the previous cycle (power and direction can thus both be set
    // to zero). There is just little noise in this perception, so
    // gives a good estimate
    predictStateAfterDash( 0.0, &pos, &vel, &sta, 0 );
  }

  // update particles that keep track of position of agent using vel
  updateParticlesAgent( vel, true );

  // body angle is not set since relative angle to neck is already
  // set after parsing sense_body message, same holds for stamina
  agentObject.setGlobalPosition ( pos,        getCurrentTime() );
  agentObject.setGlobalVelocity ( vel,        getCurrentTime() );
  agentObject.setGlobalNeckAngle( angGlobalNeck );

  return true;
}

/*! This methods updates the ball after a kick command. First it is checked
    whether the ball is indeed in the kickable range. If this is the case
    the accelaration of the ball is calculated and added to the current
    velocity. With this information the new global position of the ball
    is set.
    \param dPower power value that is passed to the kick comand
    \param dAngle that is passed with the kick command
    \return bool indicating whether update was succesful */
bool WorldModel::updateBallAfterKick( double dPower, AngDeg ang )
{
  if( getRelativeDistance( OBJECT_BALL ) < SS->getMaximalKickDist() )
  {
    // make angle relative to body
    // calculate added acceleration and add it to current velocity
    ang = VecPosition::normalizeAngle(ang + getAgentGlobalBodyAngle() );
    VecPosition acc( getActualKickPowerRate()*dPower, ang, POLAR ) ;
    acc += getGlobalVelocity( OBJECT_BALL );
    if( acc.getMagnitude() > SS->getBallSpeedMax() )
      acc.setMagnitude( SS->getBallSpeedMax() );
    Ball.setGlobalPosition( getGlobalPosition( OBJECT_BALL ) + acc,
                                      getCurrentTime()  );
    Ball.setGlobalVelocity( acc*SS->getBallDecay(), getCurrentTime()  );
  }
  else
  {
    updateDynamicObjectForNextCycle( OBJECT_BALL, 1 );
    Log.log( 21, "(WorldModel::%s) KICK command, but ball not kickable (%f)",
        __FUNCTION__, getRelativeDistance( OBJECT_BALL ) );
  }
  return true;
}

/*! This methods updates a dynamic object for 'iCycles' cycle. This is done by
    updating the global position with the velocity vector. After this
    the velocity vector is decreased with the associated decay.
    \param obj object that should be updated
    \param iCycles denotes for how many cycles dynamic object should be updated
    \return bool indicating whether update was succesful */
bool WorldModel::updateDynamicObjectForNextCycle( ObjectT obj, int iCycles)
{
  DynamicObject *o = (DynamicObject*) getObjectPtrFromType( obj );
  if( o == NULL )
    return false;

  // get velocity and add it to current global position
  VecPosition vel = getGlobalVelocity( obj );
  VecPosition pos = getGlobalPosition( obj );

  if( SoccerTypes::isBall( obj ) )
  {
    for( int i = 0; i < iCycles ; i++ )
    {
      pos += vel;
      vel *= SS->getBallDecay();
    }
  }
  else if( SoccerTypes::isPlayer( obj ) )
  {
    for( int i = 0; i < iCycles ; i++ )
    {
      // no idea of intention, just let him roll out
      pos += vel;
      vel *= SS->getPlayerDecay();
    }
  }

  o->setGlobalVelocity ( vel, getCurrentTime() );
  o->setGlobalPosition ( pos, getCurrentTime() );

  return true;
}


/*! This method checks updates the ball information when it overlaps with a
    player. If this is the case the player and the ball are moved to a position
    where they do not overlap any more. Both velocities are multiplied with 0.1
    \return true when collision occured, false otherwise */
bool WorldModel::updateBallForCollision( VecPosition posAgent )
{
  VecPosition posBall  = getGlobalPosition( OBJECT_BALL );
  VecPosition velBall  = getGlobalVelocity( OBJECT_BALL );

  // get the direction the ball was coming from and put it at small distance
  // and multiply the velocity with -0.1
  AngDeg ang = (posBall - posAgent).getDirection();
  Ball.setGlobalPosition( posAgent + VecPosition( 0.2, ang, POLAR ),
                            getCurrentTime() );
  Ball.setGlobalVelocity( velBall*-0.1, getCurrentTime() );
  Log.log( 101, "updateBallForCollision; vel from (%f,%f) to (%f,%f)",
        velBall.getX(), velBall.getY(), getGlobalVelocity(OBJECT_BALL).getX(),
        getGlobalVelocity(OBJECT_BALL).getY() );
  return true;
}

/*! This method uses the global position of the agent and the global
    position of all objects to update their relative information. When
    the new global position of the objects is updated after a sense message,
    their relative information is not up to date.
   \return bool indicating whether update was succesful */
bool WorldModel::updateRelativeFromGlobal()
{
  double dConfThr = PS->getPlayerConfThr();
  int    iIndex;

  // update all player objects
  for( ObjectT o = iterateObjectStart( iIndex, OBJECT_SET_PLAYERS, dConfThr );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, OBJECT_SET_PLAYERS, dConfThr ) )
  {
    if( o != getAgentObjectType() )
      updateObjectRelativeFromGlobal( o );
  }
  iterateObjectDone( iIndex);

  // and also the ball
  if( isConfidenceGood( OBJECT_BALL ) )
    updateObjectRelativeFromGlobal( OBJECT_BALL );

  // flags and lines are not updated, since they are not used except immediately
  // after see message when they're up to date.
  return true;
}

/*! This method updates the relative information for one object using the
    global information of this object and the global information of the
    agent.
    \param o object information that should be updated.
    \return bool indicating whether the update was succesfull. */
bool WorldModel::updateObjectRelativeFromGlobal( ObjectT o )
{
  Object *obj = (Object*) getObjectPtrFromType( o );
  if( obj == NULL )
    return false;

  // get global position and translate and rotate it to agent neck
  VecPosition rel = obj->getGlobalPosition();
  rel = rel.globalToRelative( getAgentGlobalPosition(), getAgentGlobalNeckAngle() );
  obj->setRelativePosition( rel, obj->getTimeGlobalPosition() );
  return true;
}

/*! This method calculates the different state information of the agent, that
    is the global position, global velocity and global neck angle using the
    available information in the world model. This method uses a particle
    filter to determine this information.
    \param posGlobal will be filled with global position of the agent
    \param velGlobal will be filled with global velocity of the agent
    \param angGlobal will be filled with global neck angle of the agent
    \return global position of the agent */
bool WorldModel::calculateStateAgent( VecPosition *posGlobal,
                     VecPosition *velGlobal, AngDeg *angGlobal )
{
  int    iNrLeft;

  // first determine global neck angle
  ObjectT objLine = getFurthestRelativeInSet( OBJECT_SET_LINES );
  if( objLine != OBJECT_ILLEGAL )
  {
    double angGlobalLine = getGlobalAngle  ( objLine );
    AngDeg angRelLine    = getRelativeAngle( objLine );
    *angGlobal           = angGlobalLine - angRelLine;
    *angGlobal           = VecPosition::normalizeAngle( *angGlobal );
  }
  else
  {
    Log.log( 21, "(WorldModel::%s) no line in last see message", __FUNCTION__ );
    *angGlobal           = getAgentGlobalNeckAngle();
  }

  // use global neck angle to determine estimate of current global velocity
  // neck angle is better estimate than after sense -> better estimate velocity
  // update all position particles of the agent with this velocity
  //  'false' denotes update after see message. Since global neck angle can
  //  be determined more precise after 'see' it is better to predict position
  //  again (although it was already done after sense)
  // and then check which particles are possible given current perceptions
  *velGlobal = agentObject.getSpeedRelToNeck().rotate( *angGlobal );
  velGlobal->setMagnitude( velGlobal->getMagnitude()/SS->getPlayerDecay() );

  updateParticlesAgent          ( *velGlobal, false );
  iNrLeft = checkParticlesAgent ( *angGlobal        );

  if( iNrLeft == 0 ) // if no particles left, initialize all particles
  {
    // initialize particles (from random samples using closest flag)
    // check particles are then checked with other flags
    initParticlesAgent ( *angGlobal );
    iNrLeft = checkParticlesAgent( *angGlobal );
    if( iNrLeft == 0 )
    {
      // not succeeded, use second method (weighted average flags)
      // and initialize all particles to this position
      calculateStateAgent2( posGlobal, velGlobal, angGlobal );
      initParticlesAgent( *posGlobal );
      return false;
    }
  }

  // determine global position (= average of all particles)
  // and resample all particles
  *posGlobal = averageParticles( particlesPosAgent, iNrLeft );
  resampleParticlesAgent( iNrLeft );

  // use the position to calculate better global neck angle of the agent
  // and recalculate global velocity with improved neck angle
  AngDeg ang = calculateAngleAgentWithPos( *posGlobal );
  if( ang != UnknownAngleValue )
    *angGlobal = ang;

  *velGlobal = agentObject.getSpeedRelToNeck().rotate(*angGlobal);
  return true;
}

/*! This method initializes all particles that represent the global position
    of the agent. This is done using the closest perceived flag. Points are
    generated from the area that could generate the perceived information.
    The global neck angle of the agent 'angGlobal' is used to determine the
    global position of the agent based on the perceived relative information
    to the closest flag.
    \param angGlobal global neck angle of the agent */
void WorldModel::initParticlesAgent( AngDeg angGlobal )
{
  double  dDist, dMaxRadius, dMinRadius, dInput;
  AngDeg  ang;

  // get closest flag from which samples will be generated
  ObjectT objFlag = getClosestRelativeInSet( OBJECT_SET_FLAGS );

 if( objFlag == OBJECT_ILLEGAL )
 {
    Log.log( 21, "(WorldModel::%s) no flag in last see message", __FUNCTION__ );
    return;
 }

  // get the distance to this flag and the possible range it was derived from.
  dInput = getRelativeDistance( objFlag );
  getMinMaxDistQuantizeValue( dInput, &dMinRadius, &dMaxRadius,
                                               SS->getQuantizeStepL(), 0.1 ) ;

  // get the perceived angle to this flag (add 180 to get angle relative from
  // flag to agent ) and make it global by adding global neck angle agent.
  AngDeg angFlag   = getRelativeAngle( objFlag ) + 180 + angGlobal ;

  // for all particles
  for( int i = 0 ; i < iNrParticlesAgent ; i++ )
  {
    // determine random point from distance range and
    // determine random point from the range it could be generated from
    // angles are rounded and since noise is in global neck angle and relative
    // angle flag, maximum error is in range [-0.5,0.5] + [-0.5,0.5] = [-1,1].
    dDist = dMinRadius + drand48()*(dMaxRadius-dMinRadius);
    ang   = VecPosition::normalizeAngle( angFlag - 1.0 + 2*drand48() );

    // create random point from possible interval
    particlesPosAgent[i] = getGlobalPosition( objFlag ) +
                           VecPosition( dDist, ang, POLAR );
  }
}

/*! This method initializes all particles that represent the global position
    of the agent. All points are initialized with the position 'posInitial'.
    This method can be used when you are completely sure of the position of
    the agent (for example after a 'move' command).
    \param posInitial global position of the agent */
void WorldModel::initParticlesAgent( VecPosition posInitial )
{
  for( int i = 0 ; i < iNrParticlesAgent ; i++ )
    particlesPosAgent[i] = posInitial;
}

/*! This method checks all the particles that represent the global position
    of the agent using the flags in the last perceived see message. The
    global neck angle of the agent is used to make the perceived information
    global. Particles that are not possible are deleted and all legal particles
    will be shifted to the beginning of the array. The number of legal particles
    is returned.
    \param angGlobalNeck global neck of the angle
    \return number of legal particles */
int WorldModel::checkParticlesAgent( AngDeg angGlobalNeck  )
{
  double dMaxRadius, dMinRadius, dInput, dDist;
  AngDeg ang;
  int    iIndex, iNrLeft = 0, iLength = iNrParticlesAgent;

  // for all current perceived flags
  for( ObjectT o = iterateObjectStart( iIndex, OBJECT_SET_FLAGS, 1.0 );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, OBJECT_SET_FLAGS, 1.0 ) )
  {
    iNrLeft = 0;                        // reset number of correct particles
    dInput = getRelativeDistance( o );  // get possible distance range
    getMinMaxDistQuantizeValue( dInput, &dMinRadius, &dMaxRadius,
                                    SS->getQuantizeStepL(), 0.1 )  ;

    // find all "correct points"
    for( int i = 0; i < iLength; i ++ )
    {
      // determine distance particle to flag
      // determine difference in direction between direction between global
      // flag and agent position and global perceived direction
      dDist = particlesPosAgent[i].getDistanceTo( getGlobalPosition( o ) );
      ang = (getGlobalPosition(o) - particlesPosAgent[i]).getDirection();
      ang = ang - ( getRelativeAngle( o ) + angGlobalNeck );

      // if in possible range, save it (maximum angle range is 0.5 for
      // neck angle and 0.5 for relative flag angle gives 1.0)
      if( dDist > dMinRadius && dDist < dMaxRadius &&
          fabs(VecPosition::normalizeAngle( ang )) <= 1.0 )
        particlesPosAgent[iNrLeft++] = particlesPosAgent[i];
    }
    // change maximum of correct particles
    iLength = iNrLeft;
  }
  return iNrLeft;
}

/*! This method updates all the particles that represent the global position
    of the agent.The velocity 'vel' is added to each particle to move to
    the next location. When the last boolean argument is true, the velocity
    is simple added. If this value equals true the previously added velocity
    is subtracted before the specified velocity 'vel' is added. This for
    instance occurs when a better velocity estimate (after a see message) is
    determined with which the particles should be updated.
    \param velocity that should be added to each particle
    \param bAfterSense bool denoting whether this update is done after sense
           or see. When true last added velocity is first subtracted. */
void WorldModel::updateParticlesAgent( VecPosition vel, bool bAfterSense )
{
  // used to denote last added velocity
  static VecPosition prev_vel;

  for( int i = 0; i < iNrParticlesAgent  ; i ++ )
  {
    if( bAfterSense == false ) // if after see, subtract last added 'vel'
    {
      particlesPosAgent[i].setX( particlesPosAgent[i].getX() - prev_vel.getX());
      particlesPosAgent[i].setY( particlesPosAgent[i].getY() - prev_vel.getY());
    }

    particlesPosAgent[i].setX( particlesPosAgent[i].getX( ) + vel.getX() );
    particlesPosAgent[i].setY( particlesPosAgent[i].getY( ) + vel.getY() );
  }
  prev_vel = vel;
}


/*! This method returns the average of the particles contained in 'posArray'.
    \param posArray containing all possible global positions of the agent
    \param iLength number of particles in posArray
    \return average position of all the particles */
VecPosition WorldModel::averageParticles( VecPosition posArray[], int iLength )
{
  if( iLength == 0 )
    return VecPosition( UnknownDoubleValue, UnknownDoubleValue );

  // take average of particles
  double x = 0, y = 0;
  for( int i = 0; i < iLength  ; i ++ )
  {
     x += posArray[ i ].getX( );
     y += posArray[ i ].getY( );
  }

  return VecPosition( x/iLength, y/iLength );
}

/*! This method resamples the particles that represent the global position of
    the agent. 'iLength' denotes the number of particles contained in
    'posArray' and 'iLeft' the points that do do not have to be resampled.
    A particle is resampled by taking a random particle from the first part
    of the array.
    \param iLeft number of particles that should be contained */
void WorldModel::resampleParticlesAgent( int iLeft )
{
  for ( int i = iLeft; i < iNrParticlesAgent; i ++ )
    particlesPosAgent[ i ] = particlesPosAgent[ (int)(drand48()*iLeft) ];
}

/*! This method calculates the different state information of the agent, that
    is the global position, global velocity and global neck angle using the
    available information in the world model. This method uses a weighted
    average of all currently perceived flags.
    \param posGlobal will be filled with global position of the agent
    \param velGlobal will be filled with global velocity of the agent
    \param angGlobal will be filled with global neck angle of the agent
    \return global position of the agent */
bool WorldModel::calculateStateAgent2( VecPosition *posGlobal,
                                   VecPosition *velGlobal, AngDeg *angGlobal)
{
  double      x=0.0, y=0.0, dMinRadius1, dMaxRadius1, dMinRadius2, dMaxRadius2;
  double      dTotalVar = UnknownDoubleValue, dVar, K;
  int         iIndex1, iIndex2;
  ObjectT     obj2;
  VecPosition pos;

  // for all flags that were perceived in last see message
  for( ObjectT obj1 = iterateObjectStart( iIndex1, OBJECT_SET_FLAGS, 1.0 );
       obj1 != OBJECT_ILLEGAL;
       obj1 = iterateObjectNext ( iIndex1, OBJECT_SET_FLAGS, 1.0 ) )
  {
    // calculate global position with all other flags using two flags
    iIndex2 = iIndex1;
    for( obj2 = iterateObjectNext ( iIndex2, OBJECT_SET_FLAGS, 1.0 ) ;
         obj2 != OBJECT_ILLEGAL;
         obj2 = iterateObjectNext ( iIndex2, OBJECT_SET_FLAGS, 1.0 ) )
    {
      // calculate the position using the two flags
      pos = calculatePosAgentWith2Flags( obj1, obj2 );

      // get distance range from which perceived value can originate from
      // calculate variance (=weighted factor) based on the distance to the
      // two flags, use variance corresponding to uniform distribution
      // this is not completely correct, better would be to use the
      // intersection area size of the two circle, but is too computational
      // intensive
      getMinMaxDistQuantizeValue(getRelativeDistance(obj1),
                 &dMinRadius1, &dMaxRadius1, SS->getQuantizeStepL(), 0.1 )  ;
      getMinMaxDistQuantizeValue(getRelativeDistance(obj2),
                 &dMinRadius2, &dMaxRadius2, SS->getQuantizeStepL(), 0.1 )  ;
      dVar =  (dMaxRadius1-dMinRadius1)*(dMaxRadius1-dMinRadius1)/12;
      dVar += (dMaxRadius2-dMinRadius2)*(dMaxRadius2-dMinRadius2)/12;

      if( pos.getX() != UnknownDoubleValue &&
          dTotalVar  == UnknownDoubleValue )
      {
        dTotalVar = dVar;                     // initialize the position
        x         = pos.getX();
        y         = pos.getY();
      }
      else if( pos.getX() != UnknownDoubleValue )
      {
        K = dTotalVar / ( dVar + dTotalVar ); // otherwise use new position
        x += K*( pos.getX() - x );            // based on weighted variance
        y += K*( pos.getY() - y );
        dTotalVar -= K*dTotalVar;
      }
    }
    iterateObjectDone( iIndex2 );
  }
  iterateObjectDone( iIndex1 );
  posGlobal->setVecPosition( x, y );

  // now calculate global position (experiments show best results are obtained
  // when average with all perceived flags is taken).
  *angGlobal = calculateAngleAgentWithPos( *posGlobal );

  // update velocity since after 'see' we have a better estimate of neck angle
  *velGlobal = agentObject.getSpeedRelToNeck().rotate(*angGlobal);

  return true;
}

/*! This method calculates the global position of the agent using two flags.
    Using the perceived distance to the two flag, two circles are created with
    as center the global position of the flag. The intersection of these two
    circles is returned as the global position of the agent (note that the
    relative direction to both flags can be used to determine which of the
    two possible intersection points is the correct position).
    It is assumed that the relative information of the specified flags are
    from the last see message.
    \param objFlag1 object type of first flag
    \param objFlag2 object type of second flag
    \return global position of the agent */
VecPosition WorldModel::calculatePosAgentWith2Flags( ObjectT objFlag1,
                                                     ObjectT objFlag2 )
{
    double      xA, yA, xB, yB, rA, rB;
    AngDeg      aA, aB;

    // get all information of the two flags
    xA = getGlobalPosition  ( objFlag1 ).getX();
    yA = getGlobalPosition  ( objFlag1 ).getY();
    rA = getRelativeDistance( objFlag1 );
    aA = getRelativeAngle   ( objFlag1 );
    xB = getGlobalPosition  ( objFlag2 ).getX();
    yB = getGlobalPosition  ( objFlag2 ).getY();
    rB = getRelativeDistance( objFlag2 );
    aB = getRelativeAngle   ( objFlag2 );

    double      L, dx, dy, d_par, d_orth;
    double      x, y;
    // Sign is like this 'because' y-axis increases from top to bottom
    double sign = ((aB - aA) > 0.0) ? 1.0 : -1.0;

    // From Cosinus rule: rB rB = L L + rA rA - 2 L rA cos(aB)
    // with:              rA cos(aB) = d_par
    // and:               d_par^2 + d_orth^2 = rA^2
    // Finally:
    // Position from landmark position and vectors parallel and orthogonal
    // to line from landmark A to B

    dx = xB - xA;
    dy = yB - yA;
    L = sqrt(dx*dx + dy*dy);                   // distance between two flags

    dx /= L; dy /= L;                          // normalize

    d_par = (L*L + rA*rA - rB*rB) / (2.0 * L); // dist from flag1 to orth proj
    double arg = rA*rA - d_par*d_par;
    d_orth = (arg > 0.0) ? sqrt(arg) : 0.0;

    x = xA + d_par * dx - sign * d_orth * dy;
    y = yA + d_par * dy + sign * d_orth * dx;

    return VecPosition( x, y );
}

/*! This method calculates the global neck angle of the agent using all
    available flag information and the estimation of the current global position
    of the agent 'pos'. For each perceived flag, the global direction is
    calculated using the known global position of the flag and the specified
    global position 'pos' of the agent. Then the global neck direction of the
    agent is determined using the relative direction to this flag.
    The average neck angle of all perceived flags is returned.
    \param pos current global position of the agent
    \return global neck angle of the agent */
AngDeg WorldModel::calculateAngleAgentWithPos( VecPosition pos )
{
  int    iNrFlags = 0, iIndex;
  double dCosX=0, dSinX=0 ,dAngleNow, xA, yA, aA;

  for( ObjectT obj = iterateObjectStart( iIndex, OBJECT_SET_FLAGS, 1.0 );
       obj != OBJECT_ILLEGAL;
       obj = iterateObjectNext ( iIndex, OBJECT_SET_FLAGS, 1.0 ) )
  {
    xA = getGlobalPosition( obj ).getX();
    yA = getGlobalPosition( obj ).getY();
    aA = getRelativeAngle( obj );

    // calculate global direction between flag and agent
    // calculate global neck angle by subtracting relative angle to flag
    dAngleNow = atan2Deg( yA - pos.getY(), xA - pos.getX() );
    dAngleNow = VecPosition::normalizeAngle( dAngleNow - aA );

    // add cosine part of angle and sine part separately; this to avoid
    // boundary problem when computing average angle (average of -176 and
    // 178 equals -179 and not 1).
    dCosX += cosDeg( dAngleNow );
    dSinX += sinDeg( dAngleNow );
    iNrFlags++;

  }
  iterateObjectDone( iIndex );

  // calculate average cosine and sine part and determine corresponding angle
  dCosX /= (double)iNrFlags;
  dSinX /= (double)iNrFlags;
  if( iNrFlags == 0 )
    return UnknownAngleValue;

  return VecPosition::normalizeAngle( atan2Deg( dSinX, dCosX  ) )  ;
}


/*! This method returns the velocity from the object o given the perceptions
    from the see message. It uses the soccer server formula directly
    (thus assuming no noise).
    \param o object type for which velocity is determined
    \return global velocity of the ball */
VecPosition WorldModel::calculateVelocityDynamicObject( ObjectT o )
{
  DynamicObject * dobj = (DynamicObject*) getObjectPtrFromType( o );
  if( dobj == NULL )
    return VecPosition( UnknownDoubleValue, UnknownDoubleValue );
  double dDistCh = dobj->getRelativeDistanceChange(   );
  double angCh   = dobj->getRelativeAngleChange   (   );
  double dDist   = getRelativeDistance            ( o );
  double ang     = getRelativeAngle               ( o );

  double velx = dDistCh * cosDeg(ang) - Deg2Rad(angCh) * dDist * sinDeg( ang );
  double vely = dDistCh * sinDeg(ang) + Deg2Rad(angCh) * dDist * cosDeg( ang );

  VecPosition vel( velx, vely );
  return vel.relativeToGlobal( getAgentGlobalVelocity(),
                               getAgentGlobalNeckAngle() );
}


/*! This method calculates the global position and velocity of the ball using
    the newest visual information.
    \param *posGlobal will be filled with the global position
    \param *velGlobal will be filled with the global velocity
    \return true when calculations were succesful, false otherwise */
bool WorldModel::calculateStateBall( VecPosition *posGlobal,
                                     VecPosition *velGlobal )
{
  // set the global position of the ball as follows:
  //  - get the relative position from the agent to it in world-axis
  //  - add global position agent to this relative position
  VecPosition posRelWorld =
      VecPosition( getRelativeDistance( OBJECT_BALL ),
                   getRelativeAngle( OBJECT_BALL ) + getAgentGlobalNeckAngle(),
                   POLAR );
  *posGlobal = getAgentGlobalPosition() + posRelWorld;

  if( isBeforeKickOff() )
    posGlobal->setVecPosition( 0, 0 );

  *velGlobal = getGlobalVelocity(OBJECT_BALL);
  if( Ball.getTimeChangeInformation( ) == getTimeLastSeen( OBJECT_BALL ) )
  {
    *velGlobal = calculateVelocityDynamicObject( OBJECT_BALL );;
    Log.log( 458, "vel based on change info: (%f,%f)", velGlobal->getX(),
                          velGlobal->getY() );
    // average result with predicted velocity from last see message
    // this has the best result for the ball: see program absvelocity.C
    // only use average when no big chance (kick,turn or dash) has occured
    if( fabs(velGlobal->getX() - getGlobalVelocity(OBJECT_BALL).getX())
                                      <= 2*SS->getBallRand()*getBallSpeed() &&
        fabs(velGlobal->getY() - getGlobalVelocity(OBJECT_BALL).getY())
                                      <= 2*SS->getBallRand()*getBallSpeed() )
    {
      *velGlobal = (*velGlobal + getGlobalVelocity(OBJECT_BALL))/2.0;
      Log.log( 458, "average ball vel to (%f,%f)", velGlobal->getX(),
                          velGlobal->getY() );
    }
  }
  else if( getRelativeDistance(OBJECT_BALL) < SS->getVisibleDistance() &&
           getTimeLastSeeMessage() - Ball.getTimeGlobalPosDerivedFromSee() < 3 )
  {
    // no change information but have got feel info -> use position based vel.
    // get the difference with the previous known global position
    // and subtract the position difference (this difference is caused by
    // the fact that the global position calculation contains a lot of noise
    // now the noise is filtered, since we compare the velocity as if its
    // position was seen "with the noise" of the last cycle).

    VecPosition posGlobalDiff   = *posGlobal - Ball.getGlobalPositionLastSee()
                                         - agentObject.getPositionDifference();
    Log.log( 101, "2 pos: ball(%f,%f), ball_prev(%f,%f), agentdiff(%f,%f)",
       posGlobal->getX(), posGlobal->getY(),
       Ball.getGlobalPositionLastSee().getX(), 
       Ball.getGlobalPositionLastSee().getY(),
       agentObject.getPositionDifference().getX(),
       agentObject.getPositionDifference().getY() );
       
    // difference in global positions is distance traveled, we have to make
    // distinction whether this distance is traveled in one or two cycles.
    // 1 cycle:
    //   distance difference equals last velocity so only have to multiply with
    //   decay
    // 2 cycles:
    //   do not update, since kick can be performed causing large change
    //   of which we have no see information (so thus use sense update)

    if( getTimeLastSeeMessage() - Ball.getTimeGlobalPosDerivedFromSee() == 1 &&
        m_bWasCollision == false )
    {
      *velGlobal = posGlobalDiff*SS->getBallDecay();
      Log.log( 458, "vel based on 2 pos: (%f,%f)", velGlobal->getX(),
                          velGlobal->getY() );
      Log.log( 101, "vel based on 2 pos: (%f,%f)", velGlobal->getX(),
                          velGlobal->getY() );
    }
    else if( getTimeLastSeeMessage()-Ball.getTimeGlobalPosDerivedFromSee() > 2 )
    {
      Log.log( 20, "(WorldModel:%s) time difference too large" ,__FUNCTION__ );
    }
  }
  else
    Log.log( 458, "vel ball not updated", velGlobal->getX(), velGlobal->getY() );
    ;  // object too far away do not estimate velocity, sense has updated it
       // already

  // change velocity when the ball
  //   has been catched or play mode is not play_on
  //   is in kickable distance opponent
  //   higher than max speed with added noise
  // change position when it is a kick_in: know y coordinate
  // change position when it is a back_pass_[rl]: know positions
  if( getTimeSinceLastCatch() < 2 ||
      (getPlayMode() != PM_PLAY_ON && !isGoalKickUs() && !isGoalKickThem()))
    velGlobal->setMagnitude( 0.0 );
  else if( getNrInSetInCircle( OBJECT_SET_OPPONENTS,
                            Circle(*posGlobal,SS->getMaximalKickDist())) > 0 
           && getRelativeDistance( OBJECT_BALL ) > SS->getMaximalKickDist() )
    velGlobal->setMagnitude( 0.0 );
  else if( velGlobal->getMagnitude() >
                           ( 1.0 + SS->getBallRand() )*SS->getBallSpeedMax() )
    velGlobal->setMagnitude( SS->getBallSpeedMax() );

  if( isKickInUs() || isKickInThem() )
    posGlobal->setY( sign( posGlobal->getY() ) * PITCH_WIDTH/2.0 );
  else if( isBackPassUs() )
    posGlobal->setVecPosition( - PENALTY_X,
                               sign(posGlobal->getY())*PENALTY_AREA_WIDTH/2.0 );
  else if( isBackPassThem() )
    posGlobal->setVecPosition( + PENALTY_X,
                               sign(posGlobal->getY())*PENALTY_AREA_WIDTH/2.0 );


  Log.log( 458, "final ball vel: (%f,%f)", velGlobal->getX(),velGlobal->getY());
  if( getRelativeDistance(OBJECT_BALL) < SS->getVisibleDistance() )
    Log.log( 101, "direction old: %f, new: %f",
         ( getGlobalPosition( OBJECT_BALL ) - getAgentGlobalPosition()).getDirection(),
         ( *posGlobal - getAgentGlobalPosition()).getDirection() );
  return true;
}


/*! This method determines the current state of a player and is called after
    a see message has arrived.
    \param o object type of the player
    \param *posGlobal will be filled with global position of the player
    \param *velGlobal will be filled with global velocity of the player
    \return bool indicating whether calculations succeeded. */
bool WorldModel::calculateStatePlayer( ObjectT o, VecPosition *posGlobal,
                                       VecPosition *velGlobal )
{
  PlayerObject *pob = (PlayerObject*) getObjectPtrFromType( o );
  if( pob == NULL )
    return false;

  // set the global position of this dynamic object as follows:
  //  - get the relative position from the agent to it in world-axis
  //  - add global position agent to this relative position
  VecPosition posRelWorld =
      VecPosition( getRelativeDistance( o ),
                   getRelativeAngle( o ) + agentObject.getGlobalNeckAngle(),
                   POLAR );
  *posGlobal = getAgentGlobalPosition() + posRelWorld;

  velGlobal->setVecPosition( 0, 0);
  if( pob->getTimeChangeInformation( ) == getTimeLastSeen( o ) )
  {
    // calculate the global velocity using the distance and angle change
    // with the formula from the soccermanual
    *velGlobal = calculateVelocityDynamicObject( o );
  }
  else
      ; // object too far away do not estimate velocity, sense has updated it
        // already and does not really matter then

  if( velGlobal->getMagnitude() >=
                      ( 1.0 + SS->getPlayerRand())*SS->getPlayerSpeedMax() )
    velGlobal->setMagnitude( SS->getPlayerSpeedMax() );

  return true;
}


/*! This method determines the minimum and maximum input values that will
    produce a quantized (noise used in the soccer server) distance of 'dOutput'.
    With other words, this is the range of values that will have the same
    quantized value 'dOutput'. The quantized steps are defined by x1 and x2.
    See the soccer server manual for details.
    \param dOutput resulting quantized value
    \param *dMin will be filled with minimum possible value
    \param *dMax will be filled with maximum possible value
    \param x1 value of inner quantize call (0.1 for player/ball, 0.01 for flags)
    \param x2 value of outer quantize call (normally 0.1)
    \return bool indicating whether values were filled correctly */
bool WorldModel::getMinMaxDistQuantizeValue( double dOutput, double *dMin,
                                        double *dMax,   double x1, double x2 )
{
  // change output a little bit to circumvent boundaries
  // q = quantize(e^(quantize(ln(V),x1)),x2) with quantize(V,Q) = rint(V/Q)*Q
  // e^(quantize(ln(V),x1)_min = invQuantize( q, x2 )
  // quantize(ln(V),x1) = ln ( invQuantize( q, x2 ) )
  // ln(V) = invQuantize( ln ( invQuantize( q, x2 ) ), x1 )
  // V_min = e^( invQuantize( ln ( invQuantize( q, x2 ) ), x1 ) )
  // apply inverse quantize twice to get correct value
  dOutput -= 1.0e-10;



  /* Edited by Dongryeol Lee: Applied bug fix for taking a log of negative
     number.
  */
  
  /* The following line is the original line

  *dMin = exp( invQuantizeMin( log( invQuantizeMin(dOutput,x2) ), x1 ) );

  */


  /* This is our fix: 
   *dMin = exp( invQuantizeMin( log( max(
   invQuantizeMin(dOutput,x2), 1e-10) ), x1 ) );  
  */
  double invResult = invQuantizeMin(dOutput, x2);
  

  if(invResult < 1.0e-10)
    *dMin = exp( invQuantizeMin( log( 1.0e-10 ), x1 ) );
  else
    *dMin = exp( invQuantizeMin( log( invResult ), x1 ) );
    

  dOutput += 2.0e-10;
  *dMax = exp( invQuantizeMax( log( invQuantizeMax(dOutput,x2) ), x1 ) );
  return true;
}

/*! This method determines the minimum and maximum input values that will
    produce a quantized (noise used in the soccer server) for the direction
    change of 'dOutput'.
    See the soccer server manual for details.
    \param dOutput resulting quantized value for direction change
    \param *dMin will be filled with minimum possible value
    \param *dMax will be filled with maximum possible value
    \param x1 value of outer quantize call (normally 0.1)
    \return bool indicating whether values were filled correctly */
bool WorldModel::getMinMaxDirChange( double dOutput, double *dMin,
                                     double *dMax,   double x1     )
{
 *dMin = invQuantizeMin( dOutput, x1 ) ;
 *dMax = invQuantizeMax( dOutput, x1 ) ;
 return true;
}

/*! This method determines the minimum and maximum input values that will
    produce a quantized (noise used in the soccer server) for the distance
    change of 'dOutput'.
    See the soccer server manual for details.
    \param dOutput resulting quantized value for direction change
    \param dDist distance to the perceived object
    \param *dMin will be filled with minimum possible value
    \param *dMax will be filled with maximum possible value
    \param x1 quantize step for distance change
    \param xDist1 value of inner quantize call to determine quantize value
                  corresponding to distance
    \param xDist2 value of outer quantize call to determine quantize value
                  corresponding to distance
    \return bool indicating whether values were filled correctly */
bool WorldModel::getMinMaxDistChange( double dOutput, double dDist,
         double *dMin, double *dMax, double x1, double xDist1, double xDist2)
{
  // Q_dist = quantize(e^(quantize(ln(V),xDist1)),xDist2)
  // q = Q_dist * Quantize( distance_change/distance, x1 )
  // dOutput = q/Q_dist = Quantize( distance_change/distance, x1 )
  // (distance_change/distance)_min = invQmin(q/Q_dist, x1 )
  // real distance is not know so should take into account distance range
  double dMinDist, dMaxDist;
  getMinMaxDistQuantizeValue( dDist, &dMinDist, &dMaxDist, xDist1, xDist2 );
  dOutput = dOutput/dDist;
  double dMinCh = invQuantizeMin( dOutput, x1 ) ;
  double dMaxCh = invQuantizeMax( dOutput, x1 ) ;
  *dMin = min( dMinDist*dMinCh, dMaxDist*dMinCh );
  *dMax = max( dMinDist*dMaxCh, dMaxDist*dMaxCh );
  return true;
}

/*! This method returns the minimum value that generates dOutput as a quantized
    value when 'dQuantizeStep' is used as the quantized step.
    \param dOutput quantized output
    \param dQuantizeStep quantize step
    \return minimum value that when quantized produces 'dOutput' */
double WorldModel::invQuantizeMin( double dOutput, double dQuantizeStep )
{
  // q = quantize( V, Q ) = rint(V/Q)*Q -> q/Q = rint( V/Q)
  // min = rint(q/Q)-0.5 = V_min/Q -> V_min = (rint(q/Q)-0.5)*Q
  return (rint( dOutput / dQuantizeStep )-0.5 )*dQuantizeStep;
}

/*! This method returns the maximum value that generates dOutput as a quantized
    value when 'dQuantizeStep' is used as the quantized step.
    \param dOutput quantized output
    \param dQuantizeStep quantize step
    \return maximum value that when quantized produces 'dOutput' */
double WorldModel::invQuantizeMax( double dOutput, double dQuantizeStep )
{
  // q = quantize( V, Q ) = rint(V/Q)*Q -> q/Q = rint( V/Q)
  // max = rint(q/Q)+0.5 = V_max/Q -> V_max = (rint(q/Q)-0.5)*Q
  return (rint( dOutput/dQuantizeStep) + 0.5 )*dQuantizeStep;
}

/*! This method maps the information in the array of unknown players (players
    of which we do not know the number and/or team) to the player information
    located in the WorldModel. This is done by comparing the predicted position
    of players we haven't seen this cycle and the information of players of which
    we do not have the number and/or teamname.If this difference in the distance
    is smaller than tolerated (see PlayerSettings::getPlayerDistTolerance()),
    this player information is updated with the specified information.
    \param time time of the current cycle */
void WorldModel::mapUnknownPlayers( Time time)
{
  double      dist, dMinDist;
  VecPosition rel;
  ObjectT     o, o_new, objTmp;

  // for all unknown players, try to map it to closest teammate or opponent
  for( int j = 0; j < iNrUnknownPlayers; j ++ )
  {
    rel.setVecPosition( UnknownPlayers[j].getRelativeDistance(),
                        UnknownPlayers[j].getRelativeAngle(), POLAR );
    dist  = dMinDist = 1000.0;
    o     = UnknownPlayers[j].getType();
    o_new = OBJECT_ILLEGAL;
    if( ! SoccerTypes::isOpponent( o ) ) // TEAMMATE_UNKNOWN or PLAYER_UNKNOWN
    {
      for( int i = 0 ; i < MAX_TEAMMATES ; i++ )
      {
        objTmp = Teammates[i].getType();
        if( isConfidenceGood( objTmp ) && getTimeLastSeen( objTmp ) != time &&
            ( o == OBJECT_PLAYER_UNKNOWN ||
              UnknownPlayers[j].isInRange( objTmp ) ) )
        {
          dist = rel.getDistanceTo( Teammates[i].getRelativePosition( ) );
          if( dist < dMinDist )
          {
            o_new    = objTmp;
            dMinDist = dist;
          }
        }
      }
    }
    if( ! SoccerTypes::isTeammate( o ) ) // OPPONENT_UNKNOWN or PLAYER_UNKNOWN
    {
      for( int i = 0 ; i < MAX_OPPONENTS ; i++ )
      {
        objTmp = Opponents[i].getType();
        if( isConfidenceGood( objTmp ) && getTimeLastSeen( objTmp ) != time &&
            ( o == OBJECT_PLAYER_UNKNOWN ||
              UnknownPlayers[j].isInRange( objTmp ) ) )
        {
          dist = rel.getDistanceTo( Opponents[i].getRelativePosition( ) );
          if( dist < dMinDist )
          {
            o_new    = objTmp;
            dMinDist = dist;
          }
        }
      }
    }
    // if player found and in tolerated distance
    //   update player information.
    // else if not mapped to player
    //   put information in first player position of which we have no info.
    // if we do not know it is a teammate or opponent only assume opponent when
    // it is very close since then it is probably an opponent trying to get ball
    if( SoccerTypes::isKnownPlayer(o_new)
        && dMinDist < PS->getPlayerDistTolerance())
    {
      UnknownPlayers[j].setType( o_new );
      if( SoccerTypes::isTeammate( o_new ) )
        Teammates[SoccerTypes::getIndex(o_new)] = UnknownPlayers[j];
      else if( SoccerTypes::isOpponent( o_new ) )
        Opponents[SoccerTypes::getIndex(o_new)] = UnknownPlayers[j];
    }
    else if( UnknownPlayers[j].getType() == OBJECT_TEAMMATE_UNKNOWN )
    {
      o_new = getFirstEmptySpotInSet( OBJECT_SET_TEAMMATES, j );

      if( o_new != OBJECT_ILLEGAL )
      {
	/* cout << "["<<getPlayerNumber()<<"] "<<getCurrentTime() 
	     <<": Unknown teammate, so placing in first empty slot: "
	     <<SoccerTypes::getIndex(o_new)<<endl;
	cout << "closest player was "<<o_new<<" at "
	     <<dMinDist<<" tol="<<PS->getPlayerDistTolerance()<<endl;
	*/
        UnknownPlayers[j].setType( o_new );
        Teammates[SoccerTypes::getIndex(o_new)] = UnknownPlayers[j];
      }
    }
    else if( UnknownPlayers[j].getType() == OBJECT_OPPONENT_UNKNOWN )
    {                                      // could not map info to a player
      o_new = getFirstEmptySpotInSet( OBJECT_SET_OPPONENTS, j );

      if( o_new != OBJECT_ILLEGAL )
      {
        UnknownPlayers[j].setType( o_new );
        Opponents[SoccerTypes::getIndex(o_new)] = UnknownPlayers[j];
      }
    }
    else if( UnknownPlayers[j].getType() == OBJECT_PLAYER_UNKNOWN &&
             UnknownPlayers[j].getRelativeDistance() < SS->getVisibleDistance() )
    {
      o_new = getFirstEmptySpotInSet( OBJECT_SET_OPPONENTS );

      if( o_new != OBJECT_ILLEGAL )
      {
        UnknownPlayers[j].setType( o_new );
        Opponents[SoccerTypes::getIndex(o_new)] = UnknownPlayers[j];
      }
    }
  }

  iNrUnknownPlayers  = 0;
}

/*! This method updates the ServerSettings of this agent using the player type
    information of the heterogeneous player at index 'iIndex'. This method is
    usually called when the player type of the agent is changed by the coach.
    It updates its parameters, such that all calculations to determine the
    next action are based on the correct parameters.
    \param iIndex index of the new player type
    \return bool indicating whether update was succesfull. */
bool WorldModel::updateSSToHeteroPlayerType( int iIndex )
{
   SS->setPlayerSpeedMax( pt[iIndex].dPlayerSpeedMax );
   SS->setStaminaIncMax ( pt[iIndex].dStaminaIncMax  );
   SS->setPlayerDecay   ( pt[iIndex].dPlayerDecay    );
   SS->setInertiaMoment ( pt[iIndex].dInertiaMoment  );
   SS->setDashPowerRate ( pt[iIndex].dDashPowerRate  );
   SS->setPlayerSize    ( pt[iIndex].dPlayerSize     );
   SS->setKickableMargin( pt[iIndex].dKickableMargin );
   SS->setKickRand      ( pt[iIndex].dKickRand       );
   SS->setExtraStamina  ( pt[iIndex].dExtraStamina   );
   SS->setEffortMax     ( pt[iIndex].dEffortMax      );
   SS->setEffortMin     ( pt[iIndex].dEffortMin      );

   return true;
}

/*! This methods resets all the object information stored in the worldmodel.
    This means that the last see time of these objects are set to UnknownTime.
    \return bool indicating whether update was succesfull */
bool WorldModel::resetTimeObjects( )
{

  Ball.setTimeLastSeen             ( Time( -1, 0) );
  for( int i = 0 ; i < MAX_TEAMMATES ; i ++ )
    Teammates[i].setTimeLastSeen   ( Time( -1, 0) );
  for( int i = 0 ; i < MAX_OPPONENTS ; i ++ )
    Opponents[i].setTimeLastSeen   ( Time( -1, 0) );
  for( int i = 0 ; i < MAX_FLAGS     ; i ++ )
    Flags[i].setTimeLastSeen   ( Time( -1, 0) );
  for( int i = 0 ; i < MAX_LINES     ; i ++ )
    Lines[i].setTimeLastSeen   ( Time( -1, 0) );
  agentObject.setTimeLastSeen      ( Time( -1, 0) );
  return true;
}

/*! This method removes ghosts from the WorldModel. Ghosts are objects that
    should have been seen in the last see message but aren't. Currently only
    the ball is removed from the WorldModel. */
void WorldModel::removeGhosts( )
{
  AngDeg dAngle= getAgentHalfViewAngleValue(agentObject.getViewAngle());
  dAngle -= 0.15*dAngle; // make somewhat smaller for error

  if( fabs( getRelativeAngle( OBJECT_BALL ) ) < dAngle
       && getTimeLastSeen( OBJECT_BALL ) != getTimeLastSeeMessage() )
  {
    Log.log( 556, "ball not in cone: set time ball at -1 %f %f",
       fabs( getRelativeAngle( OBJECT_BALL ) ), dAngle );
    //cerr << "removing ball from world model"<<endl;
    //Ball.setTimeLastSeen( Time( -1, 0 ) );
    Ball.setTimeLastSeen( Ball.getTimeLastSeen() - 5 );
  }

  // ball should be "seen" when it is in visible distance. 0.9 is for noise
  if( fabs( 0.9*getRelativeDistance( OBJECT_BALL ) ) < SS->getVisibleDistance()
       && getTimeLastSeen( OBJECT_BALL ) != getTimeLastSeeMessage() )
  {
    Log.log( 556, "ball not in vis. dist: set time ball at -1 %f %f",
       fabs( getRelativeAngle( OBJECT_BALL ) ), dAngle );
    //cerr << "removing ball from world model"<<endl;
    //Ball.setTimeLastSeen( Time( -1, 0 ) );
    Ball.setTimeLastSeen( Ball.getTimeLastSeen() - 5 );
  }  

  // now remove all opponents. If an opponent is not seen while you expect him
  // somewhere, he has moved a lot. It's no use keeping his current position
  // in the world model, since he would still be considered in passing 
  // confidence, etc.
  int iIndex;
  for( ObjectT o = iterateObjectStart( iIndex, OBJECT_SET_OPPONENTS );
       o != OBJECT_ILLEGAL;
       o = iterateObjectNext ( iIndex, OBJECT_SET_OPPONENTS ) )
  {
    if( fabs( getRelativeAngle( o ) ) < dAngle
         && getTimeLastSeen( o ) != getTimeLastSeeMessage() )
    {
      Log.log( 556, "opp %d not in cone: set time at -1 %f %f",
         SoccerTypes::getIndex( o ) + 1, 
         fabs( getRelativeAngle( o ) ), dAngle );
      //setTimeLastSeen( o, Time( -1, 0 ) );
      setTimeLastSeen( o, getTimeLastSeen(o) - 5 );
    }
  }
}

/*! This method calculates the global position and velocity of the ball by means
    of a particle filter. Currently not used due to CPU constraints. 
    \param *posGlobal will be filled with the global position
    \param *velGlobal will be filled with the global velocity
    \return true when calculations were succesful, false otherwise */
bool WorldModel::calculateStateBall2( VecPosition *posGlobal,
                                     VecPosition *velGlobal )
{
  // initialize particles left with total number of particles
  int  iNrParticlesLeft = iNrParticlesBall;

  // check which particles are correct. If all gone, re-initialize all particles
  checkParticlesBall ( particlesPosBall, particlesVelBall,
                                         iNrParticlesBall, &iNrParticlesLeft );
  if( iNrParticlesLeft == 0 )
  {
    initParticlesBall ( particlesPosBall, particlesVelBall, iNrParticlesBall );
    iNrParticlesLeft = iNrParticlesBall;
  }

  // determine position and velocity and resample all deleted particles
  *posGlobal = averageParticles( particlesPosBall, iNrParticlesLeft );
  *velGlobal = averageParticles( particlesVelBall, iNrParticlesLeft );
  resampleParticlesBall( particlesPosBall, particlesVelBall,
                                           iNrParticlesBall, iNrParticlesLeft );

  // if we had no chance information, but but ball is within feel distance
  // and we have also seen it in the previous cycle -> use position based vel.
  if( Ball.getTimeChangeInformation( ) != getTimeLastSeen( OBJECT_BALL ) &&
      getRelativeDistance(OBJECT_BALL) < SS->getVisibleDistance() &&
      getTimeLastSeeMessage() - Ball.getTimeGlobalPosDerivedFromSee() == 1 )
  {
    // get the difference with the previous known global position
    // and subtract the position difference (this difference is caused by
    // the fact that the global position calculation contains a lot of noise
    // now the noise is filtered, since we compare the velocity as if its
    // position was seen "with the noise" of the last cycle).

    VecPosition posGlobalDiff = *posGlobal - Ball.getGlobalPositionLastSee()
                                           - agentObject.getPositionDifference();

    *velGlobal = posGlobalDiff*SS->getBallDecay();
  }

  // adapt velocity and position based on specific situations
  // set velocity to zero when in kick distance opponent for example
  if( getTimeSinceLastCatch() < 2 || getPlayMode() != PM_PLAY_ON )
    velGlobal->setMagnitude( 0.0 );
  else if( getNrInSetInCircle( OBJECT_SET_OPPONENTS,
                            Circle(*posGlobal,SS->getMaximalKickDist())) > 0 )
    velGlobal->setMagnitude( 0.0 );
  else if( velGlobal->getMagnitude() >
                           ( 1.0 + SS->getBallRand() )*SS->getBallSpeedMax() )
    velGlobal->setMagnitude( SS->getBallSpeedMax() );

  if( isBeforeKickOff() )
    posGlobal->setVecPosition( 0, 0 );

  return true;
}

/*! This method initializes all particles that represent the global position
    and global velocity of the ball. This is done by taking the last
    perceived information and generate particles within the range that can
    produce the perceived values. Note that each position is related to the
    velocity with the same index. The state of the ball is implicitly
    represented as a 4-tuple (pos_x, pos_y, vel_x, vel_y).
    \param posArray array that will be filled with the position particles
    \param velArray array that will be filled with the velocity particles
    \param iLength number of particles that have to be initialized. */
void WorldModel::initParticlesBall( VecPosition posArray[],
                                    VecPosition velArray[], int iLength )
{
  // declare a bunch of variables
  double dDistBall = UnknownDoubleValue, dDistChange = UnknownDoubleValue;
  AngDeg angBall   = UnknownAngleValue,  angChange   = UnknownAngleValue;
  double dMinDist, dMaxDist, dMinCh, dMaxCh, dDistTmp, dDistChTmp, dVelX, dVelY;
  AngDeg angChMin, angChMax,     angTmp,   angChTmp;

  // no information  received -> no initialization
  if( Ball.getTimeRelativePosition() != getTimeLastSeeMessage() )
    return;

  // get perceived values for distance and angle with ball
  dDistBall = getRelativeDistance( OBJECT_BALL );
  angBall   = getRelativeAngle( OBJECT_BALL );

  // get perceived values for distance and direction change
  if( Ball.getTimeChangeInformation( ) == getTimeLastSeeMessage() )
  {
    dDistChange = Ball.getRelativeDistanceChange();
    angChange   = Ball.getRelativeAngleChange();
  }

  // get ranges from which values could originate
  getMinMaxDistQuantizeValue( dDistBall, &dMinDist, &dMaxDist, 0.1, 0.1 );
  getMinMaxDistChange( dDistChange, dDistBall, &dMinCh, &dMaxCh, 0.02, 0.1,0.1);
  getMinMaxDirChange ( angChange, &angChMin, &angChMax, 0.1 );

  for( int i = 0; i < iLength; i ++ )
  {
    // make random distance and angle from range (angle is rounded) and make pos
    dDistTmp   = dMinDist + drand48()*fabs(dMaxDist - dMinDist); // angle->sign
    angTmp     = angBall  + drand48() - 0.5;

    posArray[i].setVecPosition( dDistTmp, angTmp, POLAR );
    posArray[i] = posArray[i].relativeToGlobal( getAgentGlobalPosition(),
						getAgentGlobalNeckAngle() );

    if( dDistChange != UnknownDoubleValue )
    {
      // make random values for direction and distance change and make velocity
      angChTmp   = angChMin + drand48()*(angChMax-angChMin);
      dDistChTmp = dMinCh   + drand48()*(dMaxCh-dMinCh);

      dVelX=dDistChTmp*cosDeg(angTmp)-Deg2Rad(angChTmp)*dDistTmp*sinDeg(angTmp);
      dVelY=dDistChTmp*sinDeg(angTmp)+Deg2Rad(angChTmp)*dDistTmp*cosDeg(angTmp);

      velArray[i].setVecPosition( dVelX, dVelY );
      velArray[i] = velArray[i].relativeToGlobal( getAgentGlobalVelocity(),
						  getAgentGlobalNeckAngle() );
    }
    else
      velArray[i].setVecPosition( 0, 0 );
  }
}

/*! This method checks all particles that represent the global position
    and global velocity of the ball. This is done by using the last
    perceived information. Using the possible ranges from which these values
    coule originate from, it can be checked which particles are legal. These
    particles are moved to the beginning of the array. Afterwards the number of
    legal particles is stored in 'iNrParticlesLeft'.
    \param posArray array that contains the position particles
    \param velArray array that contains the velocity particles
    \param iLength number of particles that have to be checked.
    \param iNrParticlesLeft will contain the number of legal particles*/
void WorldModel::checkParticlesBall( VecPosition posArray[],
             VecPosition velArray[], int iLength, int *iNrParticlesLeft )
{
  bool   bIllegal;
  double dMinDist, dMaxDist, dMinCh, dMaxCh, dMag;
  double dDistBall = UnknownDoubleValue, dDistChange = UnknownDoubleValue;
  AngDeg angBall   = UnknownAngleValue,  angChange   = UnknownAngleValue;
  double dDistChTmp;
  AngDeg angChTmp, angChMin, angChMax;
  VecPosition pos_rel, vel_rel;

  int i1, i2, i3, i4;
  i1 = i2 = i3 = i4 = 0;

  // no new perceptions, do not check
  if( getTimeLastSeen( OBJECT_BALL ) != getTimeLastSeeMessage() )
    return;

  // initialize values distance, direction, distance change and direction change
  // and get the associated ranges
  dDistBall = getRelativeDistance( OBJECT_BALL );
  angBall   = getRelativeAngle( OBJECT_BALL );
  getMinMaxDistQuantizeValue( dDistBall, &dMinDist, &dMaxDist, 0.1, 0.1 );

  if( getTimeLastSeen( OBJECT_BALL ) == Ball.getTimeChangeInformation( ) )
  {
    dDistChange = Ball.getRelativeDistanceChange();
    angChange   = Ball.getRelativeAngleChange();
    getMinMaxDirChange ( angChange, &angChMin, &angChMax, 0.1);
    getMinMaxDistChange( dDistChange, dDistBall, &dMinCh, &dMaxCh,0.02,0.1,0.1);

  }

  *iNrParticlesLeft = 0;
  for( int i = 0; i < iLength; i ++ )
  {
    // get particles and make them relative to the agent to compare
    pos_rel = posArray[i];
    vel_rel = velArray[i];
    pos_rel = pos_rel.globalToRelative( getAgentGlobalPosition(),
					getAgentGlobalNeckAngle() );
    vel_rel = vel_rel.globalToRelative( getAgentGlobalVelocity(),
					getAgentGlobalNeckAngle() );

    bIllegal = false;

    dMag = pos_rel.getMagnitude();
    if( dMag <  dMinDist || dMag > dMaxDist )
    {
      i1++;
      bIllegal = true;
    }
    if( fabs( VecPosition::normalizeAngle(pos_rel.getDirection() - angBall) )
         > 0.5 )
    {
      bIllegal = true;
      i2++;
    }

    if( dDistChange != UnknownDoubleValue )
    {
      dDistChTmp = (vel_rel.getX()*(pos_rel.getX()/dMag)) +
                   (vel_rel.getY()*(pos_rel.getY()/dMag));
      angChTmp   = Rad2Deg( ((vel_rel.getY()*pos_rel.getX()/dMag) -
                             (vel_rel.getX()*pos_rel.getY()/dMag)))/dMag ;

      if( angChTmp < angChMin || angChTmp > angChMax )
      {
        bIllegal = true;
        i3++;
      }
      if( dDistChTmp < dMinCh || dDistChTmp > dMaxCh )
      {
        bIllegal = true;
        i4++;
      }
    }

    // if not illegal, save particles and raise iNrParticlesLeft
    if( bIllegal == false )
    {
      posArray[*iNrParticlesLeft]     = posArray[i];
      velArray[(*iNrParticlesLeft)++] = velArray[i];
    }
  }
}

/*! This method updates all particles that represent the global position
    and global velocity of the ball to the next cycle. This is done using the
    same formula as the soccer server.
    \param posArray array that contains the position particles
    \param velArray array that contains the velocity particles
    \param iLength number of particles that have to be checked.
    \param dPower power with which the ball is accelerated.
    \param ang angle (relative to body) to which the ball is accelerated. */
void WorldModel::updateParticlesBall( VecPosition posArray[],
               VecPosition velArray[], int iLength, double dPower, AngDeg ang )
{
  double dRand = SS->getBallRand();
  double dMaxRand;

  for( int i = 0; i < iLength; i ++ )
  {
    // if power supplied, assume ball (and thus particles) are kicked
    if( dPower > EPSILON )
    {
      ang = VecPosition::normalizeAngle(ang + getAgentGlobalBodyAngle() );
      velArray[i] += VecPosition( getActualKickPowerRate()*dPower, ang, POLAR) ;
      if( velArray[i].getMagnitude() > SS->getBallSpeedMax() )
        velArray[i].setMagnitude( SS->getBallSpeedMax() );
    }

    // add noise in same way server does.
    dMaxRand = dRand * velArray[i].getMagnitude();
    velArray[i] += VecPosition(
                   (-1 + 2*drand48())*dMaxRand,
                   (-1 + 2*drand48())*dMaxRand );
    posArray[i] += velArray[i];
    velArray[i] *= SS->getBallDecay();
  }
}


/*! This method resamples all particles that represent the global position
    and global velocity of the ball. This is by copying existing legal
    particles at random. Since noise will be added to each particle, they will
    blur each time the particles are updated. The first 'iLeft' particles are
    legal particles and copied to the position at the end of the array.
    \param posArray array that contains the position particles
    \param velArray array that contains the velocity particles
    \param iLength number of particles that have to be checked.
    \param iLeft number of particles that are legal */
void WorldModel::resampleParticlesBall( VecPosition posArray[],
VecPosition velArray[], int iLength, int iLeft )
{
  int iRand = 0;
  for ( int i = iLeft; i < iLength; i ++ )
  {
    iRand = (int)(drand48()*iLeft);       // pick random particle
    posArray[ i ] = posArray[ iRand ];    // and copy contents
    velArray[ i ] = velArray[ iRand ];
  }
}

/*! This method receives the part of a see message that is not processed yet.
    Based on the ordening of teammates and opponents in this see message
    information can be extracted of the range of possible object types for the
    current object. This method returns the maximum object type that this object
    be by looking to the first object next in the remainder of the see
    message.
    \param obj last read object type
    \param strMsg remainder of the see messge
    \return maximum object type that 'obj' can possible be. */
ObjectT WorldModel::getMaxRangeUnknownPlayer( ObjectT obj, char* strMsg )
{
  if( obj == OBJECT_PLAYER_UNKNOWN )
    return OBJECT_ILLEGAL;

  list<ObjectT> l;
  ObjectT o;
  bool    isGoalie, bCont = true;
  int     i;
  bool    isTeammate = SoccerTypes::isTeammate( obj );
  ObjectT objMax = (isTeammate==true) ? OBJECT_TEAMMATE_11 : OBJECT_OPPONENT_11;

  while( bCont == true )
  {
    i = Parse::gotoFirstOccurenceOf( '(', &strMsg );
    if( i == -1 )
      bCont = false;                             // no more objects
    else
    {
      strMsg++;
      o = SoccerTypes::getObjectFromStr(&strMsg,&isGoalie,getTeamName());
      if( ( isTeammate  && SoccerTypes::isTeammate( o ) ) ||
          ( !isTeammate && SoccerTypes::isOpponent( o ) ) )
        l.push_back( o );
    }
  }

  while( ! l.empty() )
  {
    o = l.back();
    l.pop_back();
    if( SoccerTypes::isKnownPlayer( o ) )      // max range is one lower
      objMax = o;

    i = SoccerTypes::getIndex( objMax );
    if( isTeammate )
      objMax = SoccerTypes::getTeammateObjectFromIndex( i - 1 );
    else
      objMax = SoccerTypes::getOpponentObjectFromIndex( i - 1 );
  }

  return objMax;
}

