/*
 * A visitor class to handle DefineMsg types. 
 *
 * John Davin
 */

#ifndef MSGSORTERDEFINE_H
#define MSGSORTERDEFINE_H

#include "visitor.h"
#include "clangmsg.h"
#include "clangdefmsg.h"
#include "coach_lang_comp.h"
#include "hasa.h"
#include "config.h"
#include "msgstore.h"
#include "msgrulestore.h"
#include "RuleVisitor.h"
#include "NamedObjectUpdate.h"
#include "NamedObjectLookup.h"
#include "ClangIterator.h"
#include <iostream>


MsgStore<Cond*> defcondstorage;
MsgStore<Dir*> defdirstorage; 
MsgStore<Region*> defregstorage; 
MsgStore<Action*> defactstorage; 
MsgRuleStore defrulestorage; 


class MsgSorterDefine : public Def::Visitor
{
  /* Source of the advice */
  ActionBinding::SourceType sourceType;

 public:

  MsgSorterDefine(ActionBinding::SourceType source)
    : sourceType(source)
    {}

  /* Helper function to update named things */
  void pointerUpdate(const NamedObjectUpdate::UpdateType& ut,
		     NamedObjectUpdate namedUpdate)
    {
      /* Declare updating iterator */
      ClangIterator namedUpdateIterator(&namedUpdate);

      
      /* If a Cond or Region object was added/redefined in our storage, go
	 check our Cond storage and update the pointers of each Cond object */
      if( ut == NamedObjectUpdate::COND || ut == NamedObjectUpdate::REGION ) {

	for(MsgStore <Cond *>::MsgIterator i = defcondstorage.getIterBegin(); 
	    i != defcondstorage.getIterEnd(); i++)
	  
	  namedUpdateIterator.iterateOn((*i).second);
      }

      
      /* If a Region or Action object was added/redefined in our storage, go
	 check our Dir storage and update the pointers of each Dir object */
      if(ut == NamedObjectUpdate::REGION || ut == NamedObjectUpdate::ACTION) {

	for(MsgStore <Dir *>::MsgIterator i = defdirstorage.getIterBegin();
	  i != defdirstorage.getIterEnd(); i++)
	
	namedUpdateIterator.iterateOn((*i).second);
      }

      /* If a Region object was added/redefined in our storage, go check our
	 Region storage and update the pointers of each Region object */
      if( ut == NamedObjectUpdate::REGION ) {

	for(MsgStore <Region *>::MsgIterator i = defregstorage.getIterBegin(); 
	    i != defregstorage.getIterEnd(); i++)
	  
	  namedUpdateIterator.iterateOn((*i).second);
      
	
	/* And also check our Action storage and update the pointers of each
	   Action object */
	for(MsgStore <Action *>::MsgIterator i = defactstorage.getIterBegin();
	    i != defactstorage.getIterEnd(); i++)
	  
	  namedUpdateIterator.iterateOn((*i).second);
      }

      
      /* Finally, we have to go through our Rule storage and update the
	 pointers of each Rule object */
      for(MsgRuleStore::MsgIterator i = defrulestorage.getIterBegin(); 
	  i!=defrulestorage.getIterEnd(); i++)
	
	namedUpdateIterator.iterateOn((*i).second.ptr);
    

    }

  void visit( DefCond* msg ){
    //PR(4, "Define: DefCond received\n");

    /* Retrieve the Cond object from the DefCond message */
    Cond *cond = (msg->detachCond()).release();
    
    
    /* Retrieve the name of the Cond object */
    std::string name = msg->getName();


    /* Declare a NamedObjectLookup to link namedy thingys inside the cond
       object, and pass in our message storage pointers */
    NamedObjectLookup namedLookup(&defcondstorage, &defdirstorage,
				  &defregstorage, &defactstorage,
				  &defrulestorage);


    /* Cretae a Named lookup ClangIterator */
    ClangIterator namedLookupIterator(&namedLookup);

    /* Call iterate on the cond object */
    namedLookupIterator.iterateOn(cond);

        
    /* Add the Cond object to our storage */
    bool exists = defcondstorage.add(name, cond);


    /* We need to go through appropriate message storages and update their
       pointers if the Cond object already existed... */
    if(exists) {

      /* Call pointerUpdate function */
      NamedObjectUpdate namedUpdate(NamedObjectUpdate::COND, name, cond);
    
      pointerUpdate(NamedObjectUpdate::COND, namedUpdate);
    }
  }

  void visit( DefDir* msg ){
    //PR(4, "Define: DefDir received\n");
    
    /* Retrieve the Dir object from DefDir message */
    Dir *dir = (msg->detachDir()).release();
    
    /* Retrieve the name of Dir object */
    std::string name = msg->getName();


    /* Declare a NamedObjectLookup to link namedy thingys inside the cond
       object, and pass in our message storage pointers */
    NamedObjectLookup namedLookup(&defcondstorage, &defdirstorage,
    				  &defregstorage, &defactstorage,
				  &defrulestorage);
    
    /* Cretae a Named lookup ClangIterator */
    ClangIterator namedLookupIterator(&namedLookup);

    /* Call iterate on the Dir object */
    namedLookupIterator.iterateOn(dir);

    
    /* Add the Dir object to our storage */
    bool exists = defdirstorage.add(name, dir);


    /* Next, we need to go through appropriate message storages and update
       their pointers if the Dir object already existed */

    if(exists) {
      
      NamedObjectUpdate namedUpdate(NamedObjectUpdate::DIR, name, dir);
    
      /* Call pointerUpdate function */
      pointerUpdate(NamedObjectUpdate::DIR, namedUpdate);
    }
  }

  void visit( DefReg* msg ){
    //PR(4, "Define: DefReg received\n");

    /* Retrieve the Region object from DefReg message */
    Region *reg = (msg->detachReg()).release();

    
    /* Retrieve the region name */
    std::string name = msg->getName();


    /* Declare a NamedObjectLookup to link namedy thingys inside the cond
       object, and pass in our message storage pointers */
    NamedObjectLookup namedLookup(&defcondstorage, &defdirstorage,
    				  &defregstorage, &defactstorage,
				  &defrulestorage);
    
    /* Cretae a Named lookup ClangIterator */
    ClangIterator namedLookupIterator(&namedLookup);

    /* Call iterate on the Region object */
    namedLookupIterator.iterateOn(reg);


    /* Add the region to our storage */
    bool exists = defregstorage.add(name, reg);


    /* Next, we need to go through appropriate message storages and update
       their pointers if the Region object already existed */
    if(exists) {
      
      NamedObjectUpdate namedUpdate(NamedObjectUpdate::REGION, name, reg);
      
      /* Call pointerUpdate function */
      pointerUpdate(NamedObjectUpdate::REGION, namedUpdate);
    }
  }

  void visit( DefAct* msg ){
    //PR(4, "Define: DefAct received\n");

    /* Retrieve the Action object from DefAct message */
    Action *act = (msg->detachAction()).release();


    /* Get the name of the action */
    std::string name = msg->getName();


    /* Declare a NamedObjectLookup to link namedy thingys inside the cond
       object, and pass in our message storage pointers */
    NamedObjectLookup namedLookup(&defcondstorage, &defdirstorage,
				  &defregstorage, &defactstorage,
				  &defrulestorage);
    
    /* Cretae a Named lookup ClangIterator */
    ClangIterator namedLookupIterator(&namedLookup);

    /* Call iterate on the Action object */
    namedLookupIterator.iterateOn(act);


    bool exists = defactstorage.add(name, act);
    

    /* Next, we need to go through appropriate storages if our Action
       object already existed */
    if(exists) {

      NamedObjectUpdate namedUpdate(NamedObjectUpdate::ACTION, name, act);
      
      /* Call pointerUpdate function */
      pointerUpdate(NamedObjectUpdate::ACTION, namedUpdate);
    }
  }  

  void visit( DefRule* msg ){
    //PR(4, "Define: DefRule received\n");
    //cout << "visiting DefRule "<<msg->getName()<<": "<<*msg<<endl;

    /* Retrieve the Action object from DefAct message */
    Rule *rule = (msg->detachRule()).release();

    /* Retrieve the rule name */
    std::string name = msg->getName();

    //std::cout << "[MsgSorterDefine]: being adding rule named " << name << endl;

    /* The RulePair object we are adding to the storage; each RulePair object
       contains the Rule name, the pointer to the Rule object, and its
       source (Rulefile, coach, etc.) */
    RulePair pair;
    pair.rulename = name;
    pair.ptr = rule;
    pair.sourceType = sourceType;
  

    /* Declare a NamedObjectLookup to link namedy thingys inside the cond
       object, and pass in our message storage pointers */
    NamedObjectLookup namedLookup(&defcondstorage, &defdirstorage,
				  &defregstorage, &defactstorage,
				  &defrulestorage);
    
    /* Cretae a Named lookup ClangIterator */
    ClangIterator namedLookupIterator(&namedLookup);

    /* Call iterate on the Rule object */
    namedLookupIterator.iterateOn(rule);

    /* Add the rule to our storage */
    bool exists = defrulestorage.add(name, pair);


    /* Next, we need to go through all of our five message storages and
       update to the right pointer, if the Rule object already existed */
    if(exists) {

      NamedObjectUpdate namedUpdate(NamedObjectUpdate::RULE, name, rule);
      
      /* Call pointerUpdate function */
      pointerUpdate(NamedObjectUpdate::RULE, namedUpdate);
    }
  }

};


#endif // MSGSORTERDEFINE_H
