/*
    Coachable Agent communication encoding and decoding library
    Copyright (C) 2004, John Davin, Carnegie Mellon University

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2.1 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

/*
 * main.C - run some tests of the SayMsg library. 
 *
 * Creation Date: 11/18/03
 * Author: John Davin
 *
 *
 *
 */

#include "SayMsgEncoder.h"
#include "SayMsgDecoder.h"
#include "SampleSayMsgFactory.h"



int main(){
  bool rtn = true;

  /* run tests of encoding and decoding by first 
   * encoding a message, checking that it was encoded correctly by 
   * asserting that the result string equals the correct string. 
   * Then decode the message and print out the values. We should then manually
   * verify that the printed output from decoding matches the encoded information
   * (data values for X,Y,dX,dY will not match exactly, but should be within
   * the precision range for that datatype */

  /* TEST 1 */

  cout << "Test 1: Encoding: "<<endl;
  cout << "OurPos(20.,-10.)\nWantPass()"<<endl;
  SayMsgEncoder *myencoder = new SayMsgEncoder();
  rtn = myencoder->add( new OurPos(20.,-10.) ); 
  rtn &= myencoder->add( new WantPass() ); // add any other message units that you want.
  if(!rtn)
    cerr <<"[main]: WARNING: add failed!"<<endl;

  string str = myencoder->getEncodedStr();
  myencoder->clear();

  cout << "Encoded string: "<<str<<endl;
  assert(str == "0Op7");

  cout << "Decoding:"<<endl;
  /* run decoding tests */
   int playernum = 3; // you also get this with the say msg
   SayMsgFactory *msgfactory = new SampleSayMsgFactory();
   SayMsgDecoder *decoder = new SayMsgDecoder(msgfactory);
   bool success = decoder->decodeStr(str, playernum);
   if(!success) cerr << "[main]: failed to decode string!"<<endl;
   SayMsgDecoder::DecodedMsgIterator iter = decoder->getMsgIterator();
   for( ; iter != decoder->getMsgIteratorEnd(); iter++){
     // iterates over SayMsgTypes* pointers. 
     SayMsgTypes *curunit = *iter;
     curunit->process();
     // calls the process function that you defined for the current SayMsgType 
   }



  /* TEST 2 */
  cout << "-----------------------------\n"<<endl;
  cout << "Test 2: Encoding: "<<endl;
  cout << "BallPos(-52.,33.9,5)\nBallVel(1.2,-2.1,5)"<<endl;
  cout << "WeHaveBall(7)"<<endl;
  rtn = myencoder->add( new BallPos(-52.,33.9,5) ); 
  rtn &= myencoder->add( new BallVel(1.2,-2.1,5) ); 
  rtn &= myencoder->add( new WeHaveBall(7) );
  if(!rtn) cerr <<"[main]: WARNING: add failed!"<<endl;

  str = myencoder->getEncodedStr();
  myencoder->clear();

  cout << "Encoded string: "<<str<<endl;
  assert(str == "10_52Q8537");

  cout << "Decoding:"<<endl;
  /* run decoding tests */

  success = decoder->decodeStr(str, playernum);
  if(!success) cerr << "[main]: failed to decode string!"<<endl;

  for( iter = decoder->getMsgIterator(); iter != decoder->getMsgIteratorEnd(); iter++){
    // iterates over SayMsgTypes* pointers. 
    SayMsgTypes *curunit = *iter;
    curunit->process();
    // calls the process function that you defined for the current SayMsgType 
  }



  /* TEST 3 */
  cout << "-----------------------------\n"<<endl;
  cout << "Test 3: Encoding: "<<endl;
  cout << "TeammatePos(pnum 5, 0.,5., cycles 2)"<<endl;
  cout << "OppHasBall( 0 )"<<endl;
  cout << "PassToPlayer( 2 )"<<endl;
  cout << "PassToCoord(15,-4.5) - exceeds 10 char limit, should *not* be added."<<endl;

  rtn = myencoder->add( new TeammatePos(5, 0.,5.,2) ); // cost 4
  rtn &= myencoder->add( new OppHasBall(0) ); // cost 2
  rtn &= myencoder->add( new PassToPlayer(2) ); // cost 2
  rtn &= myencoder->add( new PassToCoord(15.,-4.5) ); // cost 3; add should fail!

  if(!rtn) cerr <<"[main]: WARNING: add failed!"<<endl;

  str = myencoder->getEncodedStr();
  myencoder->clear();

  cout << "Encoded string: "<<str<<endl;
  assert(str == "nAF24052");

  cout << "Decoding:"<<endl;
  /* run decoding tests */

  success = decoder->decodeStr(str, playernum);
  if(!success) cerr << "[main]: failed to decode string!"<<endl;

  for( iter = decoder->getMsgIterator(); iter != decoder->getMsgIteratorEnd(); iter++){
    // iterates over SayMsgTypes* pointers. 
    SayMsgTypes *curunit = *iter;
    curunit->process();
    // calls the process function that you defined for the current SayMsgType 
  }



  /* TEST 4 */
  cout << "-----------------------------\n"<<endl;
  cout << "Test 4: Encoding: "<<endl;
  cout << "OppPos(pnum 11, 0.,-34., cycles 20)"<<endl;
  cout << "PassToCoord( 52.9,33.9)"<<endl;
  cout << "PassToPlayer( 11 )"<<endl;

  rtn = myencoder->add( new OppPos(11, 0.,-34.,20) ); // cost 4
  rtn &= myencoder->add( new PassToCoord(52.9,33.9) ); // cost 3
  rtn &= myencoder->add( new PassToPlayer(11) ); // cost 2
  if(!rtn) cerr <<"[main]: WARNING: add failed!"<<endl;

  str = myencoder->getEncodedStr();
  myencoder->clear();

  cout << "Encoded string: "<<str<<endl;
  assert(str == "iA0k6__5b");

  cout << "Decoding:"<<endl;
  /* run decoding tests */

  success = decoder->decodeStr(str, playernum);
  if(!success) cerr << "[main]: failed to decode string!"<<endl;

  for( iter = decoder->getMsgIterator(); iter != decoder->getMsgIteratorEnd(); iter++){
    SayMsgTypes *curunit = *iter;
    curunit->process();
  }


  /* TEST 5 */
  cout << "-----------------------------\n"<<endl;
  cout << "Test 5: Encoding: "<<endl;
  cout << "OurPos(20.,-10., cycles 5)"<<endl;
  cout << "BallPos( 30.,29., cycles 5)"<<endl;
  cout << "PassToPlayer( 4 )"<<endl;

  rtn = myencoder->add( new OurPos(20.,-10.) ); // cost 3
  rtn &= myencoder->add( new BallPos(30.,29.,5) ); // cost 4
  rtn &= myencoder->add( new PassToPlayer(4) ); // cost 2
  if(!rtn) cerr <<"[main]: WARNING: add failed!"<<endl;

  str = myencoder->getEncodedStr();
  myencoder->clear();

  cout << "Encoded string: "<<str<<endl;
  assert(str == "0Op1V*554");

  cout << "Decoding:"<<endl;
  /* run decoding tests */

  success = decoder->decodeStr(str, playernum);
  if(!success) cerr << "[main]: failed to decode string!"<<endl;

  for( iter = decoder->getMsgIterator(); iter != decoder->getMsgIteratorEnd(); iter++){
    SayMsgTypes *curunit = *iter;
    curunit->process();
  }





  /****** TEST 6 ********/
  cout << "-----------------------------\n"<<endl;
  cout << "Test 6: Encoding: "<<endl;
  cout << "BallPos( 15.,-15., 10 cycles )"<<endl;
  cout << "BallVel( 1.2, -1.2, 10 cycles )"<<endl;
  cout << "WeHaveBall( 0 )"<<endl;

  rtn = myencoder->add( new BallPos(15., -15., 10) ); // cost 4
  rtn &= myencoder->add( new BallVel(1.2, -1.2, 10) ); // cost 4
  rtn &= myencoder->add( new WeHaveBall( 0 ) ); // cost 2
  if(!rtn) cerr <<"[main]: WARNING: add failed!"<<endl;

  str = myencoder->getEncodedStr();
  myencoder->clear();

  cout << "Encoded string: "<<str<<endl;
  assert(str == "1Kka2Qka30");

  cout << "Decoding:"<<endl;
  /* run decoding tests */

  success = decoder->decodeStr(str, playernum);
  if(!success) cerr << "[main]: failed to decode string!"<<endl;

  for( iter = decoder->getMsgIterator(); iter != decoder->getMsgIteratorEnd(); iter++){
    SayMsgTypes *curunit = *iter;
    curunit->process();
  }




  /****** TEST 7 ********/
  cout << "-----------------------------\n"<<endl;
  cout << "Test 7: Encoding: "<<endl;
  cout << "OppHasBall( 7 )"<<endl;
  cout << "OppPos(pnum 7, 15., -15., cycles 3)"<<endl;
  cout << "WantPass()"<<endl;

  rtn = myencoder->add( new OppHasBall(7) ); // cost 2
  rtn &= myencoder->add( new OppPos(7, 15.,-15.,3) ); // cost 4
  rtn &= myencoder->add( new WantPass() ); // cost 1
  if(!rtn) cerr <<"[main]: WARNING: add failed!"<<endl;

  str = myencoder->getEncodedStr();
  myencoder->clear();

  cout << "Encoded string: "<<str<<endl;
  assert(str == "47eKk37");

  cout << "Decoding:"<<endl;
  /* run decoding tests */

  success = decoder->decodeStr(str, playernum);
  if(!success) cerr << "[main]: failed to decode string!"<<endl;

  for( iter = decoder->getMsgIterator(); iter != decoder->getMsgIteratorEnd(); iter++){
    SayMsgTypes *curunit = *iter;
    curunit->process();
  }


  /****** TEST 8 ********/
  cout << "-----------------------------\n"<<endl;
  cout << "Test 8: Confirm failure of attempt to decode illegal string: "<<endl;
  success = decoder->decodeStr("ueopueht", 4);
  success |= decoder->decodeStr("%^#(*&^(*&(^*^%", 4);
  if(!success) cout << "[main]: failed to decode string (Test Passed)"<<endl;
  else cerr << "[main]: ERROR: decoded string that we shouldn't have been able to."<<endl;

  cout <<endl;

  return 0;
}
